#include "BrokerDefines.h"
#include "BrokerInterface.h"
#include "ProxyLocalBrokerInstance.h"
#include "ProxyModuleInstance.h"
#include "SBW.h"
#include "Broker.h"

#include "SBWBrokerRPC.h"
#include "StartupNotification.h"
#include "BrokerInstance.h"

#include "SBW/SBW.h"
#include "SBW/SBWApplicationException.h"
#include "SBW/SBWModuleStartException.h"
#include "SBW/SBWLowLevel.h"

#ifndef WIN32
#include <ext/hash_set>
#include <string>
using namespace __gnu_cxx;
#else
#include <hash_set>
#endif

using namespace SystemsBiologyWorkbench;
using namespace SystemsBiologyWorkbench::Broker;
using namespace std;

SystemsBiologyWorkbench::Broker::Broker::Broker()
{
	rpc = dynamic_cast<SBWBrokerRPC *>(SystemsBiologyWorkbench::Broker::SBW::getInternalAPI()); 
	instanceTable = InstanceTable::getTable();
	moduleRegistry = ModuleRegistry::getRegistry();

	addBrokerRegistryEntry();
	addBrokerInstance();
}
void SystemsBiologyWorkbench::Broker::Broker::addBrokerInstance()
{
	instanceTable->addInstance(new ProxyLocalBrokerInstance());
}
void SystemsBiologyWorkbench::Broker::Broker::addBrokerRegistryEntry()
{
	DataBlockWriter *oWriter = new DataBlockWriter();
	(*oWriter) << "BROKER" << "SBW Broker" <<  UNIQUE <<  "" << "The SBW Broker manages all modules.";
	ModuleDescriptor *md = new ModuleDescriptor(oWriter->getReader());
	delete oWriter;
	oWriter = new DataBlockWriter();
	(*oWriter) << "BROKER" << "BROKER" << "BROKER" << "/" << "The BROKER service implements the services of the SBW Broker.";
	md->addService(new ServiceDescriptor( oWriter->getReader(), *md) );
	moduleRegistry->addModuleDescriptor("BROKER", md, false);
	delete oWriter;
}
int SystemsBiologyWorkbench::Broker::Broker::getModuleInstance(string name)
{
	TRACE("Received request for instance of '" << name << "'");
	string moduleName = getModuleNamePart(name);
	Host *host = getHostPart(name);
	if (host == NULL || rpc->getHost()->equals(host))
		return getModuleInstanceLocal(moduleName);
	else
		return getModuleInstanceRemote(moduleName, host);
}
int SystemsBiologyWorkbench::Broker::Broker::getModuleInstanceLocal(string moduleName)
{
	Instance *instance = instanceTable->getInstance(rpc->getHost(), moduleName);
	ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);

	if (instance != NULL)
	{
		// If we find an instance (meaning there's already a copy of
		// the module running), and its management type is UNIQUE, then
		// return the existing instance, else try to start up a new one.

		if (md == NULL)
		{
			// There's an instance, but the module isn't registered.
			// Unfortunately, without registration info, we can't
			// know how to start up an instance of the module.  So punt.

			TRACE("Module '" << moduleName << "' is not registered;"  <<" returning existing instance.");
			return instance->getId();
		}
		else if (md->getManagementType() == UniqueModule)
		{
			TRACE("Reusing instance of unique module '" << moduleName <<  "' on '" << rpc->getHost()->getName() << "'");
			return instance->getId();
		}
		else
			return startNewLocalModuleInstance(md);
	}
	else  // There's no instance yet, so start one if we know how.
	{
		TRACE("No existing instance of '" << moduleName << "' found");

		if (md != NULL)
			return startNewLocalModuleInstance(md);
		else
		{
			stringstream a,b;
			a <<  "Unknown/unregistered module '" << moduleName << "'";
			b << "An attempt was made to obtain an instance of a module"
				<< " named '" << moduleName << "' from the broker on '"
				<< rpc->getHost()->getName() << "', but no such module is known";
			throw new SBWModuleNotFoundException(a.str(), b.str());
		}
	}
}
int SystemsBiologyWorkbench::Broker::Broker::getModuleInstanceRemote(string moduleName, Host *host)
{
	int brokerId = getBrokerId(host);

	// Part A: If the module is unique & there is an instance running,
	// we return that w/o contacting the remote broker.  If the module
	// is unregistered (==> descriptor null) & *not* unique, then we try
	// to return a running instance if there is one, or else we fail.
	// If the module is registered and not unique, we go to part B below.

	DataBlockWriter md = getModuleDescriptorRemote(moduleName, host, true);
	string sTemp; Integer nManagement;
	md.getReader() >> sTemp >> sTemp >> nManagement;

	if (nManagement == UNIQUE || nManagement == SELF_MANAGED)
	{
		TRACE("'" << host->getName() << ":" << moduleName << "' " << (nManagement == UNIQUE ? " is unique" : "is self managed" ));

		// Look for a running instance of this module on the broker.

		TRACE("Checking if an instance of '" << host->getName() << ":"
			<< moduleName << " is already running");

		Instance *inst = instanceTable->getInstance(host, moduleName);
		if (inst != NULL)
		{
			TRACE("Returning existing instance " << inst->getId());
			return inst->getId();
		}

		throw new SBWModuleNotFoundException(
			string("Unknown/unregistered module '") + host->getName()
			+ ":" + moduleName + "'",
			"An attempt was made to obtain an instance of a module"
			+ string(" named '") + moduleName + "' from the broker on '"
			+ host->getName() + "', but no such module is known");
	} 
	else
	{
		throw new SBWModuleNotFoundException(
			string("The module '") + host->getName()
			+ ":" + moduleName + "' is not running",
			"An attempt was made to obtain an instance of a module"\
			" named '" + moduleName + "' from the broker on '"
			+ host->getName() + "', but that module is currently"\
			"not running. Remote startup is not allowed.");
	}
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptor(string name, bool includeRunning)
{
	string moduleName = getModuleNamePart(name);
	Host *host = getHostPart(name);

	if (host == NULL || rpc->getHost()->equals(host))
		return getModuleDescriptorLocal(moduleName, includeRunning);
	else
		return getModuleDescriptorRemote(moduleName, host, includeRunning);
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptorLocal(string moduleName, bool includeRunning)
{
	ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);
	if (md != NULL)
	{
		TRACE("Found '" << moduleName << "' in the local registry");
		return listForModuleDescriptor(md);
	}
	else
	{
		// It's not in the registry.  Should we check running instances?

		TRACE("'" + moduleName + "' is not registered");
		if (includeRunning)
		{
			// We didn't find a module descriptor, but the module may
			// be running unregistered.  Go through the instances.
			// Note that we want to return the most recently started
			// module by this name, so we start from the end of the list.

			vector< Instance *> instances = instanceTable->getInstances(BROKER_MODULE);
			for (int i = instances.size() - 1; i >= 0; i--)
			{
				Instance *inst = instances[i];
				if (inst->getName() == moduleName)
				{
					TRACE("Found running instance of '" <<  moduleName << "'");
					return getModuleDescriptorLocal(inst->getId(), inst);
				}
			}
		}
	}

	// If we get here, we didn't find it.
	TRACE("There is no module with name '" << moduleName << "'");
	throw new SBWModuleNotFoundException( "Unknown/unregistered module '" + moduleName + "'", 
		"An attempt was made to obtain an instance of a module named '" + moduleName + "', but no such module is known to the SBW registry.");
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptorRemote(string moduleName, Host *host,
																				   bool includeRunning)
{
	TRACE("Asking broker on " + host->getName()
		+ " about '" + moduleName + "'");

	BrokerInterface *broker = getBrokerInterface(host);
	DataBlockWriter result = broker->getModuleDescriptor(moduleName, includeRunning);
	return mungeModuleDescriptor(result, moduleName, host);
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptor(int moduleId)
{
	checkModuleId(moduleId);
	Instance *inst = getInstanceForModuleId(moduleId);

	if (inst->getBrokerId() == BROKER_MODULE)
		return getModuleDescriptorLocal(moduleId, inst);
	else
		return getModuleDescriptorRemote(moduleId, inst);
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptorLocal(int moduleId, Instance *moduleInst)
{	
	string moduleName = moduleInst->getName();

	if (StringUtil::empty(moduleName))
	{
		// We may have been given a module id for a module connected
		// anonymously.  In that case, there will be no names or command
		// line associated with it.  Treat it as a unique module.

		TRACE("Module " << moduleId << " is anonymous");
		return listForModuleDescriptor("","",UniqueModule,"","");
	}
	else
	{
		// We have an instance and the module is not anonymous.  See if
		// it's in the registry.  If it's not in the registry, it's a
		// module that connected with a name but did not register.  In
		// that case, return what info we can.

		ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);
		if (md != NULL)
		{
			TRACE("Found module " << moduleId << " ('" << moduleName << "'), in the registry")
				return listForModuleDescriptor(md);
		}
		else
		{
			TRACE("Module " << moduleId << " is not in the registry");
			return listForModuleDescriptor(moduleName, "", UniqueModule , "", "");
		}
	}
}
void SystemsBiologyWorkbench::Broker::Broker::checkBrokerId(int brokerId)
{
	BrokerInstance *bInst = dynamic_cast<BrokerInstance *>( instanceTable->getInstance(brokerId));
	if (bInst == NULL)
	{
		vector <Instance *> instances = instanceTable ->getInstances(brokerId);
		vector <Instance *>::iterator oIterator;
		for (oIterator = instances.begin(); oIterator != instances.end(); oIterator++)
		{
			Instance* oInstance = *oIterator;
			ProxyModuleInstance *oModule = dynamic_cast<ProxyModuleInstance *>(oInstance);
			try
			{
				if (oModule != NULL)
				{
					rpc->notifyOthersOfDisconnect(oModule,false);
					instanceTable->removeInstance(oModule);
				}
			}
			catch(...)
			{
				//
			}
		}
		throw new SBWApplicationException(("error broker: " + Properties::toString(brokerId) + " is no longer reachable").c_str());
	}
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptorRemote(int moduleId, Instance *moduleInst)
{
	// This must be a remote instance, so cast it to access getRemoteId().

	ProxyModuleInstance *oModule = dynamic_cast<ProxyModuleInstance *>(moduleInst);
	BrokerInstance *oBroker = dynamic_cast<BrokerInstance *>(moduleInst);
	if (oModule != NULL)
	{
		BrokerInterface *broker = getBrokerInterface(oModule->getBrokerId());

		// Note that moduleId refers to a module on the local broker --
		// must translate it to what it is on the *remote* broker.

		DataBlockWriter oWriter = broker->getModuleDescriptor(oModule->getRemoteId());

		return mungeModuleDescriptor(
			oWriter, 
			oModule->getName(), 
			oModule->getHost());
	}
	if (oBroker != NULL)
	{
		BrokerInterface *broker = getBrokerInterface(oBroker->getBrokerId());

		// Note that moduleId refers to a module on the local broker --
		// must translate it to what it is on the *remote* broker.

		DataBlockWriter oWriter = broker->getModuleDescriptor(oBroker->getRemoteId());

		return mungeModuleDescriptor(
			oWriter, 
			oBroker->getName(), 
			oBroker->getHost());
	}
	throw new SBWApplicationException("instance was no remote instance");
}
vector< DataBlockWriter > SystemsBiologyWorkbench::Broker::Broker::getModuleDescriptors(bool localOnly, bool includeRunning)
{
	vector< DataBlockWriter > records;
	vector<int> brokerIds;
	if (localOnly)                  // Only want local broker's modules.
		brokerIds.push_back(BROKER_MODULE );
	else                            // Get all broker's modules.
		brokerIds = getBrokerIds();

	// Ask each broker for a list of descriptors for registered modules,
	// and if includeRunning is true, also check our table of running
	// instances to see if any unregistered instances are known to be
	// running on that broker.

	for (unsigned int i = 0; i < brokerIds.size(); i++)
	{
		int brokerId = brokerIds[i];
		vector<ModuleDescriptor *> mds;

		try
		{
			mds = getDescriptors(brokerId);

			for (unsigned int j = 0; j < mds.size(); j++)
				records.push_back(listForModuleDescriptor(mds[j]));

			if (includeRunning)
			{
				// We have to combine the info in the registry with
				// the known running instances, since some modules may
				// not have registered.  Modules that haven't
				// registered won't have a registry entry, but will
				// have an instance.

				vector <Instance *> instances = instanceTable->getInstances(brokerId);
				for (unsigned int j = 0; j < instances.size(); j++)
				{
					string name = instances[j]->getName();
					if (getDescriptor(brokerId, name) == NULL)
						records.push_back(listForModuleDescriptor(name, "",
						UniqueModule , "", ""));
				}
			}
		}
		catch (...)
		{
			TRACE("Lost contact with SBW Broker " << brokerId);
			continue;
		}
	}

	return records;
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptor(int moduleId, int serviceId)
{
	vector< string > serviceNames = SBW::getServiceNamesFromModule(moduleId);
	if (serviceId >= 0 && serviceId <= (int)serviceNames.size()) 
	{
		return getServiceDescriptor(moduleId, serviceNames[serviceId]);
	}
	else
	{
		throw new SBWServiceNotFoundException("Service id " + Properties::toString(serviceId) + " is out of range" + 
			"Service id " + Properties::toString(serviceId) + " is not within the range of"
			+ " service identifiers recognized by module " + Properties::toString(moduleId));
	}
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptor(int moduleId, string servName)
{
	checkModuleId(moduleId);
	Instance *inst = getInstanceForModuleId(moduleId);

	if (inst->getBrokerId() == BROKER_MODULE)
		return getServiceDescriptorLocal(moduleId, servName, inst);
	else
		return getServiceDescriptorRemote(moduleId, servName, inst);
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptorLocal(int moduleId, string servName,
																				   Instance *moduleInst)
{
	servName = normalizeServiceName(servName);
	string moduleName = moduleInst->getName();
	ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);

	// Easiest case: module is registered & has a service descriptor.
	if (md != NULL)
	{
		vector<ServiceDescriptor *> servicesHash = md->getServices();
		if (servicesHash.size() > 0 )
		{
			for (unsigned int i = 0; i < servicesHash.size(); i++)
			{
				ServiceDescriptor *sd = servicesHash[i];
				if (sd != NULL && sd->getName() == servName)
					return listForServiceDescriptor(moduleName, sd);
			}
		}
	}
	// Harder case: there's a registry entry for the module, but no
	// associated ServiceDescriptor for the service.  We may have been
	// given a module id for a module connected anonymously.  In that
	// case, there will be no service record available for it, and the
	// best we can do is return a service descriptor with only the name
	// of the service in it.  However, to be safe, we first check with
	// the module itself to make sure that a service by that name
	// really does exist.

	vector<string> serviceNames = SBW::getServiceNamesFromModule(moduleId);
	for (unsigned int i = 0; i < serviceNames.size(); i++)
		if (serviceNames[i] == servName)
			return listForServiceDescriptor(moduleName, servName,
			"", "", "");

	// If we get to this point here, we didn't find it.

	throw new SBWServiceNotFoundException(
		"Cannot find a service named '" + servName + "' on " + moduleName
		+ "A service descriptor for a service named '" + servName
		+ "' has been requested from " + moduleName
		+ " but the module does not define a service by that name.");
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptorRemote(int moduleId, string servName,
																					Instance *moduleInst)
{
	ProxyModuleInstance *oProxy = dynamic_cast<ProxyModuleInstance *>(moduleInst);
	if (oProxy != NULL)
	{
		BrokerInterface *broker = getBrokerInterface(moduleInst->getBrokerId());

		// Note that moduleId refers to a module on the local broker --
		// must translate it to what it is on the *remote* broker.

		DataBlockWriter list = broker->getServiceDescriptor(oProxy->getRemoteId(), servName);
		return mungeServiceDescriptor(list, moduleInst->getFullName(),moduleInst->getHost());
	}
	BrokerInstance *oBroker = dynamic_cast<BrokerInstance *> (moduleInst);
	if (oBroker != NULL)
	{
		BrokerInterface *broker = getBrokerInterface(moduleInst->getBrokerId());

		// Note that moduleId refers to a module on the local broker --
		// must translate it to what it is on the *remote* broker.

		DataBlockWriter list = broker->getServiceDescriptor(oBroker->getRemoteId(), servName);
		return mungeServiceDescriptor(list, moduleInst->getFullName(),moduleInst->getHost());
	}

	return DataBlockWriter();
}
vector<DataBlockWriter> SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptors(int moduleId)
{
	checkModuleId(moduleId);
	Instance *inst = getInstanceForModuleId(moduleId);

	if (inst->getBrokerId() == BROKER_MODULE)
		return getServiceDescriptorsLocal(inst->getName());
	else
	{
		try
		{
			return getServiceDescriptorsRemote(inst->getName(), inst->getHost());
		}
		catch (SBWException *e)
		{
			// We should never get this -- we have a module id, therefore
			// the corresponding remote broker should be running.

			TRACE("Problem contacting broker " << inst->getBrokerId() << e->getMessage());
			throw new SBWCommunicationException(
				"Problem communicating with the Broker responsible for module "
				+ Properties::toString(moduleId),
				"Although the module identifier " + Properties::toString(moduleId) + " is known to"
				+ " SBW, the SBW Broker responsible for that module appears to"
				+ " be unreachable or otherwise unavailable.");
			delete e;
		}
	}
}
vector<DataBlockWriter> SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptors(string name)
{
	string moduleName = getModuleNamePart(name);
	Host *host = getHostPart(name);

	if (host == NULL|| rpc->getHost()->equals(host))
		return getServiceDescriptorsLocal(moduleName);
	else
		return getServiceDescriptorsRemote(moduleName, host);
}
vector<DataBlockWriter> SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptorsLocal(string moduleName)
{
	vector<DataBlockWriter> oResult;
	moduleName = normalizeModuleName(moduleName);

	// If the module is not registered, it's an exception.  If it's
	// registered, return the services found (even if that means
	// returning a null list).

	ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);
	if (md == NULL) 
	{
		stringstream a,b;
		a << "Unknown/unregistered module '" << moduleName << "'";
		b << "An attempt was made to obtain a list of services for a"
			<< " module named '" << moduleName << "', but no such module"
			<< " is known to the SBW registry.";
		throw new SBWModuleNotFoundException( a.str(),b.str());
	}
	else
	{
		vector<ServiceDescriptor *> services = md->getServices();
		if (services.size()  == 0)
			return oResult;

		for (unsigned int j = 0; j <  services.size(); j++ )
		{                
			oResult.push_back(listForServiceDescriptor(moduleName, services[j]));
		}
		return oResult;
	}
}
vector<DataBlockWriter> SystemsBiologyWorkbench::Broker::Broker::getServiceDescriptorsRemote(string moduleName, Host *host)
{
	BrokerInterface *broker = getBrokerInterface(host);
	vector<DataBlockWriter> lists = broker->getServiceDescriptors(moduleName);

	for (unsigned int i = 0; i < lists.size(); i++)
		lists[i] = mungeServiceDescriptor(lists[i], moduleName, host);

	return lists;
}
vector<int> SystemsBiologyWorkbench::Broker::Broker::getExistingModuleInstanceIds()
{
	vector<int> oTemp;
	vector<Instance *> instances = instanceTable->getInstances();
	for (unsigned int i = 0; i < instances.size(); i++)
		oTemp.push_back( instances[i]->getId());

	return oTemp;
}
vector<string> SystemsBiologyWorkbench::Broker::Broker::getServiceCategories(string parentCategory)
{
	parentCategory = SBW::normalizeCategoryName(parentCategory);

	if (parentCategory[0] == '/')
		parentCategory = parentCategory.substr(1);

	// This is not an especially efficient implementation, but I don't
	// think it needs to be. This method isn't likely to be used often,
	// so there doesn't seem to be a point to making this super-fast.

	// Using hashset here is just a simple way of getting unique keys.
	// The "* 5" is just an estimate of the space that might be needed.
	// The number is not founded in any empirical data whatsoever :-).

	vector<ModuleDescriptor *> mds = getDescriptors();	
	hash_set< string > categories;

	for (unsigned int i = 0; i < mds.size(); i++)
	{
		vector<DataBlockWriter> sds = getServiceDescriptors(mds[i]->getName());
		for (unsigned int j = 0; j < sds.size(); j++)
		{
			DataBlockWriter oWriter = sds[j];
			string sTemp;
			oWriter.getReader() >> sTemp >> sTemp >> sTemp >> sTemp;
			categories.insert( sTemp );
		}
	}

	// Now we have a list of unique category names.
	// Do a substring match to find subcategories of the given parent.

	hash_set< string > subcategories;
	hash_set <string> ::iterator oIterator;
	for ( oIterator = categories.begin(); oIterator != categories.end(); oIterator++ )
	{
		string thisCat = SBW::normalizeCategoryName( *oIterator );

		if (thisCat[0] == '/')
			thisCat = thisCat.substr(1);

		if (parentCategory.length() == 0)
		{
			if (thisCat.length() != 0)
				subcategories.insert( thisCat );
		}
		else if (thisCat.find(parentCategory) == 0 &&
			thisCat.length() > parentCategory.length() &&
			thisCat[parentCategory.length()] == '/')
		{
			subcategories.insert (thisCat.substr(parentCategory.length() + 1));
		}
	}

	if (subcategories.size() < 1)
		return vector<string>();
	else
	{
		vector<string> oResult;
		hash_set< string >::iterator oSubIterator;
		for (oSubIterator = subcategories.begin(); oSubIterator != subcategories.end(); oSubIterator++)
			oResult.push_back( *oSubIterator );
		return oResult;
	}
}
vector < DataBlockWriter > SystemsBiologyWorkbench::Broker::Broker::findServices(string category, bool recursive)
{
	category = SBW::normalizeCategoryName(category);

	// This is not an especially efficient implementation, but I don't
	// think it needs to be.  The number of modules and categories is
	// likely to always remain small (e.g., less than 100) and this
	// method isn't likely to be used often, so there doesn't seem to
	// be a point to making this super-fast.

	vector< ModuleDescriptor *> mds = getDescriptors();
	vector < DataBlockWriter > descriptors;

	for (unsigned int i = 0; i < mds.size(); i++)
	{
		vector<DataBlockWriter> sds = getServiceDescriptors(mds[i]->getName());
		for (unsigned int j = 0; j < sds.size(); j++)
		{
			ServiceDescriptor *oTemp = new ServiceDescriptor (sds[j].getReader(),*mds[i]);
			string thisCategory = oTemp->getCategory ();
			if ((!recursive && thisCategory.find(category) == 0)
				|| (recursive && matchCategory(thisCategory, category, 0)))
				descriptors.push_back ( listForServiceDescriptor(mds[i]->getName(),oTemp));
		}
	}
	return descriptors;
}
void SystemsBiologyWorkbench::Broker::Broker::registerModule(string moduleName, string moduleDisplayName,
															 int moduleType, string commandLine, string help)
{
	// This has a lot of similarity to what ModuleImpl.initModuleImpl
	// does.  We have to do the same tests here too b/c this method can
	// be called from the low-level API directly.

	if (StringUtil::empty(moduleName))
		throw new SBWModuleDefinitionException(
		"Null or empty module name",
		"'Null' and empty strings are invalid module names.");

	if (getHostPart(moduleName) != NULL)
		throw new SBWModuleDefinitionException(
		"Module registration cannot use host:modulename format",
		"The program attempted to register a module whose name"\
		" has the form 'host:modulename'.  This form is illegal"\
		" during module registration because modules are always"\
		" registered with their local broker only.");

	if (StringUtil::empty(moduleDisplayName))
		throw new SBWModuleDefinitionException(
		"Null or empty module display name",
		"Module display names cannot be null or empty strings.");

	if (moduleType != UNIQUE
		&& moduleType != SELF_MANAGED){
			string msg = "Invalid module type specified";
			string detailed = "Module type must be one of ModuleImpl.SELF_MANAGED"\
				"ModuleImpl.SBW_MANAGED or ModuleImpl.UNIQUE";
			throw new SBWModuleDefinitionException(msg, detailed);
		}

		if (StringUtil::empty(commandLine)) {
			string msg = "Empty command line given for module";
			string detailed = "The command line for starting the module must be"\
				" non-empty; otherwise, the SBW Broker will not"\
				" be able to start up the module.";

			throw new SBWModuleDefinitionException(msg,detailed);
		}

		//moduleName = moduleName.trim();

		TRACE("Registering module '" << moduleName
			<< "', display name '" << moduleDisplayName
			<< "', type '" << moduleType
			<< "', cmdLine = " << commandLine
			<< "', help = '" << help << "'")

			moduleRegistry->addModuleDescriptor(moduleName, moduleDisplayName,
			moduleType, commandLine, help);
		rpc->notifyOthersOfRegistrationChange(BROKER_MODULE, true);
}
void SystemsBiologyWorkbench::Broker::Broker::changeModuleName(int moduleId, string newModuleName)
{
	checkModuleId(moduleId);

	// Before going much further, check if newModuleName is an
	// acceptable module name.  The high-level APIs also check this,
	// but we have to do it again here because this method may be
	// called from the low-level API.

	if (getHostPart(newModuleName) != NULL)
	{
		throw new SBWModuleDefinitionException(
			"New module name cannot have host:modulename format",
			"The program attempted to change the name of module "
			+ Properties::toString(moduleId)
			+ " to a module whose name has the form 'host:modulename'."
			+ "  This form is illegal in a call to changeModuleName.");
	}

	// The new name looks ok, so get on with it.

	Instance *inst = getInstanceForModuleId(moduleId);

	if (inst->getBrokerId() == BROKER_MODULE)
		changeModuleNameLocal(moduleId, inst, newModuleName);
	else
		changeModuleNameRemote(moduleId, inst, newModuleName);
}
void SystemsBiologyWorkbench::Broker::Broker::changeModuleNameLocal(int moduleId, Instance *moduleInst,
																	string newModuleName)
{
	string moduleName = moduleInst->getName();

	TRACE("Changing name for local module " << Properties::toString(moduleId) << " ('" << moduleName << "') to '" << newModuleName << "'");

	// Get the current module descriptor for this module.  Because we
	// started with a moduleId, we know that if getModuleDescriptor
	// returns null, the module is currently unregistered.

	ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);
	if (md != NULL)
	{
		TRACE("Found '" << moduleName << "' in the local registry");
		moduleRegistry->addModuleDescriptor(newModuleName,
			md->getDisplayName(),
			md->getManagementType(),
			md->getCommandLine(),
			md->getHelp());
	}
	else
	{
		TRACE("Found " << moduleId << " is currently unregistered;"
			<< " adding new entry with default fields");
		moduleRegistry->addModuleDescriptor(newModuleName, "",
			SELF_MANAGED, "", "");
	}
	rpc->notifyOthersOfRegistrationChange(BROKER_MODULE, true);
}
void SystemsBiologyWorkbench::Broker::Broker::changeModuleNameRemote(int moduleId, Instance *moduleInst,
																	 string newModuleName)
{
	// This must be a remote instance, so cast it to access getRemoteId().
	int nRemoteId = 0;
	ProxyModuleInstance *oProxy = dynamic_cast<ProxyModuleInstance *> (moduleInst);
	BrokerInstance *oBroker = dynamic_cast<BrokerInstance *> (moduleInst);
	if (oProxy != NULL) nRemoteId = oProxy->getRemoteId();
	else if (oBroker != NULL) nRemoteId = oBroker->getRemoteId();
	else return;

	BrokerInterface *broker = getBrokerInterface(moduleInst->getBrokerId());

	// Note that moduleId refers to a module on the local broker --
	// must translate it to what it is on the *remote* broker.

	broker->changeModuleName(nRemoteId, newModuleName);
}
void SystemsBiologyWorkbench::Broker::Broker::registerService(string moduleName, string serviceName,
															  string serviceDisplayName, string category,
															  string help)
{	
	// This has a lot of similarity to what ModuleImpl.addService
	// does.  We have to do the same tests here too b/c this method can
	// be called from the low-level API directly.

	if (StringUtil::empty(moduleName))
		throw new SBWModuleDefinitionException(
		"Null or empty module name",
		"'Null' and empty strings are invalid module names.");

	if (getHostPart(moduleName) != NULL)
		throw new SBWModuleDefinitionException(
		"Service registration cannot use host:modulename format",
		"The program attempted to register a service with a"\
		" module whose name has the form 'host:modulename'."\
		" This form is illegal during module registration"\
		" because modules are always registered with their"\
		" local broker only.");

	if (StringUtil::empty(serviceName))
		throw new SBWModuleDefinitionException(
		"Null or empty service name",
		"'Null' and empty strings are invalid service names.");

	if (StringUtil::empty(serviceDisplayName)) 
	{
		//throw new SBWModuleDefinitionException(
		//    "Null or empty service display name",
		//    "Service display names cannot be null or empty strings.");
		TRACE("Service display names cannot be null or empty strings.");
		serviceDisplayName = serviceName;
	}

	moduleName = normalizeModuleName(moduleName);
	serviceName = serviceName;
	category = SBW::normalizeCategoryName(category);

	ModuleDescriptor *md = moduleRegistry->getModuleDescriptor(moduleName);

	if (md == NULL)
	{
		throw new SBWModuleNotFoundException( string() +
			"No such module '" + moduleName + "'",
			"A module must be defined before services can be defined on it");
	}

	TRACE("Registering service on module '" << moduleName
		<< "' named '" << serviceName
		<< "', display name = '" << serviceDisplayName
		<< "', category = '" << category
		<< "', help = '" << help << "'");

	// People should be able to overwrite the definition of a service
	// on a module.  We print a notice just in case they didn't mean to.

	vector< ServiceDescriptor *> services = md->getServices();
	int nPosition = -1;
	for (unsigned int i = 0; i < services.size(); i ++) {
		if (services[i]->getName() == serviceName)
			nPosition = (int)i;
	}
	if (nPosition != -1) 
	{
		TRACE("Overwriting service '" << serviceName << "' on module '" << moduleName << "'");
	}

	DataBlockWriter oWriter;
	oWriter << md->getName() << serviceName << serviceDisplayName << category << help;
	ServiceDescriptor *service = new ServiceDescriptor(oWriter.getReader(), *md);
	md->addService(service);

	// Overwrite the previous registry entry for this module, forcing
	// the registry to be checkpointed.

	moduleRegistry->addModuleDescriptor(moduleName, md);
	rpc->notifyOthersOfRegistrationChange(BROKER_MODULE, true);
}
void SystemsBiologyWorkbench::Broker::Broker::unregisterModule(string moduleName)
{
	if (getHostPart(moduleName) != NULL)
		throw new SBWModuleDefinitionException(
		"Module registration cannot use host:modulename format",
		"The program attempted to unregister a module whose name"\
		" has the form 'host:modulename'.  This form is illegal"\
		" during module registration because modules are always"\
		" registered with their local broker only.");

	moduleName = normalizeModuleName(moduleName);
	TRACE("Removing registration for module '" << moduleName << "'");

	// This doesn't throw an exception if the module name is empty.
	// Should it?

	if (! StringUtil::empty(moduleName))
	{
		moduleRegistry->removeModuleDescriptor(moduleName);
		rpc->notifyOthersOfRegistrationChange(BROKER_MODULE, true);
	}
}
int SystemsBiologyWorkbench::Broker::Broker::linkBroker(string hostName)
{
	return getBrokerId(new Host(hostName));
}
void SystemsBiologyWorkbench::Broker::Broker::shutdownBroker()
{
	//! TODO fix that ...
	rpc->pleaseStop();
}
string SystemsBiologyWorkbench::Broker::Broker::getVersion()
{
#ifdef _DEBUG
	return "1.0.5 (C++ Debug)";
#else
	return "1.0.5 (C++ Release)";
#endif
}
int SystemsBiologyWorkbench::Broker::Broker::startNewLocalModuleInstance(ModuleDescriptor *md)
{
	string moduleName = md->getName();
	TRACE("Attempting to start an instance of '" << moduleName << "'")
		rpc->waitingForModule = moduleName;
	rpc->nModuleId = -1;
	SBWOS::startProcess((char*)(md->getCommandLine().c_str()));
	for (int i = 0; i < 600; i++) {
		SBWThread::sleep(100);
		if (rpc->nModuleId != -1) 
			return rpc->nModuleId;
	}
	TRACE("Module '" << moduleName << "' couldn't be started in time.");	
	throw new SBWModuleStartException( moduleName + " couldn't be started" , moduleName + " couldn't be started" );
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::listForModuleDescriptor(ModuleDescriptor *md)
{	
	DataBlockWriter oResult;
	oResult << md->getName() << md->getDisplayName() << md->getManagementType() << md->getCommandLine() << md->getHelp();
	return oResult;
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::listForModuleDescriptor(string moduleName, string displayName,
																				 int mgmtType, string cmdLine, string help)
{
	DataBlockWriter oResult;
	oResult << moduleName << displayName << mgmtType << cmdLine << help;
	return oResult;
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::listForServiceDescriptor(string moduleName, string serviceName,
																				  string serviceDisplayName, string category,
																				  string helpString)
{
	DataBlockWriter oResult;
	oResult << moduleName << serviceName << serviceDisplayName << category << helpString;
	return oResult;
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::listForServiceDescriptor(string moduleName, ServiceDescriptor *sd)
{
	DataBlockWriter oResult;
	oResult << moduleName << sd->getName() << sd->getDisplayName() << sd->getCategory() << sd->getHelp();
	return oResult;
}
bool SystemsBiologyWorkbench::Broker::Broker::matchCategory(string category, string toMatch, int start)
{
	if (category.length() <= (unsigned)start || toMatch.length() > category.length())
	{
		return false;
	}
	else if (category.substr(start).find(toMatch, start) == (unsigned)start)
	{
		return true;
	}
	else if (category.substr(start).find('/', start) >= 0)
	{
		int nFound = category.substr(start).find('/', start);
		if (nFound < 0)
			return false;
		int newPos = nFound + 1;
		return matchCategory(category.substr(start), toMatch,	newPos);
	}
	else
	{
		return false;
	}
}
string SystemsBiologyWorkbench::Broker::Broker::normalizeModuleName(string moduleName)
{
	if (StringUtil::empty(moduleName))
		throw new SBWModuleNotFoundException( "Null or empty module name", "'Null' and empty strings are invalid module names.");
	return moduleName;

}
string SystemsBiologyWorkbench::Broker::Broker::normalizeServiceName(string serviceName)
{
	return serviceName;
}
void SystemsBiologyWorkbench::Broker::Broker::checkModuleId(int moduleId)
{
	if (moduleId != BROKER_MODULE
		&& ! instanceTable->exists(moduleId))
	{
		TRACE("Given unknown module id " << moduleId)
			stringstream a,b;
		a << "Unknown module identifier " << moduleId;
		b << "An attempt was made to obtain an instance of a module"
			<< " with identifier " << moduleId  << ", but no such module"
			<< " is known to the SBW Broker.";
		throw new SBWModuleNotFoundException(a.str(), b.str());
	}
}
Host *SystemsBiologyWorkbench::Broker::Broker::getHostPart(string name)
{
	//       name = name.trim();
	int pos = name.find(':');
	if (pos != name.npos)
		return new Host(name.substr(0, pos));
	else
		return NULL;
}

string SystemsBiologyWorkbench::Broker::Broker::getModuleNamePart(string name)
{
	int pos = name.find(':');
	if (pos != name.npos)
		return normalizeModuleName(name.substr(pos + 1));
	else
		return normalizeModuleName(name);
}
Instance *SystemsBiologyWorkbench::Broker::Broker::getInstanceForModuleId(int moduleId)
{
	Instance *inst = instanceTable->getInstance(moduleId);
	if (inst != NULL)
		return inst;

	stringstream a,b;
	a <<"Unknown module " << moduleId ;
	b << "An attempt was made to obtain an instance of a module"
		<< " with identifier " << moduleId << " from the broker on '"
		<< rpc->getHost()->getName() 
		<< ", but no such module is known to the Broker.";
	throw new SBWModuleNotFoundException( a.str(),b.str());
}
ModuleDescriptor *SystemsBiologyWorkbench::Broker::Broker::getDescriptor(int brokerId, string moduleName)
{
	if (brokerId == BROKER_MODULE)
		return moduleRegistry->getModuleDescriptor(moduleName);
	else
	{
		BrokerInstance *brokerInst = (BrokerInstance *) instanceTable->getInstance(brokerId);
		BrokerInterface *broker = getBrokerInterface(brokerId);

		DataBlockWriter list = broker->getModuleDescriptor(moduleName, false);

		Host *host = brokerInst->getHost();
		return new ModuleDescriptor(mungeModuleDescriptor(list, moduleName, host).getReader());
	}
}

vector<ModuleDescriptor *> SystemsBiologyWorkbench::Broker::Broker::getDescriptors()
{
	vector<ModuleDescriptor *> oTemp;
	vector<int> brokerIds = getBrokerIds();
	for (unsigned int i = 0; i < brokerIds.size(); i++)
	{
		vector<ModuleDescriptor *> descriptors = getDescriptors(brokerIds[i]);
		for (unsigned int j = 0; j < descriptors.size(); j++)
			oTemp.push_back(descriptors[j]);
	}
	return oTemp;
}

vector<ModuleDescriptor *> SystemsBiologyWorkbench::Broker::Broker::getDescriptors(int brokerId)
{
	if (brokerId == BROKER_MODULE)
		return moduleRegistry->getModuleDescriptors();
	else
	{
		BrokerInstance *brokerInst = dynamic_cast<BrokerInstance *> (instanceTable->getInstance(brokerId));
		BrokerInterface *broker = getBrokerInterface(brokerId);
		Host *host = brokerInst->getHost();

		vector< DataBlockWriter> records = broker->getModuleDescriptors(true, false);

		vector<ModuleDescriptor *>  mds;
		for (unsigned int i = 0; i < records.size(); i++)
		{
			int nLength = 0;
			DataBlockWriter list = records[i];
			DataBlockWriter *oWriter = new DataBlockWriter;
			unsigned char * copy = list.getDataCopy(nLength);
			oWriter->insertData(copy,nLength); 
			string moduleName;
			oWriter->getReader() >> moduleName;
			delete oWriter;
			mds.push_back(new ModuleDescriptor(mungeModuleDescriptor(list, moduleName, host).getReader()));
		}
		return mds;
	}	
}
int SystemsBiologyWorkbench::Broker::Broker::getBrokerId(Host *host)
{
	if (rpc->getHost()->equals(host))
		return BROKER_MODULE;
	else
	{
		// See if we already have a connection to a Broker on the host.

		vector<Instance *> instances = instanceTable->getInstances();
		for (int i = instances.size() - 1; i >= 0; i--)
		{
			BrokerInstance *inst = dynamic_cast<BrokerInstance *>(instances[i]);
			if (inst != NULL && inst->getHost()->equals(host))
				return inst->getId();
		}

		// If we get here, we haven't connected to the host yet.

		return rpc->connectBroker(host);
	}
}

vector<int> SystemsBiologyWorkbench::Broker::Broker::getBrokerIds()
{
	vector<int> oTemp;
	vector<Instance *> instances = instanceTable->getInstances();
	oTemp.push_back( BROKER_MODULE );
	for (unsigned int i = 0; i < instances.size(); i++)
	{
		Instance *inst = instances[i];
		BrokerInstance *oBroker = dynamic_cast<SystemsBiologyWorkbench::Broker::BrokerInstance *> (inst);

		if (oBroker != NULL)
			oTemp.push_back(inst->getId());
	}
	return oTemp;
}
BrokerInterface *SystemsBiologyWorkbench::Broker::Broker::getBrokerInterface(Host *host)
{
	return getBrokerInterface(getBrokerId(host));
}

BrokerInterface *SystemsBiologyWorkbench::Broker::Broker::getBrokerInterface(int bId)
{    
	checkBrokerId(bId);
	BrokerInstance *bInst = dynamic_cast<BrokerInstance *>( instanceTable->getInstance(bId));
	if (bInst != NULL)
		return bInst->getBrokerInterface();
	else
		throw new SBWApplicationException("instance: " + Properties::toString(bId) + " could not be found");
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::mungeModuleDescriptor(DataBlockWriter list, string modName, Host *host)
{
	DataBlockWriter oResult;
	DataBlockReader oReader = list.getReader();
	string sTemp;
	int nLength = 0;

	oResult << host->getName() + ":" + modName;
	oReader >> sTemp;
	oReader >> sTemp;
	oResult << sTemp + " on " + host->getName();
	unsigned char * copy = oReader.getRemainingDataForC(nLength);
	oResult.insertData(copy,nLength);

	return oResult;
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getListOfModules()
{
	DataBlockWriter oResult;
	vector <Instance *> instances = instanceTable->getInstances();
	vector <Instance *> ::iterator oIterator;
	for (oIterator = instances.begin(); oIterator != instances.end(); oIterator ++)
	{
		Instance *currentInstance = (*oIterator);
		DataBlockWriter oTemp;
		string moduleName;
		if (currentInstance->getHost()->equals(rpc->getHost()))
		{
			moduleName = currentInstance->getName();
		}
		else
		{
			moduleName = currentInstance->getHost()->getName() + ":" + currentInstance->getName();
		}
		oTemp << currentInstance->getId() << moduleName;
		oResult << oTemp;
	}
	return oResult;
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getMethodIds(int moduleId,int serviceId)
{
	try
	{
		checkModuleId(moduleId);
		DataBlockWriter oResult;
		DataBlockWriter args ;
		args << serviceId;
		vector<string> methods;
		rpc->call(moduleId, SBWSystemService, SBWGetMethodsMethod, args) >> methods;
		for (unsigned int i = 0; i < methods.size(); i++)
		{
			DataBlockWriter oTemp;
			oTemp << (signed)i << methods[i];
			oResult << oTemp;
		}

		return oResult;
	} 
	catch (SBWException *e)
	{
		throw e;
	}
	catch (...)
	{
		throw new SBWApplicationException ("error retrieving method ids");
	}
}
DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::getMethodIds(string moduleName,string serviceName)
{
	try
	{
		std::vector<std::string> services ;
		Integer serviceId = 0 ;
		Instance *oTemp = instanceTable->getInstance((getHostPart(moduleName) == NULL ? rpc->getHost() : getHostPart(moduleName)),  getModuleNamePart(moduleName));
		if (oTemp == NULL)
			throw new SBWApplicationException("error retrieving method ids for " + moduleName + "and " + serviceName);
		DataBlockReader result =
			rpc->call(oTemp->getId(), SBWSystemService, SBWGetServicesMethod, DataBlockWriter());
		result >> services ;

		while (serviceId != (Integer) services.size()
			&& services[serviceId].compare(serviceName))
			serviceId++ ;

		if (serviceId == (Integer) services.size())
			throw new SBWServiceNotFoundException("service not found");

		return getMethodIds(oTemp->getId(),serviceId);
	}
	catch(...)
	{
		throw new SBWApplicationException("error retrieving method ids for " + moduleName + "and " + serviceName);
	}
}
bool SystemsBiologyWorkbench::Broker::Broker::isRemoteInstance(int nModuleId)
{	
	Instance * oTemp = instanceTable->getInstance(nModuleId);
	return !(oTemp->getHost()->equals(rpc->getHost()));	
}

DataBlockWriter SystemsBiologyWorkbench::Broker::Broker::mungeServiceDescriptor(DataBlockWriter list, string modName, Host *host)
{
	DataBlockWriter oResult;
	DataBlockReader oReader = list.getReader();
	string sTemp;
	int nLength = 0;
	oReader >> sTemp;
	oResult << host->getName() + ":" + modName;
	oReader >> sTemp;
	oResult << sTemp;
	oReader >> sTemp;
	oResult << sTemp + " on " + host->getName();
	unsigned char * copy = oReader.getRemainingDataForC(nLength);
	oResult.insertData(copy,nLength);
	return oResult;
}

SystemsBiologyWorkbench::Broker::SBWBrokerRPC *SystemsBiologyWorkbench::Broker::Broker::rpc = NULL;
SystemsBiologyWorkbench::Broker::InstanceTable *SystemsBiologyWorkbench::Broker::Broker::instanceTable = NULL;
SystemsBiologyWorkbench::Broker::ModuleRegistry *SystemsBiologyWorkbench::Broker::Broker::moduleRegistry = NULL;

