/*******************************************
File qps.h: Header file for Class QPS

This class is used to parse the the input
file in the QPS (or MPS) format and return
input data  in the appropriate form 
specified by the OOQP package
********************************************/

#include <iostream.h>
#include <string>
#include <fstream>
#include "qps.h"

// Constructors: take filename as argument
QPS::QPS(char * ifilename)
{
	input.open(ifilename);
	if (!input)
	{
		cerr << " ...can't open " << ifilename << " qps file " << endl;
		exit(1);
	}
	nrows=0;  ncols=0;
	glinenum=0;

	// initialize all arrays:
	for (int i=0;i<NROWS; i++)
	{
		LB[i] = 0.0;
		UB[i] = 0.0;
		BF[i] = 0;
	}

	cout << " calling START" << endl;

	// start reading the QPS file
	START();
}
	

QPS::~QPS()
{
;
}

// this function must be called first. Read in the NAME card
void QPS::START()
{
	string s;

	// get rid of comment lines
	while (1)
	{
		getline(input, s, '\n');
		glinenum++;
		cout << " first line read = " << s << endl;
		if ( s[0] != '*')
		{
			cout << "s[0] = " << s[0] << endl; 
			break;
		}
	}
	

	cout << "s.substr(0,4) = " << s.substr(0,4) << endl;

	if (s.substr(0,4) != "NAME")
	{
		cerr << " QPS file must have NAME card " << endl;
		exit(1);
	}
	if (s.size() > 15)
		problem_name = s.substr(14,s.size());
	else
		problem_name = "unclassified";

	cout << " problem name = " << problem_name << endl;
	cout << " line size    = " << s.size() << endl;

	// now call the ROWS() function to read in the ROWS section 
	ROWS();

}

// read in the ROWS section of the input file. Make sure that this
// is called imediately after START()
void QPS::ROWS()
{
	string s;
	int temp =0;

	// get rid of comment lines
	while(1)
	{
		getline(input, s, '\n');	
		glinenum++;
		if ( s[0] != '*') break;
	}

	if (s.substr(0,4) != "ROWS")
	{
		cerr << " QPS must have ROWS section " << endl;
		exit(1);
	}

	// now get rid of comment lines again
	nrows = 0;
	while (1)
	{
		getline(input,s,'\n');
		glinenum++;
		if (s[0] != '*')
		{
			if (s.substr(0,7) == "COLUMNS")
			{
				cout << s.substr(0,7) << " will be read next! ROWS section ended \n";
				break;
			}
			temp = s.size();
			if (temp < 5) 
			{
				cerr << " Wrong format in the ROWS section at line # "<< glinenum << endl;
				exit(1);
			}
			Ctype[nrows] = (char) s[1];	
			if ((Ctype[nrows] != 'E') &&  (Ctype[nrows] != 'N'))
			{
				cerr << " Current version only deals with Ax = b" << endl;
				exit(1);
			}
			row_name[nrows] = s.substr(4,temp);
			nrows++;
		}
	}		
	#ifdef DEBUG
	cout << "ROWS" << endl;
	for (int i=0; i< nrows; i++)
	{
		cout << " " << Ctype[i] << "  " << row_name[i] << endl;
	}
	cout << " Number of Rows = " << nrows << endl;
	#endif
	COLUMNS();
}

// read in the COLUMN section of the input file. Make sure that this
// is called imediately after ROWS()
void QPS::COLUMNS()
{
	string s, st, ctoken, sr, sv;
	int temp;
	double v;
	int calling_flag=0, flag = 0;

	// get rid of comment lines if there're any
	ncols=0;
	while(1)
	{
		getline(input, s, '\n');	
		glinenum++;
		if ( s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! COLUMNS section ended " << endl;
				cout << " calling flag = " << calling_flag << endl;
				break;
			}

			temp = s.size();
			if ((temp < 25) || ( (temp >= 40) && (temp < 50)))
			{	
				cerr << " File format in COLUMNS Section is wrong at line # " << glinenum << endl;
				exit(1);
			}

			 	
			// get the first token:  column variable
			st = clean_word(s.substr(4,8));	
			// get the second token: row variable
			sr = clean_word(s.substr(14,8));
			// get the third token: value
			sv = clean_word(s.substr(24,12));
			v  = atof(sv.c_str());	

			// insert into Cnode array
			
			// takene care of the first node where ncols index = 0
			if ( (ncols==0) && (flag ==0))
			{
				col_name[ncols].id = st;
				flag = 1;
			}

			if (col_name[ncols].id != st)
			{
				ncols++;
				col_name[ncols].id = st;
				col_name[ncols].insert(sr,v);
				cout << "index = " << ncols << "   " << col_name[ncols].id << endl;
			}
			else
			{
				col_name[ncols].insert(sr,v);
			}

			// this is optional section for fields 5 and 6
			if (temp >= 40)
			{
				sr = clean_word(s.substr(39,8));
				sv = clean_word(s.substr(49,12));
				v  = atof(sv.c_str());
				col_name[ncols].insert(sr,v);
			}
		}

	}
	// ncols now holds number of primal variables (# of culumns)
	ncols++;
	print_col_vs_row();
	cout << " number of primal variables = " << ncols << endl;
	function_selection(calling_flag);

}

void QPS::RHS()
{
	string s;
	int calling_flag=0;

	// get rid of comment lines if there're any
	while(1)
	{
		getline(input, s, '\n');	
		glinenum++;
		if ( s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! RHS section ended " << endl;
				break;
			}
		}
	}
	function_selection(calling_flag);
}

void QPS::RANGES()
{
	string s;
	int calling_flag=0;

	while(1)
	{
		getline(input, s, '\n');
		glinenum++;
		if (s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! RANGES section ended " << endl;
				break;
			}
		}
	}
	function_selection(calling_flag);
}


void QPS::QSECTION()
{
	cout << " this QSECTION should not be called at the moment.  Quit now! \n";
	exit(1);

}

void QPS::CSECTION()
{
	cout << " this CSECTION should not be called at the moment.  Quit now! \n";
	exit(1);
}

void QPS::BOUNDS()
{
	cout << " in BOUNDS() \n";
	
	string s, bt, st, sv;
	double v=0.0;
	int calling_flag=0, index=0, flag = 0;

	while (1)
	{
		getline(input,s, '\n');
		glinenum++;
		if (s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! RHS section ended " << endl;
				break;
			}
		
			bt = s.substr(1,2);
			st = clean_word(s.substr(14,8));
			if (bt != "FR")
			{
				sv = clean_word(s.substr(24,12));
				v  = atof(sv.c_str());
			}
			else
			{
				v = 0;
			}
			cout << st << "\t" << v << endl;

			// need to deal with the same  variables on two sucessive lines for UP, LO keywords
			if ((index == 0) && ( flag ==0))
			{
				bound_name[index] = st;
				flag = 1;
				if (bt != "FR")
					BF[index] = 1;
			}
					
			if ( bound_name[index] != st)
			{
				index++;
				// cout << index << "\t:" << st << endl; 
				if (index >= NROWS)
				{
					cerr << " Number of bound types for variables exceeds NROWS and = " << index << endl;
				}
				bound_name[index] = st;
				if (bt != "FR")
				{
					BF[index ] = 1;
					if ( bt == "LO")
						LB[index] = v;
					else if (bt == "UP")
						UB[index] = v;
					else if (bt == "FX")
					{
						UB[index] = v;
						LB[index] = v;
					}
				}
			}
			else
			{
				if (bt == "LO")
					LB[index] = v;
				else if (bt == "UP")
					UB[index] = v;
	
			}
		}
	}
	for (int j=0;j<=index;j++)
	{
		cout << "Bount=" << BF[j] << ": " << LB[j] << " <= \t " << bound_name[j] << " \t<= " << UB[j] << endl;
	}	
	cout << "FInal index = " << index << endl;
 	function_selection(calling_flag);	
	cout << " number of lines read = " << glinenum << endl;
}


void QPS::ENDATA()
{
	cout << " number of lines read = " << glinenum << endl;
	PROCESS();
}

void QPS::PROCESS()
{
	cout << " ...now, processing input data ==> appropriate arrays ... " << endl;	
}

void QPS::print_col_vs_row()
{
	for (int i=0; i< ncols; i++)
	{
		cout << col_name[i].id << "   ";
		for (int j=0; j< col_name[i].nr; j++)
		{
			cout << col_name[i].row_name[j] << "   "  << col_name[i].values[j] << "; ";
		}
		cout << endl;
	}
}	

inline int QPS::set_calling_flag(const string & s)
{
	int cf;          
	if (s.substr(0,3) == "RHS")       
		cf = 1;
	else if (s.substr(0,6) == "RANGES")
	       	cf = 2;
	else if (s.substr(0,8) == "QSECTION")
		cf = 3;
	else if (s.substr(0,6) == "BOUNDS")
		cf = 4;
	else if (s.substr(0,8) == "CSECTION")
		cf = 5;
	else if (s.substr(0,6) == "ENDATA")
		cf = 6;
	else
		cf = 0;

	return cf;
}
void QPS::function_selection(int i)
{
	switch(i)
	{
		case 1: 
		{
			RHS();
			break;
		}
		case 2:
		{
			RANGES();
			break;
		}
		case 3:
		{
			QSECTION();
			break;
		}
		case 4:
		{
			BOUNDS();
			break;
		}
		case 5:
		{
			CSECTION();
			break;
		}
		case 6:
		{
			ENDATA();
			break;
		}
	}

}


int QPS::search_row(string s)
{
	for (int i=0;i<nrows; i++)
	{
		if (row_name[i] == s)
			return 1;
	}
	return 0;
}

int QPS::search_col(string s)
{
	/*
	for (int i=0;i<nrows; i++)
	{
		if (col_name[i] == (char) s[0])
			return 1;
	}
	*/
	return 0;
}

// this function removes all white space characters that a token has!
string QPS::clean_word(string ss)
{
	string s;
	for (int i=0;i<ss.size();i++)
	{
		if (ss[i] != ' ')  
			s += ss[i];
	}
	return s;
}

Cnode::Cnode()
{
          nr = 0;
          for (int i=0;i<NNB;i++)
          {
                    values[i] = 0.0;
                    row_name[i] = "";
          }
} 


Cnode::~Cnode()
{ ; }

void Cnode::insert(string s, double v)
{
	if (nr>=NNB)
	{
		cerr << " Number of edges for each node is > " << NNB << ". Exit now! " << endl;
		cerr << "at ID = " << id << endl;
		exit(1);
	}
	row_name[nr] = s;
	values[nr] = v;
	nr++;
}


