/*******************************************
File qps.h: Header file for Class QPS

This class is used to parse the the input
file in the QPS (or MPS) format and return
input data  in the appropriate form 
specified by the OOQP package
********************************************/

#include <iostream.h>
#include <string>
#include <fstream>
#include "qps.h"

// Constructors: take filename as argument
QPS::QPS(char * ifilename)
{
	input.open(ifilename);
	if (!input)
	{
		cerr << " ...can't open " << ifilename << " qps file " << endl;
		exit(1);
	}

	// initialize all internal data variables
	nrows=0;  ncols=0;
	glinenum=0;
	N_BOUND = 0;

	// initialize all arrays:
	for (int i=0;i<NROWS; i++)
	{
		LB[i] = 0.0;
		UB[i] = 0.0;
		BFL[i] = 1;
		BFU[i] = 0;
	}


	//initialize all variables used in OOQP
	
	NX = 0; C = NULL;

	IXUPP = NULL;  IXLOW = NULL; XUPP = NULL; XLOW = NULL;

	NNZA = 0; IROWA = NULL; JCOLA = NULL; DA = NULL;

	MY=0; B=NULL;

	MZ = 0; CLOW = NULL; CUPP = NULL; ICLOW = NULL; ICUPP = NULL;

	NNZC = 0; IROWC = NULL; JCOLC=NULL; DC = NULL;

	NNZQ = 0; IROWQ=NULL; JCOLQ=NULL; DQ = NULL;


	// start reading the QPS file	
	cout << " calling START" << endl;
	START();
}
	
QPS::~QPS()
{
;
}

// this function must be called first. Read in the NAME card
void QPS::START()
{
	string s;

	// get rid of comment lines
	while (1)
	{
		getline(input, s, '\n');
		glinenum++;
		// cout << " first line read = " << s << endl;
		if ( s[0] != '*')
		{
			break;
		}
	}
	
	if (s.substr(0,4) != "NAME")
	{
		cerr << " QPS file must have NAME card " << endl;
		exit(1);
	}
	if (s.size() > 15)
		problem_name = s.substr(14,s.size());
	else
		problem_name = "unclassified";

	// cout << " problem name = " << problem_name << endl;

	// now call the ROWS() function to read in the ROWS section 
	ROWS();

}

// read in the ROWS section of the input file. Make sure that this
// is called imediately after START()
void QPS::ROWS()
{
	string s;
	int temp =0;

	// get rid of comment lines
	while(1)
	{
		getline(input, s, '\n');	
		glinenum++;
		if ( s[0] != '*') break;
	}

	if (s.substr(0,4) != "ROWS")
	{
		cerr << " QPS must have ROWS section " << endl;
		exit(1);
	}

	// now get rid of comment lines again
	nrows = 0;
	while (1)
	{
		getline(input,s,'\n');
		glinenum++;
		if (s[0] != '*')
		{
			if (s.substr(0,7) == "COLUMNS")
			{
				cout << s.substr(0,7) << " will be read next! ROWS section ended \n";
				break;
			}
			temp = s.size();
			if (temp < 5) 
			{
				cerr << " Wrong format in the ROWS section at line # "<< glinenum << endl;
				exit(1);
			}
			// if restraint type == E, add to list 
			if (s[1] == 'E')
			{
				Ctype[nrows] = (char) s[1];
				row_name[nrows] = s.substr(4,temp);
				nrows++;
			}
			else if (s[1] =='N')
			{
				Obj_var = s.substr(4,temp);
			}
			else
			{
				cerr << " Current version only deals with Ax = b" << endl;
				exit(1);
			}
		}
	}		
	// now the number of metabolites = nrows 

	#ifdef DEBUG
	cout << "ROWS" << endl;
	for (int i=0; i< nrows; i++)
	{
		cout << " " << Ctype[i] << "  " << row_name[i] << endl;
	}
	cout << " Number of Rows = " << nrows << endl;
	#endif

	// now call COLUMN() function
	COLUMNS();
}

// read in the COLUMN section of the input file. Make sure that this
// is called imediately after ROWS()
void QPS::COLUMNS()
{
	string s, st, ctoken, sr,sr2, sv, sv2;
	int temp, nnzA;
	double v,v2;
	int calling_flag=0, flag = 0;

	// get rid of comment lines if there're any
	ncols=0;
	nnzA=0;
	while(1)
	{
		getline(input, s, '\n');	
		glinenum++;
		if ( s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! COLUMNS section ended " << endl;
				// cout << " calling flag = " << calling_flag << endl;
				break;
			}

			temp = s.size();
			if ((temp < 25) || ( (temp >= 40) && (temp < 50)))
			{	
				cerr << " File format in COLUMNS Section is wrong at line # " << glinenum << endl;
				exit(1);
			}
			 	
			// get the first token:  column variable
			st = clean_word(s.substr(4,8));	
			// get the second token: row variable
			sr = clean_word(s.substr(14,8));
			// get the third token: value
			sv = clean_word(s.substr(24,12));
			v  = atof(sv.c_str());	

			// insert into Cnode array
			// takene care of the first node where ncols index = 0
			if ( (ncols==0) && (flag ==0))
			{
				col_name[ncols].id = st;
				flag = 1;
			}

			if (col_name[ncols].id != st)
			{
				//make room for new flux and assign name
				ncols++;
				col_name[ncols].id = st;

				// if row is an objective function element ==> objective function
				if (sr == Obj_var)
				{
					Obj_fun[ncols] = v;
				}
				else
				{
					col_name[ncols].insert(sr,v);
					//count number of non-zero element of matrix A
					nnzA++;
				}
			}
			else
			{
				// in this case, objective element is in the middle of an array
				if (sr == Obj_var)
				{
					Obj_fun[ncols] = v;
				}
				else
				{
					col_name[ncols].insert(sr,v);
					nnzA++;
				}
			}

			// this is optional section for fields 5 and 6
			if (temp >= 40)
			{
				sr2 = clean_word(s.substr(39,8));
				sv2 = clean_word(s.substr(49,12));
				v2  = atof(sv2.c_str());
				if (sr2 == Obj_var)
				{
					Obj_fun[ncols] = v2;
				}
				else
				{
					col_name[ncols].insert(sr2,v2);
					nnzA++;
				}
			}
		}

	}
	// ncols now holds number of primal variables (# of culumns)
	ncols++;
	NNZA = nnzA;			// number of non-zero elements of matrix A
	// print_col_vs_row();
	// cout << " number of primal variables = " << ncols << endl;
	function_selection(calling_flag);

}

void QPS::RHS()
{
	string s;
	int calling_flag=0;

	// get rid of comment lines if there're any
	while(1)
	{
		getline(input, s, '\n');	
		glinenum++;
		if ( s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! RHS section ended " << endl;
				break;
			}
		}
	}
	function_selection(calling_flag);
}

void QPS::RANGES()
{
	string s;
	int calling_flag=0;

	while(1)
	{
		getline(input, s, '\n');
		glinenum++;
		if (s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! RANGES section ended " << endl;
				break;
			}
		}
	}
	function_selection(calling_flag);
}


void QPS::QSECTION()
{
	cout << " this QSECTION should not be called at the moment.  Quit now! \n";
	exit(1);

}

void QPS::CSECTION()
{
	cout << " this CSECTION should not be called at the moment.  Quit now! \n";
	exit(1);
}

void QPS::BOUNDS()
{
	string s, bt, st, sv;
	double v=0.0;
	int calling_flag=0, index=0, flag = 0;

	while (1)
	{
		getline(input,s, '\n');
		glinenum++;
		if (s[0] != '*')
		{
			calling_flag = set_calling_flag(s);

			if (calling_flag > 0)
			{
				cout << " Section " << s << " will be read next! RHS section ended " << endl;
				break;
			}
		
			bt = s.substr(1,2);
			st = clean_word(s.substr(14,8));
			if (bt != "FR")
			{
				sv = clean_word(s.substr(24,12));
				v  = atof(sv.c_str());
			}
			else
			{
				v = 0;
			}

			// need to deal with the same  variables on two sucessive lines for UP, LO keywords
			if ((index == 0) && ( flag ==0))
			{
				bound_name[index] = st;
				flag = 1;
				if ((bt != "FR") && (bt =="LO"))
					BFL[index] = 1;
				else if ((bt != "FR") && (bt =="UP"))
					BFU[index] = 1;
				else if (bt == "FR")
				{
					BFU[index] = 0;
					BFL[index] = 0;
				}
			}
					
			if ( bound_name[index] != st)
			{
				index++;
				// cout << index << "\t:" << st << endl; 
				if (index >= NROWS)
				{
					cerr << " Number of bound types for variables exceeds NROWS and = " << index << endl;
				}
				bound_name[index] = st;
				if (bt != "FR")
				{
					if ( bt == "LO")
					{
						LB[index] = v;
						BFL[index ] = 1;
					}
					else if (bt == "UP")
					{
						UB[index] = v;
						BFU[index ] = 1;
					}
					else if (bt == "FX")
					{
						UB[index] = v;
						LB[index] = v;
						BFL[index ] = 1;
						BFU[index ] = 1;
					}
				}
				else if (bt =="FR")
				{
					BFL[index] = 0;
					BFU[index] = 0;
				}
			}
			else
			{
				if (bt == "LO")
				{
					LB[index] = v;
					BFL[index ] = 1;
				}
				else if (bt == "UP")
				{
					UB[index] = v;
					BFU[index] = 1;
				}
				else if (bt=="FX")
				{
					BFL[index ] = 1;
					BFU[index ] = 1;
				}
				else if (bt =="FR")
				{
					BFL[index] = 0;
					BFU[index] = 0;
				}
	
			}
		}
	}
#ifdef DEBUG
	for (int j=0;j<=index;j++)
	{
		cout << "(Bount=" << BFL[j] << ") " << LB[j] << " <= \t " << bound_name[j] << " \t<= " << UB[j];
	       cout << "(Bound=" << BFU[j]	<< ")" << endl;
	}	
	cout << "FInal index = " << index << endl;
#endif
	N_BOUND = index + 1;

 	function_selection(calling_flag);	
}

void QPS::ENDATA()
{
	cout << " number of lines read = " << glinenum << endl;
	PROCESS();
}

void QPS::PROCESS()
{
	string token, primal;
	int itemp=0, jtemp;
	double dtemp=0.0;
	
	cout << " ...now, processing input data ==> appropriate arrays ... " << endl;	

	// size of primal vector = number of column fluxes
	// and linear term of the objective function
	NX = ncols;
	MY = nrows;
	MZ = 0;  		// true for now: only deal with equality constraint
				// NNZA has been determined by the COLUMN() function
	NNZQ = 0;		// true for Linear programming part
	NNZC = 0;		// No inequality constraint optimization is done here

#ifdef DEBUG
	for (int i=0;i<NX;i++)
		if (Obj_fun[i] != 0)
		{
			cout <<  "index = " << i << ", value = " << Obj_fun[i] << endl;
		}
#endif

	// cout << " Name of non-zero element with COST = " << col_name[742].id << endl;

	cout << endl << " ==============================" << endl;
	cout << " Problem Name .............................= " << problem_name << endl;
	cout << " Objective variable ...................... = " << Obj_var << endl;
	cout << " Number of primal variables ...............= " << NX << endl;
	cout << " Number of metabolites(rows) of matrix A...= " << MY << endl;
	cout << " Number of non-zero elements of matrix A...= " << NNZA << endl;
	cout << " Number of metabolites(rows) of matrix C...= " << MZ << endl;
	cout << " Number of non-zero elements of matrix C...= " << NNZC << endl;
	cout << " Number of non-zero elements of matrix Q...= " << NNZQ << endl << endl;
	
	// print_data();
	//
	// Now, copy to appropriate arrays
	
	int index=0;
	int bindex=0;
	int bcount=0;

	C     = new double[NX];
	XUPP  = new double[NX];
	XLOW  = new double[NX];
	IXUPP = new char[NX];
	IXLOW = new char[NX];
	
	IROWA = new int[NNZA];
	JCOLA = new int[NNZA];
	DA    = new double[NNZA];

	for (int i=0;i<NX;i++)
	{
		// cout << col_name[i].id << "(" << i << ") <---> "; 
		for (int j=0;j<col_name[i].nr;j++)
		{
			// token is row (metabolite) name
			token = col_name[i].row_name[j];
			jtemp = i;
			itemp = search_row(token);
			if (itemp <0)
			{
				cerr << "Token " << token <<  " not found" << endl;
				exit(1);
			}
			else
			{
				IROWA[index] = itemp;
				JCOLA[index] = jtemp;
				DA[index]    = col_name[i].values[j];
				index++;
			}
			// cout << token << " (" << itemp << ") " << col_name[i].values[j] << " |   " ;
		}
		// now deal with the bound on primal variables
		primal = col_name[i].id;	
		bindex = search_bound(primal);

		// if there is an explicit bound primal
		if (bindex > -1)
		{
			// check for lower bound
			if (BFL[bindex] == 0)			//FR
			{
				IXLOW[i] = (char) 0;
				XLOW[i]  = 0.0;
			}
			else if (BFU[bindex] == 1)		//FX,LO
			{
				IXLOW[i] = (char) 1;
				XLOW[i] = LB[bindex];
			}
			// and check for upper bound
			if (BFU[bindex] == 0)			//FR
			{
				IXUPP[i] = ( char) 0;
				XUPP[i]  = 0.0;
			}
			else if (BFU[bindex] == 1)		//FX,UP
			{
				IXUPP[i] = (char) 1;
				XUPP[i]  = UB[bindex];
			}
			bcount++;
		}
		// no explicit bound primal, it lower bound is 0 and upper bound is infinity
		else
		{
			IXLOW[i] = (char) 1;
			XLOW[i]  = 0.0;
			IXUPP[i] = (char) 0;
			XUPP[i]  = 0.0;
		}

		// now deal with objective fucntion
		C[i] = Obj_fun[i];	
	}
	if (index != NNZA)
	{
		cerr << " Error in copying number of non-zero elemtn of A" << endl;
		cerr << " Total number copied = " << index << endl;
		cerr << " Total Number read in = " << NNZA << endl;
		exit(1);
	}
	if (bcount != N_BOUND)
	{
		cerr << " Number of bound variable copied is wrong: N_BOUND = " << N_BOUND << ", Actual Bound = " << bcount << endl;
		exit(1);
	}
#ifdef DEBUG
	for (int i=0;i<NNZA;i++)
		cout << "[" << IROWA[i] << "," << JCOLA[i] <<"]: " << DA[i] << endl; 
				
	for (int i=0;i< NX;i++)
	{
		cout << i << ": (" << IXLOW[i] << ") " << XLOW[i] << "<= " << col_name[i].id ;
		cout << "\t\t\t<=" << XUPP[i] << " \t\t(" << IXUPP[i] << endl; 
	}
#endif

}

void QPS::Export_I( int & nx, int &my, int &mz, int &nnzQ, int &nnzA, int &nnzC)
{
	nx = NX;
	my = MY;
	mz = MZ;
	nnzQ = NNZQ;
	nnzA = NNZA;
	nnzC = NNZC;
}

void QPS::Export_D(double *cc, double *xlow, char *ixlow, double *xupp, char *ixupp,
			      int *irowA, int *jcolA, double *dA, double *b, string *Cname, string *Rname)
{
	for (int i=0;i<NX;i++)
	{
		cc[i]   = C[i];
		if (cc[i] != 0)
			cout << "Objective function C[" << i<< "]=" << cc[i] << endl;
		xlow[i] = XLOW[i];
		xupp[i] = XUPP[i];
		ixlow[i] = IXLOW[i];
		ixupp[i] = IXUPP[i];
		Cname[i] = col_name[i].id;
	}
	for (int i=0;i<NNZA;i++)
	{
		irowA[i] = IROWA[i];
		jcolA[i] = JCOLA[i];
		dA[i]    = DA[i];
	}
	for (int i=0;i<MY;i++)
	{
		b[i] = 0.0;
		Rname[i] = row_name[i];
	}
}

void QPS::print_col_vs_row()
{
	for (int i=0; i< ncols; i++)
	{
		cout << col_name[i].id << "   ";
		for (int j=0; j< col_name[i].nr; j++)
		{
			cout << col_name[i].row_name[j] << "   "  << col_name[i].values[j] << "; ";
		}
		cout << endl;
	}
}	

inline int QPS::set_calling_flag(const string & s)
{
	int cf;          
	if (s.substr(0,3) == "RHS")       
		cf = 1;
	else if (s.substr(0,6) == "RANGES")
	       	cf = 2;
	else if (s.substr(0,8) == "QSECTION")
		cf = 3;
	else if (s.substr(0,6) == "BOUNDS")
		cf = 4;
	else if (s.substr(0,8) == "CSECTION")
		cf = 5;
	else if (s.substr(0,6) == "ENDATA")
		cf = 6;
	else
		cf = 0;

	return cf;
}
void QPS::function_selection(int i)
{
	switch(i)
	{
		case 1: 
		{
			RHS();
			break;
		}
		case 2:
		{
			RANGES();
			break;
		}
		case 3:
		{
			QSECTION();
			break;
		}
		case 4:
		{
			BOUNDS();
			break;
		}
		case 5:
		{
			CSECTION();
			break;
		}
		case 6:
		{
			ENDATA();
			break;
		}
	}

}


int QPS::search_row(const string s)
{
	for (int i=0;i<MY; i++)
	{
		if (row_name[i] == s)
			return i;
	}
	return -1;
}

int QPS::search_bound(const string s)
{
	for (int j=0;j<N_BOUND; j++)
	{
		if (bound_name[j] == s)
			return j;
	}
	return -1;
}

int QPS::search_col(string s)
{
	/*
	for (int i=0;i<nrows; i++)
	{
		if (col_name[i] == (char) s[0])
			return 1;
	}
	*/
	return 0;
}

// this function removes all white space characters that a token has!
string QPS::clean_word(string ss)
{
	string s;
	for (int i=0;i<ss.size();i++)
	{
		if (ss[i] != ' ')  
			s += ss[i];
	}
	return s;
}

void QPS::print_data()
{
        cout << " List of metabolites .............. " << endl;
        for (int i=0;i<MY;i++)
        {
                cout << " " << Ctype[i] << "  " << row_name[i] << endl;
        }

	cout << endl << " Non-zero element of matrix A ......... " << endl;
	int count =  0;
	for (int i=0;i<NX;i++)
	{
		cout << col_name[i].id << "<======> \t" ;
		for (int j=0;j<col_name[i].nr;j++)
		{
			cout << col_name[i].row_name[j] << "   "  << col_name[i].values[j] << "| ";
			count++;
                }
	      	cout << endl;
	}
	if (count != NNZA)
	{
		cerr << " error in the number of non-zero elements of A " << endl;
		exit(1);
	}

	// cout << endl << " Bound for each primal variable ....... " << endl;
	// for (int i=0;i<NX;i++)
	// {
	// 	cout << "(bount flag =" << BFL[i] << ") " << LB[i] << " <= " << col_name[i].id ;  
	// 	cout << " \t <= " << UB[i] << "(bount flag =" << BFU[i] << ")" << endl;
	// }

}


	

Cnode::Cnode()
{
          nr = 0;
          for (int i=0;i<NNB;i++)
          {
                    values[i] = 0.0;
                    row_name[i] = "";
          }
} 


Cnode::~Cnode()
{ ; }

void Cnode::insert(string s, double v)
{
	if (nr>=NNB)
	{
		cerr << " Number of edges for each node is > " << NNB << ". Exit now! " << endl;
		cerr << "at ID = " << id << endl;
		exit(1);
	}
	row_name[nr] = s;
	values[nr] = v;
	nr++;
}


