/* OOQP                                                               *
 * Authors: E. Michael Gertz, Stephen J. Wright                       *
 * (C) 2001 University of Chicago. See Copyright Notification in OOQP */

#include "PetscIterativeSolver.h"
#include "PetscSpSymMatrix.h"
#include "PetscSparseStorage.h"
#include "petscsles.h"
#include "PetscVector.h"
#include "PetscVectorHandle.h"
 
PetscIterativeSolver::PetscIterativeSolver( PetscSpSymMatrix * Mat,
					    SLES sles_ )
{
  int dummy, n, ierr;

  
  // Create the Petsc vectors used in the solve.
  mStorage = PetscSparseStorageHandle( Mat->getStorage() );
  mStorage->getSize( dummy, n );

  int first, lastp1, m;
  
  ierr = MatGetOwnershipRange( mStorage->M, &first, &lastp1 );
  m = lastp1 - first;
  hx = PetscVectorHandle( new PetscVector( m, n ) );
  x=hx->pv;
 
  sles = sles_;
  deleteSLES = 0;
  cits = 0; // No iterations yet!
}

PetscIterativeSolver::PetscIterativeSolver( PetscSpSymMatrix * Mat,
					    KSPType ksptype,
					    PCType pctype)
{ 
  KSP  ksp;
  PC   pc;
  int ierr;
  int dummy, n;

  // Create the Petsc vectors used in the solve.
  mStorage = PetscSparseStorageHandle( Mat->getStorage() );
  mStorage->getSize( dummy, n );

  int first, lastp1, m;

  ierr = MatGetOwnershipRange( mStorage->M, &first, &lastp1 );
  m = lastp1 - first;
  hx = PetscVectorHandle( new PetscVector( m, n ) );
  x=hx->pv;
 
  /* 
     Create linear solver context
  */
  ierr = SLESCreate(PETSC_COMM_WORLD,&sles);assert(ierr == 0);
  deleteSLES = 1;
  /* 
     Set operators. Here the matrix that defines the linear system
     also serves as the preconditioning matrix.
  */
  ierr = SLESSetOperators(sles, mStorage->M, mStorage->M,
			  DIFFERENT_NONZERO_PATTERN); assert(ierr == 0);

  /* 
     Set linear solver defaults for this problem (optional).
     - By extracting the KSP and PC contexts from the SLES context,
     we can then directly call any KSP and PC routines to set
     various options.
     - The following four statements are optional; all of these
     parameters could alternatively be specified at runtime via
     SLESSetFromOptions();
  */
  ierr = SLESGetKSP(sles,&ksp);assert(ierr == 0);
  ierr = KSPSetType(ksp, ksptype);assert(ierr == 0);

  ierr = SLESGetPC(sles,&pc);assert(ierr == 0);
  ierr = PCSetType(pc, pctype);assert(ierr == 0);

  ierr = KSPSetTolerances(ksp,1.e-8,
			  PETSC_DEFAULT, PETSC_DEFAULT,
			  PETSC_DEFAULT);assert(ierr == 0);

  /* 
     Set runtime options, e.g.,
     -ksp_type <type> -pc_type <type> -ksp_monitor -ksp_rtol <rtol>
     These options will override those specified above as long as
     SLESSetFromOptions() is called _after_ any other customization
     routines.
  */
  ierr = SLESSetFromOptions(sles);assert(ierr == 0);
  cits = 0; // No iterations yet!
}

void PetscIterativeSolver::matrixChanged()
{
  int ierr;

  ierr = SLESSetOperators(sles, mStorage->M, mStorage->M,
			  SAME_NONZERO_PATTERN); assert(ierr == 0);
}

void PetscIterativeSolver::solve( OoqpVector& v_in )
{
  assert( v_in.isKindOf( kPetscVector ) );
  PetscVector & v = dynamic_cast<PetscVector &>(v_in);

  int ierr, its;
 
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
     Solve the linear system
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
     Solve linear system
  */
  ierr = SLESSolve(sles, v.pv, x, &its); assert( ierr  == 0); 
  cits += its;
  //    cout << "Call: " << ++count << " its: "
  //         << its << " cumulative: " << cits << endl;

  ierr = VecCopy( x, v.pv ); assert( ierr  == 0);
}

void PetscIterativeSolver::diagonalChanged( int /* idiag */, int /* extent */ )
{
  this->matrixChanged();
}

PetscIterativeSolver::~PetscIterativeSolver()
{
  int ierr;

  if( deleteSLES ) { // We made it, we own it.
    ierr = SLESDestroy( sles ); assert( ierr  == 0);
  }
}
