#include "sbml/common.h"
#include <stdlib.h>
#include <string.h>
#include "annotationData.h"
#include "sbml/SBMLReader.h"
#include "sbml/SBMLTypes.h"

#include "util.h"

#define MAX_NUM_REACTIONS 2000
#define MAXLIMIT  1000000
#define N_MATRIX_ELEMENT 10000   // this is used to create vector IROWA, JCOLA, and DA
#define N_METABOLITES   1000
#define N_REACTIONS     2000

int  pToken(const char * Token, const char * s, double *bound);
void MarkMetabolites(const char ** list, int * metList, int n, const char * met, int * mcount);

const char * LOWER = "LOWER";
const char * UPPER = "UPPER";
const char * OBJ = "OBJECTIVE";
const char * PRED = "PREDICTION";
const char * annotationTemplate = "<annotation xmlns:flux=\"http://arep.med.harvard.edu/fluxns\">\n\
	  <flux:limit analysis=\"%s\" upper=\"%s\" lower=\"%s\" objective=\"%f\" prediction=\"%f\"/>\n\
	  <flux:limit analysis=\"%s\" upper=\"%s\" lower=\"%s\" objective=\"%f\" prediction=\"%f\"/>\n\
	  <flux:limit analysis=\"%s\" upper=\"%s\" lower=\"%s\" objective=\"%f\" prediction=\"%f\"/>\n\
        </annotation>";

struct ReactionData 
{
	const char * ReactionID;
	int          isReversible;        // (1=yes, 0=no)
	const char ** ReactantID; const char ** ProductID; const char ** ModifierID;
	int  nReactants, nProducts, nModifiers;
	double  *ReactantStoiCoeff, *ProductStoiCoeff;
} RD[MAX_NUM_REACTIONS];



int sbmlParser (int fromString_p, char *SBMLFileName, int *NX, int *MY, int *NNZA, int *IROWA, int *JCOLA, double *DA, annotationData *AD, char **MNAME, char ** RNAME)
{
  double bound=0.0;
  // unsigned long size;
  unsigned int  nMetabolites = 0;
  unsigned int  nReactions   = 0;
  int i,j, k;
  int nRealMetabolites, *metaboliteIndexArray ;
  const char **metaboliteName, **metaboliteId ;
  const char **reactions;
  const char * stemp;
  int   * modifierFlag ;
  int   nMatrixElement = 0;

  SBMLDocument_t *d;
  Model_t * m;
  const Species_t **s;
  Reaction_t ** r;
  const ListOf_t   **lr;
  const ListOf_t   **lp;
  const ListOf_t   **lm;
  if( fromString_p ) {
    //printf("From string\n");
    d     = readSBMLFromString(SBMLFileName);
    //printf("Read SBML from string successfully\n");
  } else {
    //    printf("From file\n");
    d = readSBML( SBMLFileName );

  }
  
  m 	= SBMLDocument_getModel(d);
  // get the number of metabolites and reactions 
  nMetabolites = Model_getNumSpecies(m);
  nReactions = Model_getNumReactions(m);
  nRealMetabolites = nMetabolites;

  //printf("Number of metabolites: %d\nNumber of rxns: %d\n", nRealMetabolites, nReactions );

  metaboliteName = (const char **) malloc(nMetabolites*sizeof(char *));
  metaboliteId = (const char **) malloc(nMetabolites*sizeof(char *));
  reactions   = (const char **) malloc(nReactions*sizeof(char *));
  s           = (const Species_t **) malloc(nMetabolites*sizeof(char *));
  r           = (Reaction_t **) malloc(nReactions*sizeof(char *));
  modifierFlag= (int *) malloc(nMetabolites*sizeof(int));
  //  AD          = (annotationData *) calloc(nReactions, sizeof(annotationData) );
  
  /* now create arrays for list of reactants (r) and products (p) and modifiers (m) of each reaction */
  lr	    = (const ListOf_t **) malloc(nReactions*sizeof(char *));
  lp	    = (const ListOf_t **) malloc(nReactions*sizeof(char *));
  lm	    = (const ListOf_t **) malloc(nReactions*sizeof(char *));
  
  for (i=0; i<nMetabolites;i++)
  {
	s[i] = Model_getSpecies(m,i); 
	metaboliteName[i] = Species_getName(s[i]);
	metaboliteId[i]   = Species_getId(s[i]);
	////printf( "        species ID: %s\n", metaboliteId[i]);
	// initialize modifierFlag
	modifierFlag[i] = 0;
  } 

  // now loop thru all reactions
  for (i=0; i<nReactions; i++)
  {
	  r[i] = Model_getReaction(m,i);
	  lr[i] = Reaction_getListOfReactants(r[i]);
	  lp[i] = Reaction_getListOfProducts(r[i]);
	  lm[i] = Reaction_getListOfModifiers(r[i]);

	  // populate the data structure
	  //
	  // 1. Get reaction general info
	  RD[i].ReactionID = Reaction_getId(r[i]);
	  RD[i].isReversible = Reaction_getReversible(r[i]);
	  RD[i].nReactants = ListOf_getNumItems(lr[i]);
	  RD[i].nProducts  = ListOf_getNumItems(lp[i]);
	  RD[i].nModifiers = ListOf_getNumItems(lm[i]);
	  
	  //printf( "reaction ID: %s, reversible=%d, Nreactant=%d, Nproducts=%d, Nmodifiers=%d\n", RD[i].ReactionID, RD[i].isReversible,RD[i].nReactants, RD[i].nProducts,RD[i].nModifiers );	

	  // 2. Get reactant info
	  RD[i].ReactantID = (const char **) malloc((RD[i].nReactants)*sizeof(char *));
	  RD[i].ReactantStoiCoeff = (double *) malloc((RD[i].nReactants)*sizeof(double));
	  for (j=0; j<RD[i].nReactants;j++)
	  {
		  RD[i].ReactantID[j] = SpeciesReference_getSpecies(ListOf_get(lr[i],j));
		  RD[i].ReactantStoiCoeff[j] = SpeciesReference_getStoichiometry(ListOf_get(lr[i],j)) / (double) SpeciesReference_getDenominator(ListOf_get(lr[i],j)); 
		  
		  //printf("           reactant %d = (%f)%s\n", j, RD[i].ReactantStoiCoeff[j], RD[i].ReactantID[j] );  
				  
	  }

	  // 3.Get product info
	  RD[i].ProductID = (const char **) malloc((RD[i].nProducts)*sizeof(char *));
	  RD[i].ProductStoiCoeff = (double *) malloc((RD[i].nProducts)*sizeof(double));
	  for (j=0; j< RD[i].nProducts ;j++)
	  {
		   RD[i].ProductID[j] = SpeciesReference_getSpecies(ListOf_get(lp[i],j));
		   RD[i].ProductStoiCoeff[j] = SpeciesReference_getStoichiometry(ListOf_get(lp[i],j)) / (double) SpeciesReference_getDenominator(ListOf_get(lp[i],j)) ;

		  //printf("           product %d = (%f)%s\n", j, RD[i].ProductStoiCoeff[j], RD[i].ProductID[j]);
	  }

	  // 4. get modifier info
	  RD[i].ModifierID = (const char **) malloc((RD[i].nModifiers)*sizeof(char *));
	  for (j=0; j<RD[i].nModifiers; j++)
	  {
		  RD[i].ModifierID[j] = ModifierSpeciesReference_getSpecies(ListOf_get(lm[i],j));
		  MarkMetabolites(metaboliteId, modifierFlag , nMetabolites, RD[i].ModifierID[j], &nRealMetabolites);
		  
		  //printf("           Modifiers %d = %s\n", j, RD[i].ModifierID[j]);
	  }
	  
	  stemp = SBase_getAnnotation((const SBase_t *) r[i]);
	  AD[i].index = 0;
	  parseAnnotation(stemp, &AD[i]);
	  //printAnnotation( AD[i] );
	  //printf("Annotation[%d]: \n%s\n", i, stemp);

  }
  // Now the ReactionData data structure is populated.  
  //
  // Construct the list of real metabolites (excluding modifiers)
  metaboliteIndexArray = (int *) malloc(nRealMetabolites*sizeof(int));
  j=0; 
  for (i=0; i<nMetabolites; i++)
  {
	  if (modifierFlag[i] == 0)    // it is real metabolite
	  {
		  metaboliteIndexArray[j] = i;
		  j++;
	  }
  }
	  
  
  // loop over all real metabolites
  nMatrixElement=0;
  for (i=0; i< nRealMetabolites; i++)
  {
	  // loop over all reaction
    	  for (j=0; j<nReactions; j++)
  	  {
		// loop over all reactants of this jth reaction
  		for (k=0; k< RD[j].nReactants; k++)
		{
			// compare the metabolite name contained by this reaction
	  		if (strcmp(metaboliteId[metaboliteIndexArray[i]], RD[j].ReactantID[k]) == 0)
			{
				IROWA[nMatrixElement] = i;
				JCOLA[nMatrixElement] = j;
				DA[nMatrixElement] = -RD[j].ReactantStoiCoeff[k];
				nMatrixElement++;
				// printout stoichiometric coefficients * (-1) to specify the reactant
				// //printf("Metabolites %s in reactions %s with stoichiometric coeff = %f\n", metaboliteId[metaboliteIndexArray[i]] )
			}
		}
		// now do the same for the products
		for (k=0; k<RD[j].nProducts;k++)
		{
			if (strcmp(metaboliteId[metaboliteIndexArray[i]], RD[j].ProductID[k]) == 0)
			{
				// print out the coeffient
				IROWA[nMatrixElement] = i;
				JCOLA[nMatrixElement] = j;
				DA[nMatrixElement] = RD[j].ProductStoiCoeff[k];
				nMatrixElement++;
			}
		}
	  }
	  
  }
  // now copy the dimension
  *NNZA = nMatrixElement;
  *MY   = nRealMetabolites;
  *NX   = nReactions;

  // loop over all reactions to copy the bound values and coefficient of objective function

  //printf("Number of Reactions = %d\n",nReactions); 
  for (i=0; i< nReactions; i++)
  {


    strcpy(RNAME[i], RD[i].ReactionID);
    //printf("Reaction[%d] id: %s. RNAME[%d] = %s\n", i, RD[i].ReactionID, i, RNAME[i] );
  } 

  for (i=0; i<nRealMetabolites; i++){
	strcpy(MNAME[i], metaboliteId[metaboliteIndexArray[i]]);
  }
  //printf( "\n ----------------------- Summary ------------------------ \n" );
  //printf( "        filename:       %s\n" , SBMLFileName      );
  /* printf( " Number of Species:     %d\n",nMetabolites); */
/*   printf( " Number of Metabolites: %d\n",nRealMetabolites); */
/*   printf( " Number of Modifiers:   %d\n",nMetabolites - nRealMetabolites); */
/*   printf( " Number of Reactions:   %d\n",nReactions); */
/*   printf(" Number of non-zero elements of A = %d\n", nMatrixElement); */


  //printf("\n");
  SBMLDocument_free(d);
  return 0;
}

void writeFluxToSBML(const char *sbmldoc, const char * outputFileName, double * flux, int nsize)
{
	SBMLDocument_t *d;
	Model_t * m;
	const Reaction_t ** r;
	int nReactions, i, index;
	const char * stemp;
	char stemp2[10000], *position;
	char xval[8];
	
	d     = readSBML(sbmldoc);
	m     = SBMLDocument_getModel(d);
	nReactions = Model_getNumReactions(m);
	r = (const Reaction_t **) malloc(nReactions*sizeof(char *));
	//printf("\n");
	for (i=0; i<nReactions; i++)
	{
		// get reaction i
		r[i] = Model_getReaction(m,i);
		// now get the annotation for each reaction

		stemp = SBase_getAnnotation((const SBase_t *) r[i]);
		position = strstr(stemp,"</annotation");
		if (position == NULL)
		{
			//printf("</annotation> must exist in the sbml file, program now exits\n");
			exit(1);
		}
		index = abs(stemp - position);

		if (index > 350)
		{
			//printf("Warning: the annotation field is too large!\n %s\n", stemp);
		}

		// convert the fluxvalue to string
			
	
		SBase_setAnnotation((SBase_t *) r[i], stemp2);
		stemp = SBase_getAnnotation((const SBase_t *) r[i]);
		// printf("annotation %d= %s \n", i,stemp);
	}
	writeSBML(d,outputFileName);
	SBMLDocument_free(d);
	// printf("Writting fluxes to sbml file:......... # of Reaction = %d\n",nReactions);
}
char * getBounds(int upper_p, int ilimit, double limit, char *bound ){
  if(ilimit){
    //fprintf( stderr, "%s limit: %f\n", upper_p? "Upper":"Lower", limit );
    sprintf(bound, "%f", limit );
    return bound;
  } else if( upper_p ) {
    return "INF";
  } else {
    return "-INF";
  }
}


void writeAnnotatedFluxToSBML(const char *sbmldoc, const char * outputFileName, annotationData *AD, int fromString_p)
{
  //printf("Writing annotated flux to SBML\n");
  SBMLDocument_t *d;
  Model_t * m;
  const Reaction_t ** r;
  int nReactions, i, index;
  const char * stemp;
  char ubound[20];
  char lbound[20];
  char stemp2[10000], *position;
  char xval[8];
  //printf("SBML Doc:\n%s\n", sbmldoc);
  if(sbmldoc == 0){
    printf("Empty sbml document\n");
  } else {
    if( fromString_p ) {
      d     = readSBMLFromString(sbmldoc);
    } else {
      d = readSBML( sbmldoc );
    } 
    m     = SBMLDocument_getModel(d);
    nReactions = Model_getNumReactions(m);
    r = (const Reaction_t **) malloc(nReactions*sizeof(char *));
    //printf("\n");
    for (i=0; i<nReactions; i++)
      {
	// get reaction i
	r[i] = Model_getReaction(m,i);
	// now set the annotation for each reaction
	sprintf( stemp2, annotationTemplate, 
		 AD[i].fbaType[0], getBounds( 1, AD[i].iupperLimit[0] ,  AD[i].upperLimit[0] , ubound), getBounds( 0, AD[i].ilowerLimit[0] , AD[i].lowerLimit[0] , lbound), AD[i].objectiveValue[0], AD[i].prediction[0],
		 AD[i].fbaType[1], getBounds( 1, AD[i].iupperLimit[1] , AD[i].upperLimit[1] , ubound), getBounds( 0, AD[i].ilowerLimit[1] , AD[i].lowerLimit[1] , lbound), AD[i].objectiveValue[1], AD[i].prediction[1],
		 AD[i].fbaType[2], getBounds( 1, AD[i].iupperLimit[2] , AD[i].upperLimit[2] , ubound), getBounds( 0, AD[i].ilowerLimit[2] , AD[i].lowerLimit[2] , lbound), AD[i].objectiveValue[2], AD[i].prediction[2] );
	
	//printf(stemp2);
	SBase_setAnnotation(( SBase_t *) r[i], stemp2 );
	
	stemp = SBase_getAnnotation((SBase_t *) r[i]);
	//printf("annotation %d= %s \n", i,stemp);
      }
    if( *outputFileName ){
      writeSBML(d,outputFileName);
    } else {
      //printf("Writing to stdout\n");
      fprintf(stdout, writeSBMLToString(d) );
    }
	SBMLDocument_free(d);
	//printf("Writting fluxes to sbml file:......... # of Reaction = %d\n",nReactions);
  }
}

int pToken(const char * Token, const char * s, double *bound)
{
	char *  stemp, *svalue;
	int i,j, flag, index;
	
	stemp = strstr(s,Token); 
	if (stemp ==NULL)
		return (-1);
	svalue= (char *) malloc(25*sizeof(char));

	i = strlen(Token);  
	j = strlen(stemp);
	flag = 0;
	index = 0;

	i++;
	while (i<j)
	{
		if ((flag ==1) && (stemp[i] != '\"'))
		{
			// collect the characters
			svalue[index] = stemp[i];
			index++;
		}
		if ((stemp[i] == '\"') && (flag == 0))
			flag = 1;
		else if ((stemp[i] == '\"') && (flag == 1))
			break;
		i++;
	} 
	svalue[index] = '\0';
	*bound =atof(svalue);
	return 0;
}

void MarkMetabolites(const char ** list, int* metList, int n, const char * met, int * n_real_metabolites)
{
	int i=0;
	while (i< n)
	{
		if (strcmp(list[i],met) == 0)
		{
			metList[i] = 1;
			(*n_real_metabolites)--;
			break;
		}
		i++;
	}
}
/*
int main(int a, char *b[])
{
  int i,j;
  int NX, MY;
  double C[N_REACTIONS];
  double XLOW[N_REACTIONS], XUPP[N_REACTIONS];
  char  IXLOW[N_REACTIONS], IXUPP[N_REACTIONS];
  int   IROWA[N_MATRIX_ELEMENT], JCOLA[N_MATRIX_ELEMENT];
  char  *MetaboliteName[N_METABOLITES], *ReactionName[N_REACTIONS];
  double DA[N_MATRIX_ELEMENT];
  annotationData  AD[N_REACTIONS];

  int NNZA;
  
  for (i=0; i<N_METABOLITES;i++)
    MetaboliteName[i] = (char *) malloc(50*sizeof(char));
  
  for (i=0; i<N_REACTIONS; i++)
    ReactionName[i]   = (char *) malloc(80*sizeof(char));
  
  i = sbmlParser (b[1], &NX, &MY, &NNZA,  IROWA, JCOLA, DA, &AD[0], MetaboliteName, ReactionName);
  //printf(" number of non-zero element = %d\n", NNZA);
  
  for (i=0; i< NNZA;i++)
    {
      printf(" (row %d, col %d) dA= %f \n", IROWA[i], JCOLA[i], DA[i]);
    }
  printf(" NX = %d; MY = %d; NNZA = %d\n", NX, MY, NNZA);
  for (i=0; i<NX;i++)
    {
      
      printf(" (%d,%s))\n ", i,ReactionName[i]);
      for(j = 0; j < 3; j++) {
	printf("Analysis: %s\t LowerBound: %f\tUpperBound: %f\tObjective: %f\n",AD[i].fbaType[j], AD[i].ilowerLimit[j] ? "INF" : AD[i].lowerLimit[j], AD[i].upperLimit[j], AD[i].objectiveValue[j] );
      }
    }
  printf("Metabolites\n");
  for (i=0; i<MY;i++){
    printf("%s \n", MetaboliteName[i]);
  }
  printf("\n");
  return 0;
}
*/
