/*
##########################################################################
# Copyright 2004 Department of Genetics, Harvard Medical School and
# and the President and Fellows of Harvard University.
# This code may not be copied without acknowledgement.
# It may be freely distributed for non-commercial use.
#
# The Department of Genetics, Harvard Medical School shall not be
# liable for any damages suffered by anyone  using, modifying
# or distributing this software or its derivatives.
#
##########################################################################
*/


/**
 * This class is controling class of the FluxorSpreadsheet program.
 * It is invoked by the {@link FluxorSpreadsheet} class which deals with 
 * details such as command line arguments.
 * FluxorFrame builds the actually GUI and populates it with data from
 * a libSBML Model object, or from an sbml file (which it uses to populate
 * a libSBML Model object.
 * Acknowledgement: Some of the code and some of the ideas in this package
 * are based on ideas and code in the BioSpreadsheet
 * written by James M. McCollum <jmccoll2@utk.edu>
 * University of Tennessee - Knoxville
 * http://biocomp.ece.utk.edu
 */
package edu.harvard.med.fluxorspreadsheet;
import java.util.Vector;
import java.util.Iterator;
import java.io.IOException;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.JScrollPane;
import javax.swing.JFrame;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JTabbedPane;
import javax.swing.JFileChooser;
import javax.swing.WindowConstants;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.JOptionPane;
import java.awt.*;
import java.awt.event.*;
import java.io.PrintWriter;
import java.util.HashMap;
import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

public class FluxorFrame extends JFrame {
    private JTabbedPane tabbedPane = new JTabbedPane(JTabbedPane.BOTTOM);
    private JFileChooser fileChooser = new JFileChooser();
    private MetabolismReader metabolismReader = new MetabolismReader();
    private String displayName = "";
    private PrintWriter outputWriter = null;
    private static Metabolism metabolism = null;
    private static final String FLUXORSPREADSHEET_VERSION = "1.0";
    private static final String FLUXORSPREADSHEET_BASE_APP_NAME = 
	"Fluxor Spreadsheet";
    private static final String FLUXORSPREADSHEET_APP_NAME =
	FLUXORSPREADSHEET_BASE_APP_NAME + " " + FLUXORSPREADSHEET_VERSION;
    private static final String FLUXORSPREADSHEET_SPACER = "           ";
    static final String EMPTY_NAME = "";
    private HashMap panelMap = new HashMap();
    private static int toSeconds = 1000;
    private static int FRAME_WIDTH = 960;
    private static int FRAME_HEIGHT =640;

     

    public FluxorFrame()  throws FluxorException, IOException{
	this(new Metabolism(), EMPTY_NAME, null );
    }

    public FluxorFrame(PrintWriter outputWriter)  
	throws FluxorException, IOException {
	this(new Metabolism(), EMPTY_NAME, outputWriter);
    }


    public FluxorFrame(String fileName)  
	throws FluxorException, IOException {
	this(fileName, null);
    }

    public FluxorFrame(String fileName, PrintWriter outputWriter)  
	throws FluxorException, IOException {
	this((new MetabolismReader()).readMetabolismFromFile(fileName),
	     fileName,
	     outputWriter);
    }

    public FluxorFrame(Metabolism metabolism)  
	throws FluxorException, IOException {
	this(metabolism, metabolism.getDisplayName(), null);
    }

    public FluxorFrame(Metabolism metabolism, PrintWriter outputWriter)  
	throws FluxorException, IOException {
	this(metabolism, metabolism.getDisplayName(), outputWriter);
    }


    // We currently don't actually use the display name. It would be used 
    // to display the file name in the title bar.
    public FluxorFrame(Metabolism metabolism, 
		       String displayName, 
		       PrintWriter outputWriter)  
	throws FluxorException, IOException {
        this.metabolism = metabolism;
	if(null == metabolism) {
	    throw new FluxorException("Metabolism is null!");
	} else {
	    this.displayName = metabolism.getDisplayName();
	}
	this.outputWriter = outputWriter;
 	setTitle(FLUXORSPREADSHEET_APP_NAME + FLUXORSPREADSHEET_SPACER +  
		 this.displayName );
 		 
	try {
	    initPanels();
	    initMenu();
	    addWindowListener(new WindowAdapter() {
		    public void windowClosing(WindowEvent e) { 
			onFileExit(); }});
	    setSize(FRAME_WIDTH, FRAME_HEIGHT);
	    setLocation
		((getToolkit().getScreenSize().width - getSize().width)/2,
		 (getToolkit().getScreenSize().height - getSize().height)/2);
	    setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
	    show();
	}
	catch (Throwable e) {
	    System.out.println("exception: " + e.getMessage());
	    e.printStackTrace();
	    //    System.out.println(stackTraceToString(e));
	    throw new FluxorException(e);
	}
    }
    
    private void hangOn(int seconds){
	try {
	    Thread.sleep(seconds*toSeconds);
	}
	catch (InterruptedException e) {
	    // ignore
	}
    }
    public JFileChooser getFileChooser() {
        return fileChooser;
    }
    

    private void initMenu() {

       // Create the File Menu
        JMenu fileMenu = new JMenu("File");
        JMenuItem fileNewMenuItem = new JMenuItem("New");
	// Doesn't make sense until new rows can be added
	fileNewMenuItem.setEnabled(false);  
        JMenuItem fileOpenMenuItem = new JMenuItem("Open");
        JMenuItem fileSaveMenuItem = new JMenuItem("Save");
        JMenuItem fileExitMenuItem = new JMenuItem("Exit");
        fileNewMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) { onFileNew(); }});
	fileOpenMenuItem.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) { onFileOpen();}});
	fileSaveMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) { onFileSave();}});
	fileExitMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) { onFileExit();}});
	fileMenu.add(fileNewMenuItem);
	fileMenu.addSeparator();
	fileMenu.add(fileOpenMenuItem);
	fileMenu.add(fileSaveMenuItem);
	fileMenu.addSeparator();
	fileMenu.add(fileExitMenuItem);
                       
	// Create the View menu 
	JMenu viewMenu = new JMenu("View");
        
	try {
	    Vector panelIDs = metabolism.getPanelIDs();
	    Iterator i = panelIDs.iterator();
	    while (i.hasNext()) {
		final String panelID = (String)i.next();
		JMenuItem viewPanelMenuItem = 
		    new JMenuItem(metabolism.getPanelName(panelID));
		viewPanelMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) { 
			    onViewPanel(panelID); }});
		viewMenu.add(viewPanelMenuItem);
	    }
	} catch (FluxorException e) {
	    throw new RuntimeException( e.fillInStackTrace() );
	}
	

                       
	// Create the Continue Menu (for Biospice)
	JMenu continueMenu = new JMenu("Continue");
	continueMenu.addMenuListener(new MenuListener() {
		public void menuCanceled(MenuEvent e) { 
		    ; };
		public void menuDeselected(MenuEvent e) { 
		    ; };
		public void menuSelected(MenuEvent e) { 
		    onContinue(); };
	    });

	// Create the help menu                         
	JMenu helpMenu = new JMenu("Help");
	JMenuItem helpAboutMenuItem = new JMenuItem("About");
	helpAboutMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) { 
		    onHelpAbout(); }});
	helpMenu.add(helpAboutMenuItem);

	// Create the menu bar containing the menus            
	JMenuBar menuBar = new JMenuBar();
	menuBar.add(fileMenu);
	menuBar.add(viewMenu);
	menuBar.add(continueMenu);
	menuBar.add(helpMenu);
	setJMenuBar(menuBar);
    }
    



    // Selecting View Panel brings that panel to the top
    private void onViewPanel(String panelID) {
        tabbedPane.setSelectedComponent(((FluxorPanel)panelMap.
					 get(panelID)).getScrollPane());
    }
    
    private void onFileOpen() {

        if (checkForChanges()) {
            int returnval = fileChooser.showOpenDialog(this);
            if (returnval != fileChooser.APPROVE_OPTION) {
                return;
            }
	    String fileAbsolutePath = 
		fileChooser.getSelectedFile().getAbsolutePath();
            try {
		metabolism = 
		    metabolismReader.readMetabolismFromFile(fileAbsolutePath);
            } catch (IOException e) {
                JOptionPane.showMessageDialog(this,
					      e.getMessage(),
					      FLUXORSPREADSHEET_APP_NAME,
					      JOptionPane.ERROR_MESSAGE);
                return;
            }
            
	    // Synchronize the tables with the metabolism object
	    updatePanels();
	    //currently don't display file name
	    setTitle(FLUXORSPREADSHEET_APP_NAME + FLUXORSPREADSHEET_SPACER + 
		     metabolism.getDisplayName());
        }
    }
    
    private void onFileNew() {
        if (checkForChanges()) {
	    // Clear the metabolism object
	    metabolism = new Metabolism();
	    // Synchronize the tables with the metabolism object
	    updatePanels();

            setTitle(FLUXORSPREADSHEET_APP_NAME + " - [New Model]");
        }
    }
    private void onFileExit() {
        if (checkForChanges()) {
            try {
		// Send the file to the output (file or std.out) if desired
		if (null!= metabolism && null != outputWriter) {
		    metabolism.output(outputWriter);
		}
            } 
	    finally {
		System.exit(0);
	    }
	}
    }
    
    private void onContinue() {
	// Don't check for changes, just output the sbml file
	
	try {
	    // Send the file to the output (file or std.out) if desired
	    if (null!= metabolism && null != outputWriter) {
		metabolism.output(outputWriter);
	    }
	} 
	finally {
	    System.exit(0);
	}
    }

    private boolean checkForChanges() {
	clearPanelSelections();
	if (metabolism.isDirty()) {
	    int returnval = 
		JOptionPane.showConfirmDialog(this,
					      "Save changes to current model?",
					      FLUXORSPREADSHEET_APP_NAME,
					      JOptionPane.YES_NO_CANCEL_OPTION);
	    if (returnval == JOptionPane.CANCEL_OPTION) {
		return false;
	    } else if (returnval == JOptionPane.YES_OPTION) {
		return saveFile();
	    }
	    return true;
	}
        return true;
    }
    
    private void onFileSave() {
        saveFile();
    }
    
    private void onHelpAbout() {
        JOptionPane.showMessageDialog(this,
				      FLUXORSPREADSHEET_APP_NAME + "\n" +
				      "Church Laboratory\n" +
				      "Department of Genetics\n" +
				      "Harvard University Medical School\n" +
				      "http://harvard.med.edu",
				      "About",
				      JOptionPane.INFORMATION_MESSAGE);
    }
    
    private boolean saveFile() {
	clearPanelSelections();
 	int returnval = fileChooser.showSaveDialog(this);
 	if (returnval != fileChooser.APPROVE_OPTION) {
 	    return false;
 	}
	String fileAbsolutePath = 
	    fileChooser.getSelectedFile().getAbsolutePath();
 	try {
 	    metabolism.save(fileAbsolutePath);
 	} catch (IOException e) {
 	    JOptionPane.showMessageDialog(this,
 					  e.getMessage(),
 					  FLUXORSPREADSHEET_APP_NAME,
 					  JOptionPane.ERROR_MESSAGE);
 	    return false;
 	}
	setTitle(FLUXORSPREADSHEET_APP_NAME + FLUXORSPREADSHEET_SPACER +  
		 metabolism.getDisplayName());
	return true;
    }
    
    
    private void initPanels()  throws FluxorException {
	Vector panelIDs = metabolism.getPanelIDs();
	
	Iterator i = panelIDs.iterator();
	while (i.hasNext()) {
	    String panelID = (String)i.next();
	    FluxorPanelInfo panelInfo = metabolism.getPanelInfo(panelID);
	    FluxorPanel panel = new FluxorPanel(panelInfo);
	    // Add the panel to the list of all panels
	    panelMap.put(panelID, panel);
	    tabbedPane.add(panel.getScrollPane(), panel.getPanelName());
	}

        // All Panels have been created. Now add the tabbed pane to 
	//the Content Pane
        getContentPane().add(tabbedPane);
    }

    public void updatePanels() {
	try {
	    Iterator i = (panelMap.values()).iterator();
 	    while (i.hasNext()) {
 		FluxorPanel panel = (FluxorPanel)i.next();
 		if (null != panel) {
		    panel.updateTable();
 		}
 	    }
	} catch (FluxorException e) {
	    throw new RuntimeException( e.fillInStackTrace() );
	}
    }

    public void clearPanelSelections() {
	Iterator i = (panelMap.values()).iterator();
	while (i.hasNext()) {
	    FluxorPanel panel = (FluxorPanel)i.next();
	    if (null != panel) {
		panel.stopEditing();
		panel.clearSelection();
	    }
	}
    }

    public static Metabolism getMetabolism() {
	return metabolism;
    }

    public String stackTraceToString (Throwable e) {
	StringBuffer sb = new StringBuffer("");
	StackTraceElement[] stackTrace = e.getStackTrace();
	for (int i = 0; i< stackTrace.length; i++) {
	    sb.append(stackTrace.toString());
	}
	return sb.toString();
    }
}


    


