/*
 * Metabolism.java
 *
 * Created on July 4, 2004, 12:26 PM
 */


package edu.harvard.med.fluxorspreadsheet;

import java.util.Vector;
import java.util.HashMap;
import java.util.Iterator;
import java.io.IOException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import libsbml.*;

/**
 * This class is the interface through which other code interacts with the
 * underlying libSBML Model object.
 * The libSBML Model object is created by the {@link MetabolismReader} class.
 * The {@link FluxorTableModel}  interacts with the Metabolism class to 
 * obtain and
 * update the values in the Reaction and Species panels. The two sets of
 * values (displayed and in the libSBML Model object) are kept synchronized.
 * The Metabolism class can also produce a SBML String representation of the
 * model to be used for output.
 * 
 * The Metabolism class is also the interface for obtain information about the
 * panels (e.g. Species panel and Reaction panel). This includes panel ID, name
 * and information (metadata) about the columns of the panel.
 *
 * @author  Aaron Brandes
 * @version 1.0
 */
public class Metabolism {

    private Model model;
    private HashMap panelMap;
    private HashMap panelMetaDataMap;
    private boolean dirty = false;
    private boolean hasPredictions;
    
    
    /**
     * Creates a new instance of Metabolism
     * @param model the <code>Model</code> represented by this Metabolism
     * @param panelMap a <code>HashMap</code> of panel IDs
     * @param panelMetaDataMap a <code>HashMap</code> with data describing
     *                         the panels
     */
    public Metabolism(Model model, HashMap panelMap, HashMap panelMetaDataMap) {
	// Until the model is modified, it should be considered "not dirty"
	setDirty(false);
	this.model = model;
	this.panelMap = panelMap;
	this.panelMetaDataMap = panelMetaDataMap;
	// libsbml is alsays saying there is no annotation
	// this works around a bug 
	model.unsetAnnotation();
	model.setAnnotation(getAnalysisTypesAnnotation());

	// As long as we are playing with the annotation, we had better save
	// newly created Metabolism
	// Also when creating a ReactionRow we may modify its annotation.
	// in the future it should communicate this.
	setDirty(true);
    }
    /**
     * Creates a new empty <code>Metabolism</code> instance.
     *
     */
    public Metabolism(){
	this( new Model("foo", "bar"), new HashMap(), new HashMap() );
    }
   
    /**
     * Creates a new <code>Metabolism</code> instance.
     *
     * @param id    <code>Model</code> id
     * @param name   <code>Model</code> name
     */
    public Metabolism(String id, String name) {
	this(new Model(id, name), new HashMap(), new HashMap() );
    }
    /**
     *  <code>getPanelInfo</code> gets a <code>FluxorPanelInfo</code>
     *  object, given its key.
     * @param panelID id for desired <code>FluxorPanelInfo</code>
     * @return the <code>FluxorPanelInfo</code> object
     * @exception FluxorException if an error occurs
     */
    public FluxorPanelInfo getPanelInfo( String panelID ) 
	throws FluxorException {
	if( panelMetaDataMap.containsKey( panelID ) ){
	    FluxorPanelMetaData panelMetaData = 
		(FluxorPanelMetaData)panelMetaDataMap.get( panelID );
	    return panelMetaData.getPanelInfo();
	} else {
	    throw new FluxorException("KeyError: " + panelID + 
				      " is not a valid Panel ID" );
	}
    }


    /**
     * Describe <code>getDisplayName</code> method here.
     *
     * @return  Model name for user display
     */
    public String getDisplayName() {
	if( null == model ){
	    return "Empty model!";
	}
	else if( model.isSetName() != 0 ) {
	    return model.getName();
	} else if ( model.isSetId() ) {
	    return model.getId();
	} else {
	    return "No Name or ID found!";
	}
    }
   
    
    /**
     * Get <code>Vector</code> of all panel IDs
     *
     * @return a <code>Vector</code> of all panel IDs
     */
    public Vector getPanelIDs(){
	return new Vector(panelMap.keySet());
    }
    /**
     * Describe <code>getRowIDs</code> method here.
     *
     * @param panelID a <code>String</code> value
     * @return a <code>Vector</code> value
     * @exception FluxorException if an error occurs
     */
    public Vector getRowIDs(String panelID) throws FluxorException {
	if( !panelMap.containsKey( panelID ) ){
	    throw new FluxorException("Key error: " + panelID + 
				      " is not a valid panelID");
	}
	MetabolismPanel metPanel = (MetabolismPanel)panelMap.get(panelID);		Vector rows = metPanel.getPanelRowIDs();
	return rows ;
    }
	
    /**
     * Get name of a panel
     *
     * @param panelID panel id
     * @return panel name
     * @exception FluxorException if an error occurs
     */
    public String getPanelName( String panelID ) throws FluxorException {
	if( panelMap.containsKey( panelID ) ){
	    MetabolismPanel panel = (MetabolismPanel)panelMap.get( panelID );
	    return panel.getPanelName();
	} else {
	    throw new FluxorException("Key error: " + panelID + 
				      " is not a valid panelID");
	}
    }
    /**
     * Get <code>Vector</code> of all column IDs for a panel
     *
     * @param panelID the panel ide
     * @return a <code>Vector</code> of all column IDs for the panel
     * @exception FluxorException if an error occurs
     */
    public Vector getColumnIDs( String panelID ) throws FluxorException {
	if( panelMap.containsKey( panelID ) ){
	    Vector columnIDs = new Vector();
	    FluxorPanelMetaData panelMetaData = 
		(FluxorPanelMetaData)panelMetaDataMap.get( panelID );
	    FluxorColumnMetaData columnMetaData = 
		panelMetaData.getColumnMetaData();
	    columnIDs = columnMetaData.getColumnIDs();
	    return columnIDs;
	} else {
	    throw new FluxorException("Key error: " + panelID + 
				      " is not a valid panelID");
	}

    }

    /**
     * Indicates whether model already has predictions
     * @return true - has predictins
     *         <code>false</code> does not have predictions
     */
    public boolean isHasPredictions() {
	return hasPredictions;
    }

    /**
     * Set the HasPredictions value.
     * @param newHasPredictions The new HasPredictions value.
     */
    public void setHasPredictions(boolean newHasPredictions) {
	this.hasPredictions = newHasPredictions;
    }

    /**
     * Describe <code>getValue</code> method here.
     *
     * @param panelID a <code>String</code> value
     * @param rowID a <code>String</code> value
     * @param columnID a <code>String</code> value
     * @return an <code>Object</code> value
     * @exception FluxorException if an error occurs
     */
    public Object getValue(String panelID,
			   String rowID,
			   String columnID) throws FluxorException {
	if( !panelMap.containsKey( panelID ) ){
	    throw new FluxorException("Key error: " + panelID + 
				      " is not a valid panelID");
	} 
	MetabolismPanel metPanel = (MetabolismPanel)panelMap.get(panelID);
	PanelRow panelRow = metPanel.getPanelRow(rowID);
	if( null == panelRow  ) {
	    throw new FluxorException("Key error: " + rowID + 
				      " is not a valid rowID");
	}
	HashMap rowMap = panelRow.getColumnData();
	if( !rowMap.containsKey( columnID ) ) {
	    throw new FluxorException("Key error: " + columnID + 
				      " is not a valid columnID");
	}

	if( !rowMap.containsKey( columnID ) ) {
	    throw new FluxorException("Key error: " + columnID + 
				      " is not a valid columnID");
	}

	return rowMap.get(columnID);

    }


    /**
     * Describe <code>setValue</code> method here.
     *
     * @param panelID a <code>String</code> value
     * @param rowID a <code>String</code> value
     * @param columnID a <code>String</code> value
     * @param value an <code>Object</code> value
     * @exception FluxorException if an error occurs
     */
    public void setValue(String panelID,
			 String rowID,
			 String columnID,
			 Object value) throws FluxorException {

	if( !panelMap.containsKey( panelID ) ) {
	    throw new FluxorException("Key error: " + panelID + 
				      " is not a valid panelID");
	} 

	MetabolismPanel metPanel = (MetabolismPanel)panelMap.get(panelID);
	PanelRow panelRow = metPanel.getPanelRow(rowID);
	if( null == panelRow  ) {
	    throw new FluxorException("Key error: " + rowID + 
				      " is not a valid rowID");
	}
	HashMap rowMap = panelRow.getColumnData();
	if( !rowMap.containsKey( columnID ) ) {
	    throw new FluxorException("Key error: " + columnID + 
				      " is not a valid columnID");
	}

	rowMap.put(columnID, value);
	panelRow.setColumnData(rowMap);
	setDirty(true);
    }


    /**
     * XXX: Not implemented
     * @param panelID a <code>String</code> value
     * @param columnID a <code>String</code> value
     * @return a <code>Vector</code> value
     */
    public Vector getColumnRange(String panelID,
				 String columnID) {
	return new Vector();
    }

    public void output(PrintWriter out) {
	out.println(toString());
	out.close();
	setDirty(false);
    }

    public void save(String fileAbsolutePath)  throws IOException {
	PrintWriter out = 
	    new PrintWriter(new FileOutputStream(fileAbsolutePath));
	output(out);
    }


    public void toStdOut()  throws IOException {
	output(new PrintWriter(System.out));
    }

    /**
     * Gets the value of dirty
     *
     * @return the value of dirty
     */
    public boolean isDirty()  {
	return this.dirty;
    }

    /**
     * Sets the value of dirty
     *
     * @param argDirty Value to assign to this.dirty
     */
    public void setDirty(boolean argDirty) {
	this.dirty = argDirty;
    }

    public String toString(){
	try {
	    //return MetabolismReader.writeSBMLToString();
	    StringBuffer sb = new StringBuffer("");
	    sb.append("<sbml xmlns:biopax=\"http:");
	    sb.append("//www.biopax.org/release/biopax-level1.owl#\"\n");
	    sb.append("xmlns=\"http://www.sbml.org/sbml/level2\" ");
	    sb.append("level=\"2\" version=\"1\">\n");
	    sb.append(model.toSBML(2, 1));
	    sb.append("</sbml>\n");
	    return sb.toString();
	} catch (Exception e) {
	    return e.toString();
	}
    }

    public String toTableString() throws FluxorException {
	StringBuffer sb = new StringBuffer("Metabolism:\n\n");
	Vector panelIDs = getPanelIDs();
	Iterator i = panelIDs.iterator();
	while( i.hasNext() ){
	    String panelID = (String)i.next();
	    sb.append("Panel: " + panelID + "\n");
	    Vector columnIDs = getColumnIDs( panelID );
	    Iterator j = columnIDs.iterator();
	    while( j.hasNext() ){
		String columnID = (String)j.next();
		sb.append(columnID + "\t");
	    }
	    sb.append("\n");
	    Iterator k = getRowIDs( panelID ).iterator();
	    while( k.hasNext() ){
		String rowID = (String)k.next();
		Iterator l = columnIDs.iterator();
		while( l.hasNext() ){
		    String columnID = (String)l.next();
		    Object value = getValue( panelID, rowID, columnID );
		    sb.append(value.toString() + "\t" );
		}
		sb.append("\n");
	    }
	}
	return sb.toString();
    }

    static String getAnalysisTypesAnnotation() {
	StringBuffer sb = new StringBuffer("");
	sb.append("    <annotation xmlns:flux=\"");
	sb.append("http://arep.med.harvard.edu/fluxns\">\n");
	sb.append("       <flux:listOfAnalysisTypes>\n");
	sb.append("         <flux:analysisType id=\"fba\"");
	sb.append(" name=\"Flux Balance Analysis\"/>\n");
	sb.append("         <flux:analysisType id=\"moma\"");
	sb.append(" name=\"Minimization of Metabolic Adjustment\"/>\n");
	sb.append("       </flux:listOfAnalysisTypes>\n");
	sb.append("       <flux:listOfAnalyses>\n");
	sb.append("         <flux:analysis id=\"wild_type\"");
	sb.append(" name=\"Wild-Type FBA\" analysisType=\"fba\"/>\n");
	sb.append("         <flux:analysis id=\"mutant_fba\"");
	sb.append(" name=\"Mutant FBA\" analysisType=\"fba\"/>\n");
	sb.append("         <flux:analysis id=\"mutant_moma\"");
	sb.append(" name=\"Mutant MOMA\" analysisType=\"moma\"/>\n");
	sb.append("       </flux:listOfAnalyses>\n");
	sb.append("    </annotation>");
	return sb.toString();
    }

    static {
	System.loadLibrary("sbmlj");
    }


}


