/**
 * This class creates the underlying libSBML Model object from sbml text
 * contained in a string or file. Other classes interact with the libSBML 
 * Model object via the {@link Metabolism} class.
 *
 * The MetabolismReader class  also intiates the collection of information 
 * about the panels (e.g. Species panel and Reaction panel). This includes
 * panel ID, name and information (metadata) about the columns of the panel.
 */

package edu.harvard.med.fluxorspreadsheet;
import libsbml.*;
import java.util.*;
import java.io.IOException;

public class MetabolismReader {
    Model model;
    HashMap panelMap;
    HashMap panelMetaDataMap;

    public MetabolismReader(){
	panelMap = new HashMap();
	panelMetaDataMap = new HashMap();
    }

    public void setModelFromFile( String filename ) throws IOException {
	SBMLReader reader = 
	    new SBMLReader( libsbml.XML_SCHEMA_VALIDATION_NONE ); 
	SBMLDocument d = reader.readSBML( filename );
	Assert( d != null, "d != null" );
	Assert( d.getLevel  () == 2 , "d.getLevel() == " + d.getLevel() );
	Assert( d.getVersion() == 1 , " d.getVersion() == " + d.getVersion() );
	Assert( d.getNumErrors() == 0, " d.getNumErrors() == " + 
		d.getNumErrors() );
	model = d.getModel();
    }

    public void setModelFromString( String sbml ){
	SBMLReader reader = new SBMLReader( libsbml.XML_SCHEMA_VALIDATION_NONE ); 
	SBMLDocument d = reader.readSBMLFromString( sbml );
	Assert( d != null, "d != null" );
	Assert( d.getLevel  () == 2 , "d.getLevel() == " + d.getLevel());
	Assert( d.getVersion() == 1 , " d.getVersion() == " + d.getVersion());
	Assert( d.getNumErrors() == 0, " d.getNumErrors() == " + d.getNumErrors() );
	model = d.getModel();
    }

    public Model getModel() {
	return this.model;
    }

    Metabolism readMetabolismFromFile( String filename ) throws IOException{
	setModelFromFile( filename );
	return getMetabolism();
    }


    Metabolism readMetabolismFromString(String sbml) throws IOException{
	setModelFromString( sbml );
	
	return getMetabolism();
    }


    Metabolism getMetabolism(){
	MetabolismPanel speciesPanel = new SpeciesPanel( getModel() );
	MetabolismPanel rxnPanel = new ReactionPanel( getModel() );
	try {
	    panelMetaDataMap.put( speciesPanel.getPanelID(), 
				  speciesPanel.getPanelMetaData() );
	    panelMetaDataMap.put( rxnPanel.getPanelID(), 
				  rxnPanel.getPanelMetaData() );
	    panelMap.put( speciesPanel.getPanelID(), speciesPanel );
	    panelMap.put( rxnPanel.getPanelID(), rxnPanel );
	} catch( ClassNotFoundException  e) {
	    e.printStackTrace();
	}

	return new Metabolism( getModel(), panelMap, panelMetaDataMap );
    }
    
    static void Assert(boolean condition, String errorMessage)
    {
	if (condition == false)
	    {
		System.err.println(errorMessage);
		System.exit(0);
	    }
    }
    
}


