package edu.harvard.med.fluxorspreadsheet;
import java.util.regex.*;
import java.util.HashMap;
import libsbml.*;

/**
 * The ReactionAnnotation class handle the parsing and updating of the
 * annotation the defines flux limits, objective and predictions for the
 * wild type, mutant fba and mutant moma compuational experiments.
 * It is used by the {@link ReactionRow} class.
 */

// Some approaches taken here are just for release 1
// the analysis types output in the toString method are
// "wild_type", "mutant_moma", "mutant_fba"

public class ReactionAnnotation {
    private String reactionID;
    private HashMap analysisAnnotationMap = new HashMap();
    HashMap defaultValuesMap;
    boolean modifiedOnInitialization = false;
    
    final String DEFAULT_VALUE = "0";
    final String [] DEFAULT_VALUES = {"-INF", "INF", "0", "0"};
    final String DEFAULT_IRREVERSIBLE_LOWER_BOUND = "0";
    final static String WILD_TYPE = "wild_type";
    final static String MUTANT_MOMA = "mutant_moma";
    final static String MUTANT_FBA = "mutant_fba";

    final static String LOWER_ANNOTATION = "lower";
    final static String UPPER_ANNOTATION = "upper";

    final static String OBJECTIVE_ANNOTATION = "objective";
    final static String PREDICTION_ANNOTATION = "prediction";

    final String[] analysisTypes = {WILD_TYPE,
				    MUTANT_FBA,
				    MUTANT_MOMA,};

    final String[] annotationAttributes = {LOWER_ANNOTATION,
					   UPPER_ANNOTATION,
					   OBJECTIVE_ANNOTATION,
					   PREDICTION_ANNOTATION};
    static int NUM_ATTRIBUTES = 4;
    private static boolean existPredictions = true;

    
    private static String GET_ANALYSIS = 
	"analysis=\"([A-Za-z0-9_]*)\"(.*?)>";
    private static String MATCH_FLUX_ANNOTATION = 
	"([A-Za-z0-9_]*)=\"([A-Za-z0-9.-]*)\"";
   private static String MATCH_PREDICTION_ANNOTATION = 
	"prediction=\"[0-9.-]*\"";
    {
	defaultValuesMap = new HashMap();
	for (int j = 0; j < NUM_ATTRIBUTES; j++) {
	    defaultValuesMap.put(annotationAttributes[j],
				 DEFAULT_VALUES[j]);
	}
    }

    public ReactionAnnotation( String annotationString,
			       boolean isReversible ){
	modifiedOnInitialization = false;
	String [] defaultValues = new String[DEFAULT_VALUES.length];
	System.arraycopy(DEFAULT_VALUES,  0, 
			 defaultValues, 0, DEFAULT_VALUES.length);

	if (!isReversible) {
	    defaultValues[0] = DEFAULT_IRREVERSIBLE_LOWER_BOUND;
	}

	HashMap annotationMap;
	parseAnnotation( annotationString );

	// Handle lower and upper bounds as a special cases, because
	// the reaction may be irreversible, or wild_type lower or
	// upper bounds could be specified, and must be propagated
	// if the mutant fba or moma value is not specified
	if (analysisAnnotationMap.containsKey(WILD_TYPE) ){
	    annotationMap = 
		    (HashMap)analysisAnnotationMap.get(WILD_TYPE);
	    for (int j = 0; j < 1; j++) {
		if (annotationMap.containsKey(annotationAttributes[j])) {
		    // bound exists, set it as the default value
		    defaultValues[j] = annotationAttributes[j];
		}
	    }
	}

	//Add default values for any annotations that were not specified


	for (int i = 0; i < analysisTypes.length; i++) {
	    if (analysisAnnotationMap.containsKey(analysisTypes[i]) ){
		annotationMap = 
		    (HashMap)analysisAnnotationMap.get(analysisTypes[i]);
	    }
	    else {
		// Create a Map for this analysis type
		annotationMap = new HashMap();
		analysisAnnotationMap.put(analysisTypes[i],
					  annotationMap);
	    }

	    for (int j = 0; j < NUM_ATTRIBUTES; j++) {
		if (!annotationMap.containsKey(annotationAttributes[j])) {
		    // If attribute does not exist, give it the default value
		    annotationMap.put(annotationAttributes[j],
				      defaultValues[j]);
		    // The annotation has been modified and must be 
		    // set in the model (in ReactionRow)
		    modifiedOnInitialization = true;
		}
	    }
	}
    }

    public ReactionAnnotation(boolean isReversible){
	// since this annotation was just created, it should be considered
	// modified, so it will be put into the model
	modifiedOnInitialization = true;
	String defaultLowerBound = DEFAULT_VALUES[0];
	if (!isReversible) {
	    defaultLowerBound = DEFAULT_IRREVERSIBLE_LOWER_BOUND;
	}
	analysisAnnotationMap = new HashMap();
	for (int i = 0; i < analysisTypes.length; i++) {
	    HashMap annotationMap = new HashMap();
	    // Handle lower bound as a special case, because
	    // the reaction may be irreversible
	    annotationMap.put(annotationAttributes[0],
			      defaultLowerBound);
	    for (int j = 1; j < NUM_ATTRIBUTES; j++) {
		annotationMap.put(annotationAttributes[j],
				  DEFAULT_VALUES[j]);
	    }
	    analysisAnnotationMap.put(analysisTypes[i],
				      annotationMap);
	}
    }

    public void parseAnnotation( String annotationString ){
	Pattern analysisPattern = Pattern.compile(GET_ANALYSIS);
	Matcher analysisMatcher = analysisPattern.matcher(annotationString);
	while (analysisMatcher.find() && 
	       analysisMatcher.groupCount() > 0) {
	    String analysisType =  analysisMatcher.group(1);
	    String analysisAnnotation = analysisMatcher.group(2);
	    HashMap annotationMap = new HashMap();
	    Pattern pattern = Pattern.compile(MATCH_FLUX_ANNOTATION);
	    Matcher matcher = pattern.matcher(analysisAnnotation);
	    while(matcher.find()) {
		String annotationType = matcher.group(1);
		String annotationValue = matcher.group(2);
		annotationMap.put(annotationType, annotationValue);
	    }
	    analysisAnnotationMap.put(analysisType,annotationMap); 
	}	    
    }
    
    public String getLower(String analysis ) {
	return getValue(analysis, LOWER_ANNOTATION); 
    }

    public void setLower(String analysis,  String lowerStr){
	setValue(analysis, LOWER_ANNOTATION, lowerStr );
    }
    
    public String getUpper( String analysis ){
	return getValue(analysis, UPPER_ANNOTATION); 
    }

    public void setUpper(String analysis, String upperStr){
	setValue(analysis, UPPER_ANNOTATION, upperStr );
    }

 
    public String getObjective( String analysis ){
	return getValue(analysis, OBJECTIVE_ANNOTATION ); 
    }

    public void setObjective(String analysis,  String objectiveStr ){
	setValue(analysis, OBJECTIVE_ANNOTATION,  objectiveStr);

    }

    public String getPrediction( String analysis ){
	return getValue(analysis, PREDICTION_ANNOTATION); 
    }

    public String getValue(String analysis, String key) {
	String value = DEFAULT_VALUE;
	if (defaultValuesMap.containsKey(key)) {
	    value = (String)defaultValuesMap.get(key);
	}
	if (analysisAnnotationMap.containsKey(analysis ) ){
	    HashMap annotationMap = 
		(HashMap)analysisAnnotationMap.get(analysis);
	    if ((null != annotationMap) &&
		annotationMap.containsKey(key)) {
		value =  (String)annotationMap.get(key);
	    }
	}
	return value;
    }


    public void setValue(String analysis, String key, String value) {
	if (!analysisAnnotationMap.containsKey(analysis ) ){
	    // throw new FluxorException("Annotation error: " + analysis + 
	    // " is not a valid analysis");
	    System.out.println("Annotation error: " + analysis + 
			       " is not a valid analysis");
	}

	HashMap annotationMap = 
	    (HashMap)analysisAnnotationMap.get(analysis);
	if ((null == annotationMap) ||
	    !annotationMap.containsKey(key)) {
	    //throw new FluxorException("Annotation error: " + key + 
	    // " is not a valid annotation type");
	    System.out.println("Annotation error: " + key + 
			       " is not a valid annoation type");
	}
	annotationMap.put(key, value);
    }


    public static boolean hasPrediction( Reaction reaction ){
	boolean hasPrediction = false;
	if (reaction.isSetAnnotation()) {
	    String annotationString = reaction.getAnnotation();
	    Pattern analysisPattern = 
		Pattern.compile(MATCH_PREDICTION_ANNOTATION);
	    Matcher analysisMatcher = 
		analysisPattern.matcher(annotationString);
	    hasPrediction = analysisMatcher.find() ;
	}
	return hasPrediction;
    }

 
    // if there are no predictions, there is one less annotation
    // attribute type
    public static void setNoPredictions () {
	NUM_ATTRIBUTES -= 1;
	existPredictions = false;
    }


    /**
     * Get the ModifiedOnInitialization value.
     * @return the ModifiedOnInitialization value.
     */
    public boolean isModifiedOnInitialization() {
	return modifiedOnInitialization;
    }

    public String toString() {
	StringBuffer sb
	    = new StringBuffer
	    ("<annotation xmlns:flux=\"http://arep.med.harvard.edu/fluxns\">\n");
	sb.append("          <flux:limit analysis=\"" + WILD_TYPE + "\" ");

	sb.append(LOWER_ANNOTATION + "=\"" + 
		  getLower(WILD_TYPE) + "\" ");
	sb.append(UPPER_ANNOTATION + "=\"" + 
		  getUpper(WILD_TYPE) + "\" ");
	sb.append(OBJECTIVE_ANNOTATION + "=\"" + 
		  getObjective(WILD_TYPE) + "\"");
	if (existPredictions) {
	    sb.append(" " + PREDICTION_ANNOTATION + "=\"" + 
		      getPrediction(WILD_TYPE) + "\"");
	}
	sb.append("/>\n");

	sb.append("          <flux:limit analysis=\"" + MUTANT_FBA + "\" ");
	sb.append(LOWER_ANNOTATION + "=\"" + 
		  getLower(MUTANT_FBA) + "\" ");
	sb.append(UPPER_ANNOTATION + "=\"" + 
		  getUpper(MUTANT_FBA) + "\" ");
	sb.append(OBJECTIVE_ANNOTATION + "=\"" + 
		  getObjective(MUTANT_FBA) + "\"");
	if (existPredictions) {
	    sb.append(" " + PREDICTION_ANNOTATION + "=\"" + 
		      getPrediction(MUTANT_FBA) + "\"");
	}
	sb.append("/>\n");


	sb.append("          <flux:limit analysis=\"" + MUTANT_MOMA + "\" ");
	sb.append(LOWER_ANNOTATION + "=\"" + 
		  getLower(MUTANT_MOMA) + "\" ");
	sb.append(UPPER_ANNOTATION + "=\"" + 
		  getUpper(MUTANT_MOMA) + "\" ");
	sb.append(OBJECTIVE_ANNOTATION + "=\"" + 
		  getObjective(MUTANT_MOMA) + "\"");
	if (existPredictions) {
	    sb.append(" " + PREDICTION_ANNOTATION + "=\"" + 
		      getPrediction(MUTANT_MOMA) + "\"");
	}
	sb.append("/>\n");

	sb.append("      </annotation>");
	return sb.toString();
    }
}


       
