/**
 * The ReactionPanel class  collects (metadata) about the columns of the 
 * Reaction and libSBML Reaction objects. These libSBML Reaction objects are
 * referenced by their IDs, and form the link between the row data in the
 * GUI reaction panel, and the {@link Metabolism  }Metabolism  class and the
 * libSBML Model object.
 * Note: the libSBML Reaction objects as well as the column metadata are
 * encapsulated in the {@link ReactionRow} class.
 */
package edu.harvard.med.fluxorspreadsheet;
import libsbml.*;
import java.util.*;
import java.lang.reflect.*;

public class ReactionRow implements MetabolismReaction, PanelRow {
    Reaction reaction;
    ReactionAnnotation reactionAnnotation;
    private static final int ID = 0;
    private static final int NAME = 1;
    private static final int CLASS = 2;
    private static final int EQN = 2;
    private static final int WTLOWBND = 3;
    private static final int WTUPBND = 4;
    private static final int KO = 5;
    private static final int OBJ = 6;
    private static final int WTFBA = 7;
    private static final int MUTFBA = 8;
    private static final int MOMA = 9;
    private static boolean existPredictions = true;
    private static int NUM_COLUMNS = 10;
    private static int NUM_ANALYSES = 3;
	
    private static String [][] reactionColumnMetaData =
	//     ID           Name                         Class               
	{
	    {"rxnID",       "Reaction ID",               "java.lang.String"},
	    {"rxnName",     "Reaction Name",             "java.lang.String"},
	    {"rxnEqn",      "Reaction Equation",         "java.lang.String"},
	    //"edu.harvard.med.fluxorspreadsheet.ReactionEquation"},
	    {"rxnWTLowBnd", "Lower Bound",     "java.lang.String"},
	    //{"rxnWTLowBnd", "Wild Type Lower Bound",     "java.lang.String"},
            //"edu.harvard.med.fluxorspreadsheet.FluxLimit"},
	    {"rxnWTUpBnd",  "Upper Bound",     "java.lang.String"},
	    //{"rxnWTUpBnd",  "Wild Type Upper Bound",     "java.lang.String"},
	    //"edu.harvard.med.fluxorspreadsheet.FluxLimit"},
	    {"rxnKO",       "Knockout",         "java.lang.Boolean"},
	    //{"rxnKO",       "Reaction Knockout",         "java.lang.Boolean"},
	    {"rxnObjCoef",  "Objective Coefficent",      "java.lang.Float"},
	    {"rxnWTFBA",    "Wild Type FBA",             "java.lang.Float"},
	    {"rxnMutFBA",   "Mutant FBA",                "java.lang.Float"},
	    {"rxnMOMA",     "MOMA",                      "java.lang.Float"}   
	};


    private static Boolean[] reactionColumnEditableData =
	//  Editable            ID           


	{
	    new Boolean(false), //"rxnID"       
	    new Boolean(false), //"rxnName"     
	    new Boolean(false), //"rxnEqn"      
	    new Boolean(true), //"rxnWTLowBnd" 
	    new Boolean(true), //"rxnWTUpBnd"  
	    new Boolean(true), //"rxnKO"       
	    new Boolean(true), //"rxnObjCoef"  
	    new Boolean(false), //"rxnWTFBA"    
	    new Boolean(false), //"rxnMutFBA"   
	    new Boolean(false), //"rxnMOMA"     
	};

    private static final String rxnID = reactionColumnMetaData[ID][ID];
    private static final String rxnName = reactionColumnMetaData[NAME][ID];
    private static final String rxnEqn = reactionColumnMetaData[EQN][ID];
    private static final String rxnWTUpBnd = 
	reactionColumnMetaData[WTUPBND][ID];
    private static final String rxnWTLowBnd = 
	reactionColumnMetaData[WTLOWBND][ID];
    private static final String rxnKO = reactionColumnMetaData[KO][ID];
    private static final String rxnObjCoef = reactionColumnMetaData[OBJ][ID];
    private static final String rxnWTFBA = reactionColumnMetaData[WTFBA][ID];
    private static final String rxnMutFBA = 
	reactionColumnMetaData[MUTFBA][ID];
    private static final String rxnMOMA = reactionColumnMetaData[MOMA][ID];

   

    public ReactionRow( Reaction reaction ) {
	this.reaction = reaction;
	this.reactionAnnotation = computeReactionAnnotation();
	//If reaction did not have an annotation, or the annotation
	//needed to be modified because some values were not initialized,
	// set the annotation in the reaction string, and make sure the
	// changes will be set when the model is output

	if (reactionAnnotation.isModifiedOnInitialization()) {
	    String annotation = reactionAnnotation.toString();
	    reaction.setAnnotation(annotation);
	    // In the future we should also call
	    // metabolism.setDirty(true);
	    // but we don't have access to the metabolism object now
	    // In consequence, newly created Metabolims objects will
	    // always be considered dirty.
	}
    }

    public HashMap getColumnData() {
	HashMap columnData = new HashMap();
	columnData.put( rxnID, getReactionID() );
	columnData.put( rxnName, getReactionName() );
	columnData.put( rxnEqn, getReactionEquation() );
	columnData.put( rxnWTUpBnd, getReactionWildTypeUpperBound() );
	columnData.put( rxnWTLowBnd, getReactionWildTypeLowerBound() );
	columnData.put( rxnKO, isReactionKnockout() );
	columnData.put( rxnObjCoef, getObjectiveCoefficient() );
	// Only display prediction data if some exists
	if (existPredictions) {
	    columnData.put( rxnWTFBA, getWildTypeFBA() );
	    columnData.put( rxnMutFBA, getMutantFBA() );
	    columnData.put( rxnMOMA, getMOMA() );
	}
	
	return columnData;
    }
    public static FluxorColumnMetaData 
	getColumnMetaData() throws ClassNotFoundException{

	Vector columnIDs = new Vector();
	HashMap columnIndexMap = new HashMap();
	HashMap columnNameMap = new HashMap();
	HashMap columnClassMap = new HashMap();
	HashMap columnEditableMap = new HashMap();
	
	
	for(int i=0; i < NUM_COLUMNS; i++ ){
	    columnIDs.add( reactionColumnMetaData[i][ID] );
	    columnIndexMap.put( reactionColumnMetaData[i][ID],
				new Integer(i) );
	    columnNameMap.put( reactionColumnMetaData[i][ID],
			       reactionColumnMetaData[i][NAME] );
	    columnClassMap.put(reactionColumnMetaData[i][ID],
			       Class.forName( reactionColumnMetaData
					      [i][CLASS] ));
	    columnEditableMap.put(reactionColumnMetaData[i][ID],
				  reactionColumnEditableData[i]);
				  
	}
	return new FluxorColumnMetaData(columnIDs,
					columnIndexMap,
					columnNameMap,
					columnClassMap,
					columnEditableMap);
    }

    
    public void setColumnData( HashMap columnData ){
	if( columnData.containsKey( rxnID ) ){
	    setReactionID( (String)columnData.get( rxnID ) );
	} 
	if( columnData.containsKey( rxnName ) ){
	    setReactionName( (String)columnData.get( rxnName ) );
	} 
	// we don't allow updating of reaction eqs for now
// 	if( columnData.containsKey( rxnEqn ) ){
// 	    setReactionEquation( (ReactionEquation)columnData.get( rxnEqn ) );
//	} 
	if( columnData.containsKey( rxnWTUpBnd ) ){
	    setReactionWildTypeUpperBound( (String)
					  columnData.get( rxnWTUpBnd ) );
	} 
	if( columnData.containsKey( rxnWTLowBnd ) ){
	    setReactionWildTypeLowerBound((String)
					   columnData.get( rxnWTLowBnd ) );
	} 
	if( columnData.containsKey( rxnKO ) ){
	    setReactionKnockout( (Boolean)columnData.get( rxnKO ) );
	} 
	if( columnData.containsKey( rxnObjCoef ) ){
	    setObjectiveCoefficient( (Float)columnData.get( rxnObjCoef ) );
	} 
	
	// Only update prediction data if some exists
	if (existPredictions) {
	    if( columnData.containsKey( rxnWTFBA ) ){
		setWildTypeFBA( (Float)columnData.get( rxnWTFBA ) );
	    } 
	    if( columnData.containsKey( rxnMutFBA ) ){
		setMutantFBA( (Float)columnData.get( rxnMutFBA ) );
	    } 
	    
	    if( columnData.containsKey( rxnMOMA ) ){
		setMOMA( (Float)columnData.get( rxnWTFBA ) );
	    }
	} 
	String annotation = reactionAnnotation.toString();
	
	reaction.setAnnotation(annotation);
    }


    public String getID(){
	return getReactionID();
    }

    public void setID( String id ){
	setReactionID( id );
    }
    public String getReactionID(){
	return reaction.getId();
    }
    public void setReactionID( String sid ) {
	reaction.setId( sid );
    }
    public String getReactionName( ){
	return reaction.getName();
    }
    public void setReactionName( String reactionName ){
	reaction.setName( reactionName );
    }

    public ReactionAnnotation getReactionAnnotation( ){
	return this.reactionAnnotation;
    }

    public ReactionAnnotation computeReactionAnnotation( ){
	ReactionAnnotation myReactionAnnotation;
	boolean isReversible = reaction.getReversible();

	if (reaction.isSetAnnotation()) {
	    String annotationStr = reaction.getAnnotation();
	    myReactionAnnotation = new ReactionAnnotation(annotationStr,
							  isReversible);
	}
	else {
	    myReactionAnnotation = new ReactionAnnotation(isReversible);
	}
	return myReactionAnnotation;
    }

    //public ReactionEquation getReactionEquation( ){
    public String getReactionEquation( ){
	return createReactionEquation(reaction);
    }

    public void setReactionEquation( ReactionEquation rxnEqn ){
	
    }

    // flux limits will be strings for now
    public String getReactionWildTypeLowerBound( ){
	return new String(reactionAnnotation.
			     getLower(ReactionAnnotation.WILD_TYPE));
    }
    
    public void setReactionWildTypeLowerBound( String rxnWTLowBnd ){
	// Currently we should set the MUTANT_FBA an the MUTANT_MOMA 
        //  flux limits to be the same, unless it is a knockout.
	String rxnWTLowBndStr = rxnWTLowBnd.toString();
	reactionAnnotation.setLower(ReactionAnnotation.WILD_TYPE,
				    rxnWTLowBndStr);
	if (!isReactionKnockout().booleanValue()) {
	    reactionAnnotation.setLower(ReactionAnnotation.MUTANT_FBA,
					rxnWTLowBndStr);
	    reactionAnnotation.setLower(ReactionAnnotation.MUTANT_MOMA,	
					rxnWTLowBndStr);
	}
    }

    public String getReactionWildTypeUpperBound( ){
	return new String(reactionAnnotation.
			     getUpper(ReactionAnnotation.WILD_TYPE));
    }

    
    public void setReactionWildTypeUpperBound( String rxnWTUpBnd ){
	// Currently we should set the MUTANT_FBA an the MUTANT_MOMA 
        //  flux limits to be the same, unless it is a knockout.
	String rxnWTUpBndStr = rxnWTUpBnd.toString();
	reactionAnnotation.setUpper(ReactionAnnotation.WILD_TYPE,
				    rxnWTUpBndStr);
	if (!isReactionKnockout().booleanValue()) {
	    reactionAnnotation.setUpper(ReactionAnnotation.MUTANT_FBA,
					rxnWTUpBndStr);
	    reactionAnnotation.setUpper(ReactionAnnotation.MUTANT_MOMA,	
					rxnWTUpBndStr);
	}
    }
        
    public Boolean isReactionKnockout( ){
	String lowLimit = 
	    reactionAnnotation.getLower(ReactionAnnotation.MUTANT_FBA);
	String upLimit = 
	     reactionAnnotation.getUpper(ReactionAnnotation.MUTANT_FBA);
	return new Boolean(equalsZero(lowLimit) &&
			   equalsZero(upLimit));
    }

    
    public void setReactionKnockout( Boolean rxnKO ){
	if (rxnKO.booleanValue()) {
	    //For knockout set the MUTANT_FBA and MUTANT_MOMAflux limits to 0
	    reactionAnnotation.setLower(ReactionAnnotation.MUTANT_FBA, "0");
	    reactionAnnotation.setUpper(ReactionAnnotation.MUTANT_FBA, "0");
	    reactionAnnotation.setLower(ReactionAnnotation.MUTANT_MOMA, "0");
	    reactionAnnotation.setUpper(ReactionAnnotation.MUTANT_MOMA, "0");
	}
	else {
	    // To undo the knockout, set the flux limits to be those of
	    // the wild type
	    reactionAnnotation.
		setLower(ReactionAnnotation.MUTANT_FBA, 
			 reactionAnnotation.
			 getLower(ReactionAnnotation.WILD_TYPE));
	    reactionAnnotation.
		setUpper(ReactionAnnotation.MUTANT_FBA,
			 reactionAnnotation.
			 getUpper(ReactionAnnotation.WILD_TYPE));
	    reactionAnnotation.
		setLower(ReactionAnnotation.MUTANT_MOMA, 
			 reactionAnnotation.
			 getLower(ReactionAnnotation.WILD_TYPE));
	    reactionAnnotation.
		setUpper(ReactionAnnotation.MUTANT_MOMA,
			 reactionAnnotation.
			 getUpper(ReactionAnnotation.WILD_TYPE));
	}
    }
    
    public Float getObjectiveCoefficient( ){
	return new 
	    Float(reactionAnnotation.
		      getObjective(ReactionAnnotation.WILD_TYPE));
    }

    
    public void setObjectiveCoefficient( Float rxnObjCoef ){
	// For now, the MUTANT_FBA will share the wild type objective
	reactionAnnotation.setObjective(ReactionAnnotation.WILD_TYPE,
				    rxnObjCoef.toString());
	reactionAnnotation.setObjective(ReactionAnnotation.MUTANT_FBA,
				    rxnObjCoef.toString());	
    }
    
    public Float getWildTypeFBA( ){
	return new 
	    Float(reactionAnnotation.
		  getPrediction(ReactionAnnotation.WILD_TYPE));
    }

    
    public void setWildTypeFBA( Float rxnWTFBA ){
	// no-op for now
	// reactionAnnotation.setPrediction(ReactionAnnotation.WILD_TYPE,
// 					 rxnWTFBA.toString());
    }
    
    public Float getMutantFBA( ){
	return new 
	    Float(reactionAnnotation.
		  getPrediction(ReactionAnnotation.MUTANT_FBA));
    }

    
    public void setMutantFBA( Float rxnMutFBA ){
	// no-op for now
    }
    
    public Float getMOMA( ){
	return new 
	    Float(reactionAnnotation.
		  getPrediction(ReactionAnnotation.MUTANT_MOMA));
    }

    
    public void setMOMA( Float rxnMOMA ){
	// no-op for now
    }


   /**
     * Retrieves a string representation for this reaction.
     * @return 
     * a string representing the reaction.
     * Based on code in the BioSpreadsheet
     *  written by James M. McCollum <jmccoll2@utk.edu>
     * University of Tennessee - Knoxville
     * http://biocomp.ece.utk.edu

     */
    public String createReactionEquation(Reaction reaction) {
	StringBuffer sbuff= new StringBuffer("");
	for(int i=0; i< reaction.getNumReactants(); i++) {
	    if (i != 0) {
		sbuff.append("+ ");
	    }
	    SpeciesReference specRef = reaction.getReactant(i);
	    String speciesName = specRef.getSpecies();
	    double stochiometry = specRef.getStoichiometry();
	    if (stochiometry != 1.0) {
		sbuff.append(stochiometry + " ");
	    }
	    sbuff.append(speciesName + " ");
	}
	if (reaction.getReversible()) {
	    sbuff.append("<");
	}
	sbuff.append("=> ");
	long productCount = reaction.getNumProducts();
	if (productCount == 0) {
	    sbuff.append("*");
	} else {
	    for(int i=0; i< productCount; i++) {
		if (i != 0) {
		    sbuff.append("+ ");
		}
		SpeciesReference specRef = reaction.getProduct(i);
		String speciesName = specRef.getSpecies();
		double stochiometry = specRef.getStoichiometry();
		if (stochiometry != 1.0) {
		    sbuff.append(stochiometry  + " ");
		}
		sbuff.append(speciesName + " ");
	    }
	}
	return sbuff.toString();
    }
    
    public boolean equalsZero(String value) {
	boolean isZero = false;
	if (null == value) { 
	    return isZero;
	}
	try {
	    Float floatVal = new Float(Float.parseFloat(value));
	    isZero = (0 == floatVal.compareTo(new Float(0.0)) ||
		      0 == floatVal.compareTo(new Float(0.0))); 
	}
	catch (NumberFormatException e) {
	    // if can't parse it, it's not zero
	}
	    return isZero;
    }

    public static void setNoPredictions () {
	existPredictions = false;
	NUM_COLUMNS -= NUM_ANALYSES;
    }

    public String toString(){
	return reaction.toSBML(2, 1);
    }
    static {
	System.loadLibrary("sbmlj");
    }
}
