#ifndef SBW_HOST_ACCESS_TABLE
#define SBW_HOST_ACCESS_TABLE

#include <vector>
#include "SBW/SBWException.h"

namespace SystemsBiologyWorkbench
{
	namespace Broker 
	{
		class HostAccessTable
		{
			/** The list of acceptable addresses. **/
		private:
			std::vector<std::string> acceptAddresses;

			/** The list of explicitly denied addresses.  **/
			std::vector<std::string> denyAddresses;

			/** The single instance of this table, created at class loading time.  **/
			static HostAccessTable *singleton;
			void init(void);
			/**
			* Private constructor that suppresses the default public constructor.
			* This, and the following method for returning an instance of this
			* class ({@link #getInstance()}), implement the "Singleton" design
			* pattern (named in the Design Patterns book by Gamma et al.).  See
			* <a href="http://developer.java.sun.com/developer/qow/archive/111/">
			* http://developer.java.sun.com/developer/qow/archive/111/</a> for
			* more information.
			**/
			HostAccessTable() { }

		public:
			/**
			* Returns the (single) instance of this class.
			**/
			static HostAccessTable *getTable()
			{
				if (HostAccessTable::singleton == NULL)
				{
					HostAccessTable::singleton = new HostAccessTable;
					HostAccessTable::singleton->init();
				}
				return HostAccessTable::singleton;
			}

			/**
			* This code is based on similar code in <code>WebServer.java</code>
			* in the Helma XML-RPC implementation by Hannes Wallnoefer.
			* <p>
			* @param address the address to check for permission
			* @return <code>true</code> if the address is allowed
			**/
			bool checkAddress(std::string address);

			/**
			* Add an IP address to the list of accepted clients.  The parameter can
			* contain '*' as wildcard character, e.g., "192.168.*.*".
			* <p>
			* This code is based on similar code in <code>WebServer.java</code>
			* in the Helma XML-RPC implementation by Hannes Wallnoefer.
			* <p>
			* @param address the address to add to the list of accepted addresses
			* @see denyAddress(java.lang.std::string)
			**/
			void acceptAddress(std::string address);

			/**
			* Add an IP address to the list of denied clients. The parameter can
			* contain '*' as wildcard character, e.g. "192.168.*.*".
			* <p>
			* This code is based on similar code in <code>WebServer.java</code>
			* in the Helma XML-RPC implementation by Hannes Wallnoefer.
			* <p>
			* @param address the address to add to the list of accepted addresses
			* @see acceptAddress(java.lang.std::string)
			**/
			void denyAddress(std::string address);
			/**
			* Class that implements an entry in the accept or deny lists
			* for checking addresses.
			* 
			* This code is based on similar code in <code>WebServer.java</code>
			* in the Helma XML-RPC implementation by Hannes Wallnoefer.
			**/
			class AddressMatcher
			{
				/** The address pattern for this entry. **/
				int *pattern;
			public:
				/** Constructor that parses the address and stores it.  **/
				AddressMatcher(std::string address);
				/** Performs a match against the address stored for this instance. **/
				bool matches(std::string address);
			};

			/**
			* Exception class for representing an error in parsing an IP address.
			**/
			class AddressMatcherException : SystemsBiologyWorkbench::SBWException
			{
			public:
				AddressMatcherException(std::string msg) : SBWException(msg)
				{ }
			};
		};
	}
}
#endif
