#ifndef SBW_BROKER_INSTANCE_THREAD
#define SBW_BROKER_INSTANCE_THREAD

#include "SBW/SBWOSSocket.h"
#include "SBW/DataBlockReader.h"
#include "SBW/SBWException.h"

#include "Instance.h"
#include "InstanceBase.h"
#include "InstanceTable.h"

#include <string>
#include <vector>
#include <iostream>

namespace SystemsBiologyWorkbench
{
	namespace Broker
	{

		class InstanceThread : public InstanceBase ,  public SBWThread
		{
		public:
			/**
			* Interface for defining callbacks used to perform actions at certain
			* points in the life cycle of the instance.  The callbacks are defined
			* in {@link SBWBrokerRPC}.
			**/
			class Hook
			{
			public :
				/**
				* Callback invoked as the last step by the {@link InstanceThread}
				* class constructor.
				* <p>
				* @param instance the instance
				**/
				virtual void onCreation(InstanceThread *instance);

				/**
				* Callback invoked as the first step by the {@link InstanceThread#run}
				* method.
				* <p>
				* @param instance the instance
				**/
				virtual void onStartup(InstanceThread *instance) ;

				/**
				* Callback invoked after the {@link InstanceThread} message loop
				* has stopped being operational.
				* <p>
				* @param instance the instance
				**/
				virtual void onShutdown(InstanceThread *instance) ;
			};
			/**
			* Interface used to implement factory methods for generating new
			* instances of subclasses of {@link InstanceThread}.  This is used
			* in {@link ModuleInstance} and {@link BrokerInstance} to implement
			* the <code>newInstance</code> method that each provides.  The factory
			* methods are used by {@link SBWBrokerRPC} and {@link ConnectionThread}.
			**/
			class Factory
			{
			public:
				/**
				* Returns a new instance of this module/broker instance class.
				* @param socket the socket connecting to the application
				**/
				virtual InstanceThread *newInstance(SOCKET socket);

				/**
				* Returns a string identifying the type of instance, such as
				* "module" or "broker".
				**/
				virtual std::string getType();
			};

			void setException(SBWException *ex)
			{
				exception = ex;
			}

			SBWException *getException()
			{
				return exception;
			}

			/** The thread object for this instance.  **/
			virtual bool connect(std::string key) = 0;
			virtual void transmit(unsigned char *data, int length) = 0;
			virtual void deliverMessage(int destId, Instance *destInstance, unsigned char * message, int length) = 0;
			bool start();
			void run();
			void setThreadName();
			void runMessageLoop();
			void disconnect();
			void disconnectInstance();
			void shutdown();
			void shutdownInstanceThread();
		private:
			SBWException *exception;
			/** Boolean flag indicating whether we believe the client has started.  **/
			bool started;
			/** Boolean flag indicating whether the run loop is operational.  **/
			bool operational;
			/** The list of hooks for this instance.  **/
			std::vector<Hook *> myInstanceHooks;
			/** Cached pointer to the instance table.  **/
			InstanceTable* instanceTable;
			/**
			* Sets the exception for an exception that just occurred, so that
			* a call to {@link #getException()} will retrieve it.  This allows
			* passing exceptions in certain situations that are awkward to do
			* using direct throws.
			**/
			void runCreationHooks();
			/**
			* Run the {@link InstanceThread.Hook#onStartup} hook functions for
			* this instance.
			**/
			void runStartupHooks();
			/**
			* Run the {@link InstanceThread.Hook#onShutdown} hook functions for
			* this instance.
			**/
			void runShutdownHooks();
		protected:
			/** Socket for this instance's connection to its remote module/broker. **/
			SOCKET mySocket;
			/** How long to wait for the instance to start up.  **/
			int startupWaitTime;
			/**
			* Constructor for this class.  This sets up internal variables and
			* sets parameters on the internal thread object, but does not start
			* the thread.  The last step here is calling {@link runCreationHooks}.
			* <p>
			* @param socket the socket connected to the actual module or remote broker
			* @param type a string, either "module" or "broker"
			* @param hook an {@link InstanceThread.Hook} object providing the
			* callbacks to call for this instance.
			**/
			InstanceThread(SOCKET socket, std::string type, Hook *hook);
		};
	}
}
#endif
