#include "BrokerDefines.h"
#include "ModuleRegistryFile.h"
#include "ModuleRegistry.h" 

#include "libxml/xmlreader.h"
#include "libxml/xmlwriter.h"
#include "libxml/parser.h"
#include "libxml/tree.h"

#include "SBW/SBWConfig.h"
#include "SBW/SBWApplicationException.h"

#include <string>
#include <ios>
#include <fstream>

#ifdef WIN32
#include <direct.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif

using namespace std;
using namespace SystemsBiologyWorkbench::Broker;
using namespace SystemsBiologyWorkbench;

ModuleRegistryFile::~ModuleRegistryFile()
{
	xmlCleanupParser();
}
ModuleRegistryFile::ModuleRegistryFile(ModuleRegistry *registry)
{
	LIBXML_TEST_VERSION;
	this->registry = registry;
	this->registryFile = getRegistryFile();
	if (!exists())
	{
		std::string sDir = getPath();
#ifndef WIN32
		if (mkdir(sDir.c_str(), 755) == -1)
#else
		if (mkdir(sDir.c_str()) == -1)
#endif
		{
			int nPosLast = 0;
#ifndef WIN32
			int nPosCurrent = sDir.find("/",nPosLast);
#else
			int nPosCurrent = sDir.find("\\",nPosLast);
#endif
			std::string sTemp;
			while ((unsigned)nPosCurrent != sDir.npos)
			{
				sTemp = sDir.substr(0,nPosCurrent+1);
#ifndef WIN32
				mkdir (sTemp.c_str(), 755);
#else
				mkdir (sTemp.c_str());
#endif
				nPosLast = nPosCurrent+1;
#ifndef WIN32
				nPosCurrent = sDir.find("/",nPosLast);
#else
				nPosCurrent = sDir.find("\\",nPosLast);
#endif
			}
		}
	}
}
bool ModuleRegistryFile::exists()
{	
	FILE *regFile = fopen(registryFile.c_str(),"r");

	// Do some sanity checking on the value before returning it.

	if (regFile != NULL)
	{
		fclose(regFile);
		return true;

	}
	return false;

}
std::string ModuleRegistryFile::getPath()
{
	return Config::getSBWUserDir();
}
std::string ModuleRegistryFile::getRegistryFile()
{
#ifndef WIN32
	std::string sRegistryFile = Config::getSBWUserDir() + "/" + DEFAULT_REGISTRY_FILENAME;
#else
	std::string sRegistryFile = Config::getSBWUserDir() + "\\" + DEFAULT_REGISTRY_FILENAME;
#endif
	return sRegistryFile;
}
void ModuleRegistryFile::checkpoint()
{
	writeRegistryFile();
}
void ModuleRegistryFile::printStats()
{
}
void ModuleRegistryFile::parseNode(xmlNode *node) 
{
	xmlNode *current = NULL;
	bool bFound = false;
	std::string sName = "listOfModules";
	for (current = node->children; current; current = current->next)
	{
		if (sName == (char *)(current->name)) 
		{
			bFound = true;
			break;
		}
	}
	if (!bFound || current == NULL) 
	{
		throw new SBWApplicationException("listOfModules not found");
	}

	for (current = current->children; current; current = current->next)
	{
		if ( current->type == XML_ELEMENT_NODE) 
		{
			std::string moduleName = getAttributeValue(current, "name");
			std::string moduleDisplayName = getAttributeValue(current, "displayName");
			std::string type = getAttributeValue(current, "type");
			std::string cmd = getAttributeValue(current, "startupCmd");
			std::string help = getAttributeValue(current, "help");

			DataBlockWriter oWriter;
			oWriter << moduleName << moduleDisplayName << typeToInt(type) << cmd << help;
			ModuleDescriptor *md = new ModuleDescriptor(oWriter.getReader());
			registry->addModuleDescriptor(moduleName, md, false);

			xmlNode *services = current->children;

			if (services == NULL) 
			{
				TRACE( "Module '" << moduleName << "' has no <listOfServices>")
					continue;
			}
			else 
			{
				if ( services->type != XML_ELEMENT_NODE) 
					services = services->next;
				for (services = services->children; services; services = services->next) 
				{
					if ( services ->type == XML_ELEMENT_NODE) 
					{
						std::string serviceName = getAttributeValue(services, "name");
						std::string serviceDisp = getAttributeValue(services, "displayName");
						std::string serviceCat  = getAttributeValue(services, "category");
						std::string serviceHelp = getAttributeValue(services, "help");

						DataBlockWriter oServiceWriter;
						oServiceWriter << moduleName << serviceName << serviceDisp << serviceCat << serviceHelp;
						ServiceDescriptor *sd = new ServiceDescriptor (oServiceWriter.getReader(), *md);
						md -> addService(sd); 
					}
				}
			}
		}
	}
}
void ModuleRegistryFile::readRegistry()
{
	registryDoc		= NULL;
	registryRoot	= NULL;

	registryDoc = xmlParseFile(registryFile.c_str());
	if (registryDoc != NULL)
	{
		registryRoot = xmlDocGetRootElement(registryDoc);
		parseNode(registryRoot);
	}
	else 
	{
		throw new SBWApplicationException("couldn't read registry.");
	}
}
std::string ModuleRegistryFile::removeQuotes(std::string sString) 
{
	char *newString = (char*) malloc(sizeof(char) * sString.length());
	for ( unsigned int i=0; i < sString.length(); i++) 
	{
		if (sString[i] == '\"')
			newString[i] = '\'';
		else
			newString[i] = sString[i];			
	}
	newString[sString.length()] = '\0';
	std::string sResult (newString);
	free (newString);
	return sResult;
}
void ModuleRegistryFile::writeRegistryFile()
{
	try 
	{
		xmlTextWriterPtr writer;
		xmlChar *tmp;
		xmlDocPtr doc;

		writer = xmlNewTextWriterDoc(&doc, 0);
		xmlTextWriterStartDocument(writer, NULL, "UTF-8", NULL);
		xmlTextWriterStartElement(writer, BAD_CAST "SBWRegistry");
		xmlTextWriterWriteAttribute(writer, BAD_CAST "lastUpdated",
			BAD_CAST "today");
		xmlTextWriterStartElement(writer, BAD_CAST "listOfModules");
		std::vector<ModuleDescriptor *> modules = registry->getModuleDescriptors();
		for (unsigned int i = 0; i < modules.size(); i++) 
		{
			ModuleDescriptor *md = modules[i];
			if (md != NULL && md->getName() != "BROKER") 
			{
				xmlTextWriterStartElement(writer, BAD_CAST "module");
				tmp = ConvertInput(md->getName().c_str(), "UTF-8");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					BAD_CAST tmp);
				xmlFree(tmp);
				tmp = ConvertInput(md->getDisplayName().c_str(), "UTF-8");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "displayName",
					BAD_CAST tmp);
				xmlFree(tmp);
				tmp = ConvertInput((md->getManagementType() == UNIQUE ? "UNIQUE" : "SELF_MANAGED"), "UTF-8");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
					BAD_CAST tmp);
				xmlFree(tmp);
				tmp = ConvertInput(md->getCommandLine().c_str(), "UTF-8");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "startupCmd",
					BAD_CAST tmp);
				xmlFree(tmp);
				tmp = ConvertInput(md->getHelp().c_str(), "UTF-8");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "help",
					BAD_CAST tmp);
				xmlFree(tmp);
				std::vector<ServiceDescriptor*> services = md ->getServices();
				if (services.size() > 0 ) {
					xmlTextWriterStartElement(writer, BAD_CAST "listOfServices");
					for (unsigned int j = 0; j < services.size(); j++) 
					{
						ServiceDescriptor *sd = services[j];
						if (sd != NULL) 
						{
							xmlTextWriterStartElement(writer, BAD_CAST "service");
							tmp = ConvertInput(sd->getName().c_str(), "UTF-8");
							xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
								BAD_CAST tmp);
							xmlFree(tmp);
							tmp = ConvertInput(sd->getDisplayName().c_str(), "UTF-8");
							xmlTextWriterWriteAttribute(writer, BAD_CAST "displayName",
								BAD_CAST tmp);
							xmlFree(tmp);
							tmp = ConvertInput(sd->getCategory().c_str(), "UTF-8");
							xmlTextWriterWriteAttribute(writer, BAD_CAST "category",
								BAD_CAST tmp);
							xmlFree(tmp);
							tmp = ConvertInput(sd->getHelp().c_str(), "UTF-8");
							xmlTextWriterWriteAttribute(writer, BAD_CAST "help",
								BAD_CAST tmp);
							xmlFree(tmp);
							xmlTextWriterEndElement(writer); // end service
						}
					}
					xmlTextWriterEndElement(writer); // end listOfServices
				}
				xmlTextWriterEndElement(writer); // end module
			}
		}
		xmlTextWriterEndElement(writer); // end listOfModules
		xmlTextWriterEndElement(writer); // end SBWRegistry
		xmlTextWriterEndDocument(writer);
		xmlFreeTextWriter(writer);
		xmlSaveFileEnc(registryFile.c_str(), doc, "UTF-8");
		xmlFreeDoc(doc);
	}
	catch (...)	
	{
		throw new SBWApplicationException("couldn't write registry...");
	}	
}
/**
* ConvertInput:
* @in: string in a given encoding
* @encoding: the encoding used
*
* Converts @in into UTF-8 for processing with libxml2 APIs
*
* Returns the converted UTF-8 string, or NULL in case of error.
*/
xmlChar * ModuleRegistryFile::ConvertInput(const char *in, const char *encoding)
{
	xmlChar *out;
	int ret;
	int size;
	int out_size;
	int temp;
	xmlCharEncodingHandlerPtr handler;

	if (in == 0)
		return 0;

	handler = xmlFindCharEncodingHandler(encoding);

	if (!handler) {
		printf("ConvertInput: no encoding handler found for '%s'\n",
			encoding ? encoding : "");
		return 0;
	}

	size = (int) strlen(in) + 1;
	out_size = size * 2 - 1;
	out = (unsigned char *) xmlMalloc((size_t) out_size);

	if (out != 0) {
		temp = size - 1; 
		ret = handler->input(out, &out_size, (const xmlChar *) in, &temp);
		if (ret || temp - size + 1) {
			if (ret) {
				printf("ConvertInput: conversion wasn't successful.\n");
			} else {
				printf
					("ConvertInput: conversion wasn't successful. converted: %i octets.\n",
					temp);
			}

			xmlFree(out);
			out = 0;
		} else {
			out = (unsigned char *) xmlRealloc(out, out_size + 1);
			out[out_size] = 0;  /*null terminating out */
		}
	} else {
		printf("ConvertInput: no mem\n");
	}

	return out;
}
int ModuleRegistryFile::typeToInt(std::string type)
{
	if (type == "UNIQUE") 
	{
		return UNIQUE;
	} 
	else if (type == "SELF_MANAGED")
	{
		return SELF_MANAGED;
	}
	else
		throw new SBWApplicationException("Invalid module management type found ...");
}
std::string ModuleRegistryFile::getAttributeValue(xmlNode *elem, std::string attrName)
{
	char *result;
	result = (char*) xmlGetProp(elem, (xmlChar *)attrName.c_str());
	if (result == NULL)
		return "";
	std::string sResult (result);
	xmlFree(result);
	return sResult;
}
ModuleRegistry *ModuleRegistryFile::registry = NULL;
std::string ModuleRegistryFile::registryFile = "";
xmlNode *ModuleRegistryFile::registryRoot = NULL;
xmlDoc *ModuleRegistryFile::registryDoc = NULL;
xmlNode *ModuleRegistryFile::registryListOfModules = NULL;
