/**
 * @file SBWOSSocket.cpp
 * @brief platform independant interface to sockets
 * @author SBW Development Group <sysbio-team@caltech.edu>
 *
 * Organization: Caltech ERATO Kitano Systems Biology Project
 *
 * Created: @date 18th June 2001
 * $Id: SBWOSSocket.cpp,v 1.3 2004/09/23 22:18:59 wrindone Exp $
 * $Source: /var/lib/cvs/Biospice/FluxorPipeline/Metatool/src/C++Broker/SBWCore/SBWOSSocket.cpp,v $
 */

/*
** Copyright 2001 California Institute of Technology and
** Japan Science and Technology Corporation.
** 
** This library is free software; you can redistribute it and/or modify it
** under the terms of the GNU Lesser General Public License as published
** by the Free Software Foundation; either version 2.1 of the License, or
** any later version.
** 
** This library is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
** MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
** documentation provided hereunder is on an "as is" basis, and the
** California Institute of Technology and Japan Science and Technology
** Corporation have no obligations to provide maintenance, support,
** updates, enhancements or modifications.  In no event shall the
** California Institute of Technology or the Japan Science and Technology
** Corporation be liable to any party for direct, indirect, special,
** incidental or consequential damages, including lost profits, arising
** out of the use of this software and its documentation, even if the
** California Institute of Technology and/or Japan Science and Technology
** Corporation have been advised of the possibility of such damage.  See
** the GNU Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this library; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
**
** The original code contained here was initially developed by:
**
**     Andrew Finney, Herbert Sauro, Michael Hucka, Hamid Bolouri
**     The Systems Biology Workbench Development Group
**     ERATO Kitano Systems Biology Project
**     Control and Dynamical Systems, MC 107-81
**     California Institute of Technology
**     Pasadena, CA, 91125, USA
**
**     http://www.cds.caltech.edu/erato
**     mailto:sysbio-team@caltech.edu
**
** Contributor(s):
**
*/

// SBWOSSocket.cpp: implementation of the SBWOSSocket class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "config.h"
#include "SBWRawException.h"
#include <string.h>

#if defined(WIN32)
#include <winsock2.h>
#else
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

// FIXME what is this doing?
#include <tcpd.h>
#include <unistd.h>
#endif
#include "SBWOSSocket.h"

using namespace SystemsBiologyWorkbench ;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
/**
 * WIN32 specific - initializes the WIN32 WSA system
 */
SBWOSSocketSystem::SBWOSSocketSystem()
{
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 3, 0 );

	err = WSAStartup( wVersionRequested, &wsaData );
	assert ( err == 0 );
}

/**
 * WIN32 specific - shutdown WSA system
 */
SBWOSSocketSystem::~SBWOSSocketSystem()
{
	WSACleanup();
}

/// WIN32 specific - manages the WSA system
SBWOSSocketSystem SBWOSSocket::system;
#endif // WIN32

/**
 * Create a socket and connect to the given port on this host.
 * @param port server port on hostname to connect to.
 * @param hostname machine to connect to, "" implies local.
 */
SBWOSSocket::SBWOSSocket(Integer port, const char *hostname)
{
	if (hostname == "")
	{
		TRACE("Opening socket to localhost");
	}
	else
	{
		TRACE("Opening socket to host named '" << hostname << "'");
	}

	sock = socket(AF_INET, SOCK_STREAM, 0);

	if (sock == INVALID_SOCKET)
	{
	    TRACE("Could not create socket");
		throwError();
	}

	sockaddr_in address ;

	address.sin_family = AF_INET;// short
	address.sin_port = htons(port) ; // u_short

#if defined(WIN32)
	unsigned long hostAddress ;
#else
	in_addr_t hostAddress ;
#endif

	if (hostname == "" || strlen(hostname) == 0
		|| strcmp(hostname, "localhost") == 0)
	{
		TRACE("Host address = 127.0.0.1");
		hostAddress = inet_addr("127.0.0.1");
	}
	else
	{
		TRACE("Trying to obtain address for host '" << hostname << "'");
		std::string given(hostname);
		std::string::size_type first = given.find_first_not_of("1234567890.");

		if (first == std::string::npos)
		{
			hostAddress = inet_addr(hostname) ;
		}
		else
		{
			struct hostent *hptr = gethostbyname(hostname);

			if (hptr == NULL)
			{
				TRACE("Failed to get host entry from gethostbyname");
				throwError();
			}

			memcpy(&hostAddress, hptr->h_addr_list[0], sizeof(hostAddress));
		}
	}

#if defined(WIN32)
	address.sin_addr.S_un.S_addr = hostAddress; 
	if (address.sin_addr.S_un.S_addr == INADDR_NONE)
		throw new SBWSocketException("address error");
#else
	address.sin_addr.s_addr = hostAddress; 
	if (address.sin_addr.s_addr == INADDR_NONE)
		throw new SBWSocketException("address error");
#endif

	TRACE("Connecting socket");
	if (connect(sock, (sockaddr *)&address, sizeof(address)) == SOCKET_ERROR)
		throwError();
}

/**
 * close/disconnect socket
 */
SBWOSSocket::~SBWOSSocket()
{
	TRACE("Destroying socket object");
#ifdef WIN32
	WSAEVENT event = WSACreateEvent();

	if (event == WSA_INVALID_EVENT)
		throwError();

	if (WSAEventSelect(sock, event, FD_CLOSE) == SOCKET_ERROR)
		throwError();
#endif

	if (shutdown(sock, SD_SEND) == SOCKET_ERROR)
		throwError();

#ifdef WIN32
	if (WSAWaitForMultipleEvents(1, &event, TRUE, WSA_INFINITE, FALSE)
			== WSA_WAIT_FAILED)
	{
		throwError();
	}
#endif

	if (closesocket(sock) == SOCKET_ERROR)
		throwError();
}

/**
 * receive data from socket.
 * Blocks until all request bytes have been received or disconnection occurs.
 * @param data buffer to receive data
 * @param length size of buffer and number of bytes requested.
 */
void SBWOSSocket::receive(unsigned char *data, Integer length)
{
	Integer cursor = 0 ;

	while (cursor != length)
	{
		Integer result = recv(sock, (char *)data + cursor, length - cursor, 0);

		if (result == 0)
			throw new SBWDisconnectException("graceful disconnect while reading");

		if (result == SOCKET_ERROR)
			throwError();

		cursor += result ;
	}
}

/**
 * transmit data to socket.
 * Blocks until all bytes have been transmitted or disconnection occurs.
 * @param data data to be transmitted.
 * @param length number of bytes to be transmitted.
 */
void SBWOSSocket::transmit(unsigned char *data, Integer length)
{
	Integer cursor = 0;

	while (cursor != length)
	{
		Integer result = send(sock, (char *)data + cursor, length - cursor, 0);
		
		if (result == SOCKET_ERROR)
			throwError();

		cursor += result ;
	}
}

/**
 * throws an SBWException * corresponding to the current error condition of this socket.
 * @throw an SBWException * corresponding to the current error condition of this socket.
 */
void SBWOSSocket::throwError()
{
#if defined(WIN32)
	int code = WSAGetLastError();
	
	switch (code)
	{
		case WSANOTINITIALISED :
			throw new SBWSocketException("A successful WSAStartup must occur before using this function.");

		case WSAENETDOWN :
			throw new SBWDisconnectException("The network subsystem or the associated service provider has failed.");

		case WSAEAFNOSUPPORT :
			throw new SBWSocketException("The specified address family is not supported.");
			
		case WSAEINPROGRESS :
			throw new SBWSocketException("A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function.");

		case WSAEMFILE :
			throw new SBWSocketException("No more socket descriptors are available.");
			
		case WSAENOBUFS :
			throw new SBWSocketException("No buffer space is available. The socket cannot be created.");
			
		case WSAEPROTONOSUPPORT :
			throw new SBWSocketException("The specified protocol is not supported.");
			
		case WSAEPROTOTYPE :
			throw new SBWSocketException("The specified protocol is the wrong type for this socket.");
			
		case WSAESOCKTNOSUPPORT :
			throw new SBWSocketException("The specified socket type is not supported in this address family.");
			
		case WSAEADDRINUSE :
			throw new SBWSocketException("The socket's local address is already in use and the socket was not marked to allow address reuse with SO_REUSEADDR. This error usually occurs when executing bind, but could be delayed until this function if the bind was to a partially wild-card address (involving ADDR_ANY) and if a specific address needs to be committed at the time of this function."); 

		case WSAEINTR :
			throw new SBWSocketException("The (blocking) Windows Socket 1.1 call was canceled through WSACancelBlockingCall."); 

		case WSAEALREADY :
			throw new SBWSocketException("A nonblocking connect call is in progress on the specified socket.");

		case WSAEADDRNOTAVAIL :
			throw new SBWSocketException("The remote address is not a valid address (such as ADDR_ANY).");
			
		case WSAECONNREFUSED :
			throw new SBWConnectException("The attempt to connect was forcefully rejected.");
			
		case WSAEFAULT :
			throw new SBWSocketException("The name or the namelen parameter is not a valid part of the user address space, the namelen parameter is too small, or the name parameter contains incorrect address format for the associated address family."); 

		case WSAEINVAL :
			throw new SBWSocketException("The parameter s is a listening socket, or the destination address specified is not consistent with that of the constrained group the socket belongs to."); 

		case WSAEISCONN :
			throw new SBWSocketException("The socket is already connected (connection-oriented sockets only).");
			
		case WSAENETUNREACH :
			throw new SBWDisconnectException("The network cannot be reached from this host at this time.");
			
		case WSAENOTSOCK :
			throw new SBWSocketException("The descriptor is not a socket.");
			
		case WSAETIMEDOUT :
			throw new SBWConnectException("Attempt to connect timed out without establishing a connection.");
			
		case WSAEWOULDBLOCK :
			throw new SBWSocketException("The socket is marked as nonblocking and the connection cannot be completed immediately."); 

		case WSAEACCES :
			throw new SBWSocketException("Attempt to connect datagram socket to broadcast address failed because setsockopt option SO_BROADCAST is not enabled."); 

		case WSAENOTCONN :
			throw new SBWSocketException("The socket is not connected.");
			
		case WSAENETRESET :
			throw new SBWDisconnectException("The connection has been broken due to the keep-alive activity detecting a failure while the operation was in progress.");

		case WSAEOPNOTSUPP :
			throw new SBWSocketException("MSG_OOB was specified, but the socket is not stream-style such as type SOCK_STREAM, out-of-band data is not supported in the communication domain associated with this socket, or the socket is unidirectional and supports only send operations.");
			
		case WSAESHUTDOWN :
			throw new SBWDisconnectException("The socket has been shut down; it is not possible to recv on a socket after shutdown has been invoked with how set to SD_RECEIVE or SD_BOTH."); 

		case WSAEMSGSIZE :
			throw new SBWSocketException("The message was too large to fit into the specified buffer and was truncated.");
			
		case WSAECONNABORTED :
			throw new SBWDisconnectException("The virtual circuit was terminated due to a time-out or other failure. The application should close the socket as it is no longer usable.");
			
		case WSAECONNRESET :
			throw new SBWDisconnectException("The virtual circuit was reset by the remote side executing a \"hard\" or \"abortive\" close. The application should close the socket as it is no longer usable. On a UDP datagram socket this error would indicate that a previous send operation resulted in an ICMP \"Port Unreachable\" message."); 

		case WSAEHOSTUNREACH :
			throw new SBWDisconnectException("The remote host cannot be reached from this host at this time."); 

		case WSAHOST_NOT_FOUND :
			throw new SBWConnectException("Authoritative answer host not found");

		case WSATRY_AGAIN :
			throw new SBWConnectException("Nonauthoritative host not found, or server failure");
		
		case WSANO_RECOVERY :
			throw new SBWConnectException("A nonrecoverable error occurred");

		case WSANO_DATA :
			throw new SBWConnectException("Valid name, no data record of requested type");


		default:
			throw new SBWSocketException("Unknown code");

	}

#elif defined(HAVE_ERRNO_H)

	switch (errno)
	{
    	case EPROTONOSUPPORT:
    	case EAFNOSUPPORT:
    	case EPERM:
    	case ENFILE:
    	case EMFILE:
    	case ENOBUFS:
    	case ENOMEM:
    	case EINVAL:
    	case EACCES:
    	case EFAULT:
    	case ENOTCONN:
    	case ENOTSOCK:
    	case EBADF:
    	case EADDRINUSE:
    		throw new SBWSocketException(strerror(errno));

    	case ECONNREFUSED:
    	case EISCONN:
    	case ETIMEDOUT:
    	case EINPROGRESS:
    	case ENETUNREACH:
    	case EALREADY:
    	case EAGAIN:
    		throw new SBWConnectException(strerror(errno));

		default:
			throw new SBWSocketException("Unknown code");

	}

#endif
}

bool SBWOSSocket::isLocalAddress(const char *hostname)
{
	return strlen(hostname) == 0 || strcmp(hostname, "localhost") == 0
			|| strcmp(hostname, "127.0.0.1") == 0;
}

bool SBWOSSocket::portInUse(int port)
{
    int sock = socket(AF_INET, SOCK_STREAM, 0);
    bool inUse = false;

    if (sock > 0)
    {
		struct sockaddr_in addr;
		
		addr.sin_family = AF_INET;
		addr.sin_port = htons(port);

#if defined(WIN32)
        addr.sin_addr.S_un.S_addr = inet_addr("127.0.0.1");
		if (addr.sin_addr.S_un.S_addr == INADDR_NONE)
			throw new SBWSocketException("address error");
#else
        addr.sin_addr.s_addr = inet_addr("127.0.0.1");
		if (addr.sin_addr.s_addr == INADDR_NONE)
			throw new SBWSocketException("address error");
#endif

		// A failure to bind to the socket indicates it's in use (or unusable),
		// but a successful bind is not enough to say that it's NOT in use.

		int bindCode = bind(sock, (struct sockaddr *) &addr, sizeof(addr));

#if defined(WIN32)
		if (bindCode != 0 && WSAGetLastError() == WSAEADDRINUSE)
#elif defined(HAVE_ERRNO_H)
		if (bindCode != 0 && errno == EADDRINUSE)
#endif
		{
			TRACE("Port " << port << " appears to be in use");
			inUse = true;
		}
		else
		{
			// Binding was successful.  Now try listening on the port.
			// If we can listen, then it's not in use.

			int listenCode = listen(sock, 1);
#if defined(WIN32)
			if (listenCode != 0 && WSAGetLastError() == WSAEADDRINUSE)
#elif defined(HAVE_ERRNO_H)
			if (listenCode != 0 && errno == EADDRINUSE)
#endif
			{
				TRACE("Port " << port << " appears to be in use");
				inUse = true;
			}
		}

		closesocket(sock);
		return inUse;
    }

    return inUse;
}
