/**
  * Filename    : Metatool.cpp
  * Description : SBW Interface to Metatool
  * Author(s)   : Sri Paladugu <spaladug@kgi.edu>
  * Organization: Keck Graduate Institute (KGI)
  * Created     :
  * Revision    : $Id: Metatool.cpp,v 1.2 2004/09/23 18:22:59 wrindone Exp $
  * Source      : $Source: /var/lib/cvs/Biospice/FluxorPipeline/Metatool/src/Metatool/src/Metatool.cpp,v $
  *
  * Copyright 2004 Keck Graduate Institute
  *
  * This library is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
  * by the Free Software Foundation; either version 2.1 of the License, or
  * any later version.
  *
  * This library is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
  * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
  * documentation provided hereunder is on an "as is" basis, and the
  * California Institute of Technology and Japan Science and Technology
  * Corporation have no obligations to provide maintenance, support,
  * updates, enhancements or modifications.  In no event shall the
  * California Institute of Technology or the Japan Science and Technology
  * Corporation be liable to any party for direct, indirect, special,
  * incidental or consequential damages, including lost profits, arising
  * out of the use of this software and its documentation, even if the
  * California Institute of Technology and/or Japan Science and Technology
  * Corporation have been advised of the possibility of such damage.  See
  * the GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this library; if not, write to the Free Software Foundation,
  * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
  *
  * The original code contained here was initially developed by:
  *
  *     Sri Paladugu
  *     The Systems Biology Markup Language Development Group
  *     ERATO Kitano Symbiotic Systems Project
  *     Keck Graduate Institute of Applied Life Sciences
  *     535 Watson Drive
  *     Claremont, CA, 91711, USA
  *
  *     http://www.cds.caltech.edu/erato
  *     http://www.kgi.edu
  *     mailto:sysbio-team@caltech.edu
  * Contributor(s):
  */

#if defined(WIN32)
#include <windows.h>
#elif defined(linux)
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#endif
#include <iostream>
#include <sstream>
#include <fstream>
#include <list>
#include <algorithm>
#include "SBW.h"
#include "DataBlockType.h"
using namespace SystemsBiologyWorkbench ;
using namespace std;
#define CHECK_NOM \
	{\
		if(nomModule == -1000)\
		{\
			nomModule = SBW::getModuleInstance("edu.caltech.NOMClipboard");\
			nomService = nomModule.findServiceByName("NOM");\
		}\
	}
const string METATOOLSBW_VERSION  = "Wrapper C++ Version 1.0; Metatool by Stefan Schuster et al";
const string INPUT_METATOOL_FILE  = "m$$in.txt";
const string OUTPUT_METATOOL_FILE = "m$$out.txt";
#if defined(WIN32)
const string metaToolExeName      = "meta4.3_double.exe";
#elif defined(linux)
const string metaToolExeName      = "meta4.3_double";
#endif
const string marker1  = "INTERNAL METABOLITES:";
const string marker2  = "EXTERNAL METABOLITES:";
const string marker3  = "REACTIONS:";
const string marker4  = "STOICHIOMETRIC MATRIX";
const string marker5  = " matrix dimension";
const string marker6  = "KERNEL";
const string marker7  = "SUBSETS OF REACTIONS";
const string marker8  = "CONVEX BASIS";
const string marker9  = "CONSERVATION RELATIONS";
const string marker10 = "ELEMENTARY MODES";
const string marker11 = "-ENZREV";
const string marker12 = "-ENZIRREV";
const string marker13 = "-METINT";
const string marker14 = "-METEXT";
const int MAXLENGTH   = 512; // Hope this is long enough !

/*
* Metatool
* this class provides an implementation of the mt service
*/
class Metatool
{
private:
	int nInternalSpecies;
	int nExternalSpecies;
	int nReactions;
	vector < vector <double> > sm;
	vector < vector <double> > em;
	vector < vector <double> > cv;
	vector < vector <double> > ns;
	char metaToolPath[MAXLENGTH];
	string metaToolInputFile;
	Module nomModule;
	Service nomService;

	///This Function checks whether a file exists or not.
	///Tries to open the file in read mode. This function can
	///fail - if the file exits and is not readable.
	bool IsFileExist(const string  &name)
	{
         ifstream  in(name.c_str(), ios::in|ios::binary);
            return in.is_open();
    	}

	/// Build a MetaTool Script File from the model currently held in the NOM
	/// Returns the path and name of the input file that was created
	/// NOM = Network Object Model
	string MakeMetaToolFile()
	{
		string str;
		string name;
		string result;
#if defined(WIN32)
		str = AppPath + "\\" + INPUT_METATOOL_FILE;
#elif defined(linux)
		str = AppPath + "/" + INPUT_METATOOL_FILE;
#endif
		result = str;
		ofstream outstream(str.c_str());
		if(!outstream.is_open())
		{
			//raise Exception
	       	SBWApplicationException *ae;
			ae = new SBWApplicationException("Unable to write to file",
				"Exception raised in Module: Metatool, Method: MakeMetaToolFile");
		}

		CHECK_NOM; //Check if NOMClipboard Module is running or not
		ServiceMethod numReactionsMethod = nomService.getMethod("int getNumReactions()");
		ServiceMethod isReversibleMethod = nomService.getMethod("boolean isReactionReversible(int)");
		ServiceMethod reactionNameMethod = nomService.getMethod("string getNthReactionName(int)");
		DataBlockWriter OWrite;
		int numReactions;
		bool isReactionReversible;
		numReactionsMethod.call(OWrite) >> numReactions;
		outstream << "-ENZREV" << endl;
		for(int i=0; i<numReactions; i++)
		{
			isReversibleMethod.call(DataBlockWriter() << i ) >> isReactionReversible;
			if(isReactionReversible) // if true
			{
				reactionNameMethod.call (DataBlockWriter() << i) >> name;
				outstream << name + " ";
			}
		}
		outstream << endl;
		outstream << endl;

		outstream << "-ENZIRREV" << endl;
		for(int i=0; i<numReactions; i++)
		{
			isReversibleMethod.call(DataBlockWriter() << i ) >> isReactionReversible;
			if(!isReactionReversible) // if false
			{
				reactionNameMethod.call (DataBlockWriter() << i) >> name;
				outstream << name + " ";
			}
		}
		outstream << endl;
		outstream << endl;

		//-----------------------------------------------------------------------
		//-METINT and -METEXT should contain only those species that appear
		//in either reactant or product blocks of the reactions.
		//In other words Modifier species should not be listed under these
		//two sections. So create a list of strings containing all the reactants
		//and products in each and every reaction of the model.
		//----------------------------------------------------------------------
		ServiceMethod reactantNameMethod          = nomService.getMethod("string getNthReactantName(int, int)");
		ServiceMethod numReactantsMethod          = nomService.getMethod("int getNumReactants(int)");
		ServiceMethod productNameMethod           = nomService.getMethod("string getNthProductName(int, int)");
		ServiceMethod numProductsMethod           = nomService.getMethod("int getNumProducts(int)");

		int    numReactants;
		int    numProducts;
		string reactantName;
		string productName;
		
		list<string>           reactantsProductsList; //Holds the list of both reactants and products.
		list<string>::iterator reactantsProductsList_iter;
		//Now let us populate the list.
		for(int i = 0; i < numReactions; i++)
		{
			numReactantsMethod.call(DataBlockWriter() << i) >> numReactants;
			for(int j = 0; j < numReactants; j++)
			{
				reactantNameMethod.call( DataBlockWriter() << i << j ) >> reactantName;
				reactantsProductsList.push_back(reactantName);
			}

			numProductsMethod.call(DataBlockWriter() << i) >> numProducts;
			for(int j = 0; j < numProducts; j++)
			{
				productNameMethod.call( DataBlockWriter() << i << j ) >> productName;
				reactantsProductsList.push_back(productName);
			}
		}


		ServiceMethod numFloatingSpeciesMethod  = nomService.getMethod("int getNumFloatingSpecies()");
		ServiceMethod floatingSpeciesNameMethod = nomService.getMethod("string getNthFloatingSpeciesName(int)");
		int    numFloatingSpecies;
		string floatingSpeciesName;

		numFloatingSpeciesMethod.call(OWrite) >> numFloatingSpecies;
		outstream << "-METINT" << endl;
		for(int i=0; i < numFloatingSpecies; i++)
		{
			floatingSpeciesNameMethod.call(DataBlockWriter() << i) >> floatingSpeciesName;
			reactantsProductsList_iter       = find(reactantsProductsList.begin(), reactantsProductsList.end(), floatingSpeciesName);
			if ( reactantsProductsList_iter != reactantsProductsList.end() )
			{
				outstream << floatingSpeciesName;
				outstream << " ";
			}
		}
		outstream << endl;
		outstream << endl;

		ServiceMethod numBoundarySpeciesMethod  = nomService.getMethod("int getNumBoundarySpecies()");
		ServiceMethod boundarySpeciesNameMethod = nomService.getMethod("string getNthBoundarySpeciesName(int)");
		int    numBoundarySpecies;
		string boundarySpeciesName;

		numBoundarySpeciesMethod.call(OWrite) >> numBoundarySpecies;
		outstream << "-METEXT" << endl;
		for(int i=0; i < numBoundarySpecies; i++)
		{
			boundarySpeciesNameMethod.call(DataBlockWriter() << i) >> boundarySpeciesName;
			reactantsProductsList_iter       = find(reactantsProductsList.begin(), reactantsProductsList.end(), boundarySpeciesName);
			if ( reactantsProductsList_iter != reactantsProductsList.end() )
			{
				outstream << boundarySpeciesName;
				outstream << " ";
			}
		}

		//--------------------------------------------------------------------
		//Don't write the CAT section until you write source and sink species.
		//So append the data to a string and write it at the end after done
		//writing the source and sink species names. This should account for
		//empty reactant and product blocks that may appear in level 2 SBML.
		//-------------------------------------------------------------------
		const string sourceString      = "src";
		const string sinkString        = "sink";
		const string NL			       = "\n";
		bool         isAddDummySource  = false;
		bool         isAddDummySink    = false;
		string strCATsection; //string that contains the details that should be written in -CAT Section of metatool
		ServiceMethod reactantStoichiometryMethod = nomService.getMethod("int getNthReactantStoichiometry(int, int)");
		ServiceMethod productStoichiometryMethod  = nomService.getMethod("int getNthProductStoichiometry(int, int)");
		int    reactantStoichiometry;		
		int    productStoichiometry;		
		string reactionName;
		char   buffer[20];
		string strBuffer;

		numReactants  = 0;
		numProducts   = 0;
		reactantName  = "";
		productName   = "";
		strCATsection = "";
		strCATsection = strCATsection + NL;
		strCATsection = strCATsection + NL;
		strCATsection = strCATsection + "-CAT" + NL;

		for(int i=0; i<numReactions; i++)
		{
			reactionNameMethod.call (DataBlockWriter() << i) >> reactionName;
		    strCATsection = strCATsection + reactionName + " : ";
			
			numReactantsMethod.call(DataBlockWriter() << i) >> numReactants;
			//If there are zero reactants then add a dummy source string.
			if (numReactants == 0)
			{
				isAddDummySource = true;
				strCATsection    = strCATsection + sourceString;
			}
			else // Write the first reactant in the reaction
			{
				reactantStoichiometryMethod.call( DataBlockWriter() << i << 0 ) >> reactantStoichiometry;
				if (abs(reactantStoichiometry) != 1)
				{
					sprintf( buffer, "%d", abs(reactantStoichiometry) );
					strBuffer = buffer;
					strCATsection = strCATsection + strBuffer + " ";
				}

				reactantNameMethod.call( DataBlockWriter() << i << 0 ) >> reactantName;
				strCATsection = strCATsection + reactantName;
			}

			//Now start writing remaining reactants of the reaction
			for(int j = 1; j<numReactants; j++)
			{
        		strCATsection = strCATsection + " + ";
				reactantStoichiometryMethod.call( DataBlockWriter() << i << j ) >> reactantStoichiometry;
				if (abs(reactantStoichiometry) != 1)
				{
					sprintf( buffer, "%d", abs(reactantStoichiometry) );
					strBuffer = buffer;
					strCATsection = strCATsection + " " + strBuffer + " ";
				}
				reactantNameMethod.call( DataBlockWriter() << i << j ) >> reactantName;
				strCATsection = strCATsection + reactantName;
			}

			strCATsection = strCATsection + " = ";

			numProductsMethod.call(DataBlockWriter() << i) >> numProducts;
			//If there are zero product then add a dummy sink string.
			if (numProducts == 0)
			{
				isAddDummySink = true;
				strCATsection  = strCATsection + sinkString;
			}
			else // Write the first product of the reaction
			{
				productStoichiometryMethod.call( DataBlockWriter() << i << 0 ) >> productStoichiometry;
				if (abs(productStoichiometry)!= 1)
				{
					sprintf( buffer, "%d", abs(productStoichiometry) );
					strBuffer     = buffer;
					strCATsection = strCATsection + strBuffer + " ";
				}

				productNameMethod.call( DataBlockWriter() << i << 0 ) >> productName;
				strCATsection = strCATsection + productName;
			}

			//Now start writing remaining products of the reaction
			for(int j = 1; j<numProducts; j++)
			{
				strCATsection = strCATsection + " + ";
				productStoichiometryMethod.call( DataBlockWriter() << i << j ) >> productStoichiometry;
				if (abs(productStoichiometry) != 1)
				{
					sprintf( buffer, "%d", abs(productStoichiometry) );
					strBuffer     = buffer;
					strCATsection = strCATsection + " " + strBuffer + " ";
				}
				productNameMethod.call( DataBlockWriter() << i << j ) >> productName;
				strCATsection = strCATsection + productName;
			}

			strCATsection = strCATsection + " . ";
		}

		if (isAddDummySource)
		{
			outstream << sourceString;
			outstream << " ";
		}

		if (isAddDummySink)
		{
			outstream << sinkString;
			outstream << " ";
		}

		outstream << strCATsection;
		outstream.close();
		return result;
	}

	/// Run Metatool with the given metatool input file and wait for execution to
	/// complete. If the input file name is empty then the routine uses
	/// the default input metatool file name
	void runMetatool(string metaToolInputFile)
	{
		string ExecuteStr;
		if(IsFileExist(metaToolPath))
		{
#if defined(WIN32)
			string MetaToolPath = metaToolPath;
			// Now build the execution string
			ExecuteStr = MetaToolPath + ' ' + '"' + metaToolInputFile + '"' + ' ' + OUTPUT_METATOOL_FILE;

			if (ExecuteStr.length() > MAXLENGTH)
			{
				SBWApplicationException* ae = new SBWApplicationException("Overflow error in constructing MetaTool execution string",
					"Exception raised in Module:Metatool, Method:runMetatool");
				throw ae;
			}

			char lpExecute[MAXLENGTH];
			char lpCurrentDirectory[MAXLENGTH];
			strcpy(lpCurrentDirectory, AppPath.c_str());
			strcpy(lpExecute, ExecuteStr.c_str());
			STARTUPINFO si;
			PROCESS_INFORMATION pi;
			ZeroMemory( &si, sizeof(si) );
			si.cb = sizeof(si);
			si.dwFlags = STARTF_USESHOWWINDOW;
			//hide the child:
			si.wShowWindow = SW_HIDE;
			ZeroMemory( &pi, sizeof(pi) );
			if(!CreateProcess( NULL, // No module name (use command line).
				lpExecute,
				NULL,             // Process handle not inheritable.
				NULL,             // Thread handle not inheritable.
				FALSE,            // Set handle inheritance to FALSE.
				0,                // No creation flags.
				NULL,             // Use parent's environment block.
				lpCurrentDirectory,          // Use parent's starting directory.
				&si,              // Pointer to STARTUPINFO structure.
				&pi )             // Pointer to PROCESS_INFORMATION structure.
				)
			{
				// Needs to throw exception
				SBWApplicationException* ae = new SBWApplicationException("Unable to Create Process",
					"Exception raised in Module:Metatool, Method:runMetatool");
				throw ae;
			}

			WaitForSingleObject(pi.hProcess, INFINITE);
			DWORD dwExitCode;
			GetExitCodeProcess(pi.hProcess, &dwExitCode);

			if(dwExitCode == 0)
			{
				// Process has finished Successfully. Now we should close it.
				CloseHandle(pi.hProcess);
				CloseHandle(pi.hThread);
			}
			else
			{
				// Process has finished. Now we should close it.
				CloseHandle(pi.hProcess);
				CloseHandle(pi.hThread);
				SBWApplicationException* ae;
				string outputfile;
				outputfile = AppPath + "\\" + OUTPUT_METATOOL_FILE;
		        ofstream outstream(outputfile.c_str());
				outstream << "ERROR OCCURED IN EXECUTION OF METATOOL" << endl;
				ae = new SBWApplicationException("Error in Execution of Metatool Program",
					"Exception raised in Module:Metatool, Method:runMetatool");
				throw ae;
			}

#elif defined(linux)
			pid_t cpid;
			int status;
			SBWApplicationException* ae;
			string outputfile;
			outputfile = AppPath + "/" + OUTPUT_METATOOL_FILE;

			cpid = fork();
			if (cpid == -1)
			{
				ae = new SBWApplicationException("Unable to Create Process",
					"Exception raised in Module:Metatool, Method:runMetatool");
				throw ae;
			}
			else if( cpid == 0)
			{
				execlp(metaToolPath, metaToolExeName.c_str(), metaToolInputFile.c_str(), outputfile.c_str(), (char*) 0);
				perror("Error Occured \n");
				exit(1);
			}
		        else
			{
				fflush(stdout);
				while(wait(&status) != cpid)
					continue;
				if (WEXITSTATUS(status) == 0)
				{
					//Child Process is completed.
				}
				else
				{
		            ofstream outstream(outputfile.c_str());
					outstream << "ERROR OCCURED IN EXECUTION OF METATOOL" << endl;
					ae = new SBWApplicationException("Error in Execution of Metatool Program",
						"Exception raised in Module:Metatool, Method:runMetatool");
					throw ae;
				}

			}
#endif
		}
		else
		{
			//throw exception unable to locate file at specified path.
			SBWApplicationException* ae = new SBWApplicationException("Unable to locate Metatool executable at specified path. Does the file "
				+ metaToolExeName + " exist?",
				"Exception raised in Module:Metatool, Method:runMetatool");
			throw ae;
		}

	}

	///
	///This Function extracts the matrix based on the input matrixIdentifier
	///If the marker "matrix Identifier" is not found - Then the function returns -1
	///else it returns 0.
	int extractMatrix (ifstream& in_stream, string matrixIdentifier)
	{
		string str, rowsStr, colsStr, rows, cols;
		string buffer;
		vector<string> tList;   // Create vector to hold our tokens
		int nrows, ncols;
		getline(in_stream, str);
		getline(in_stream, str);

		if( str.substr(0,marker5.length() ) != marker5)
		{
			if (matrixIdentifier == "sm")
			{
				sm.clear();
			}

			if (matrixIdentifier == "ns")
			{
				ns.clear();
			}

			if(matrixIdentifier == "cv")
			{
				cv.clear();
			}

			if(matrixIdentifier == "em")
			{
				em.clear();
			}

			return -1; //Failure to extract matrix. The marker "marker identifier" is not found.
		}

		str = str.substr(marker5.length()+1, str.length());
		stringstream ss(str);	// Insert the string into a stream
		try
		{
		    while (ss >> buffer)
			{
				tList.push_back(buffer);
			}
			rowsStr = tList[0];
			colsStr = tList[2];
			rows	= rowsStr.substr(1,rowsStr.length());
			cols	= colsStr.substr(1,colsStr.length());
			nrows	= atoi(rows.c_str());
			ncols	= atoi(cols.c_str());

			if (nrows == 0)
			{
				if (matrixIdentifier == "sm")
				{
					sm.clear();
				}

				if (matrixIdentifier == "ns")
				{
					ns.clear();
				}

				if(matrixIdentifier == "cv")
				{
					cv.clear();
				}

				if(matrixIdentifier == "em")
				{
					em.clear();
				}
				return -1;

			}

			//If matrixIdentifier is -sm extract stoichiometric matrix
			if(matrixIdentifier == "sm")
			{
				sm.resize(nrows);
				for(int k=0; k < nrows; k++)
				{
					sm[k].resize(ncols);
				}

				for(int i=0; i < nrows; i++)
				{
					getline(in_stream, str);
					stringstream ss(str); // Insert the string into a stream
					for(int j=0; j < ncols; j++)
					{
						ss >> sm[i][j];
					}

				}
			}

			//If matrixIdentifier is -ns extract Null Space vectors
			if(matrixIdentifier == "ns")
			{
				ns.resize(nrows);
				for(int k=0; k < nrows; k++)
				{
					ns[k].resize(ncols);
				}

				for(int i=0; i < nrows; i++)
				{
					getline(in_stream, str);
					stringstream ss(str); // Insert the string into a stream
					for(int j=0; j < ncols; j++)
					{
						ss >> ns[i][j];
					}

				}
			}

			//If matrixIdentifier is -cv extract Conservation realtions vectors
			if(matrixIdentifier == "cv")
			{
				cv.resize(nrows);
				for(int k=0; k < nrows; k++)
				{
					cv[k].resize(ncols);
				}

				for(int i=0; i < nrows; i++)
				{
					getline(in_stream, str);
					stringstream ss(str); // Insert the string into a stream
					for(int j=0; j < ncols; j++)
					{
						ss >> cv[i][j];
					}
				}
			}

			//If matrixIdentifier is -em extract elementary modes matrix
			if(matrixIdentifier == "em")
			{
				em.resize(nrows);
				for(int k=0; k < nrows; k++)
				{
					em[k].resize(ncols);
				}

				for(int i=0; i < nrows; i++)
				{
					getline(in_stream, str);
					stringstream ss(str); // Insert the string into a stream
					for(int j=0; j < ncols; j++)
					{
						ss >> em[i][j];
					}

				}
			}

			return 0; // Extract of matrix Successful

		}
		catch (SBWApplicationException *ae)
		{
			ae = new SBWApplicationException("Error in Matrix extraction", "Exception raised in Method: extractMatrix,  Module: metatoolSBW");
			throw ae;
		}
	}

	///
	///This Function fetches the number listed next to a marker
	int getIntValue (ifstream &infile_stream, string marker)
	{
		bool found = false;
		string str;
		int result = -1;
		do
		{
			getline(infile_stream, str);
			if (str.substr(0, marker.length()) == marker)
			{
				str = str.substr(marker.length(), str.length());
				result = atoi (str.c_str());
				found = true;
			}
		}  while(!found && !infile_stream.eof() );
		return result;
	}

	///
	///This method extracts the information from the output file
	void extractInformation()
	{
		bool found;
		string str;
		try
		{
#if defined(WIN32)
			str = AppPath + "\\" + OUTPUT_METATOOL_FILE;
#elif defined(linux)
			str = AppPath + "/" + OUTPUT_METATOOL_FILE;
#endif
			ifstream instream(str.c_str());

			if(!instream.is_open())
			{
		        SBWApplicationException *ae;
				ae = new SBWApplicationException("Unable to extract information from - " + OUTPUT_METATOOL_FILE,
					" Exception raised in Module:Metatool, Method: extractInformation");
				throw ae;
			}

			nInternalSpecies = getIntValue(instream, marker1);
			nExternalSpecies = getIntValue(instream, marker2);
			nReactions       = getIntValue(instream, marker3);

			found = false;
			do
			{
				getline(instream, str);
				if (str.substr(0, marker4.length()) == marker4)
				{
					extractMatrix (instream, "sm"); //extract Stoichiometric Matrix
					found = true;
				}
			}while(!found && !instream.eof());

			found = false;
			do
			{
				getline(instream, str);
				if (str.substr(0, marker6.length()) == marker6)
				{
					extractMatrix(instream, "ns"); //extract NullSpace Vectors
					found = true;
				}
			}while(!found && !instream.eof());

			found = false;
			do
			{
				getline(instream, str);
				if (str.substr(0, marker9.length()) == marker9)
				{
					extractMatrix (instream, "cv"); //extract conservartion Vectors
					found = true;
				}
			}while(!found && !instream.eof());

			found = false;
			do
			{
				getline(instream, str);
				if (str.substr(0, marker10.length()) == marker10)
				{
					extractMatrix (instream, "em"); //extract elementary modes matrix
					found = true;
				}
			}while(!found && !instream.eof());

			instream.close();
		}
		catch (SBWApplicationException *ae)
		{
			throw ae;
		}
	}

	/*******************************
	 *Here Starts the Public Methods
	 *******************************/
	public :
	static string AppPath;
	///
	///Get Application Path
	static void setApplicationPath(char* FullPath)
	{
		string FullPathStr;
		int AppPathLength;
		FullPathStr = FullPath;
#if defined(WIN32)
		AppPathLength = FullPathStr.rfind("\\");
#elif defined(linux)
		AppPathLength = FullPathStr.rfind("/");
#endif
		AppPath = FullPathStr.substr(0,AppPathLength);
	}

	///
	///Metatool Constructor
	Metatool()
	{
		nInternalSpecies  = 0;
		nExternalSpecies  = 0;
		nReactions        = 0;
		metaToolInputFile = "";
#if defined(WIN32)
		string applicationPath = AppPath + "\\" + metaToolExeName;
#elif defined(linux)
		string applicationPath = AppPath + "/" + metaToolExeName;
#endif
		strcpy(metaToolPath, applicationPath.c_str());
		nomModule = -1000;
	}

	/// this method provides the implementation of getVersion method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getVersion(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << METATOOLSBW_VERSION;
	}

	/// This method provides the implementation of setMetaToolPath method
	/// Sets the metatool path and exe name.
	/// Returns false if it fails to find the path.
	DataBlockWriter setMetaToolPath(Module from, DataBlockReader reader)
	{
		char* arg;
		int fileExistsFlag;
		
		reader >> arg;
		if(IsFileExist(arg))
		{
			fileExistsFlag = 1;
		}
		else
		{
			fileExistsFlag = 0;
		}
		strcpy(metaToolPath, arg);
		return DataBlockWriter() << fileExistsFlag;
	}

	/// This method provides the implementation of runMetatool method
	/// Run Metatool with the given metatool input file and wait for execution to
	/// complete. If the input file name is empty then the routine uses
	/// the default input metatool file name
	DataBlockWriter runMetatool(Module from, DataBlockReader reader)
	{
		DataBlockWriter OWrite;
		
		try
		{
			reader >> metaToolInputFile;
			runMetatool(metaToolInputFile);
			extractInformation();
		}
		catch(SBWApplicationException* ae)
		{
			//Clear the values of variables set in previous runs.
			nInternalSpecies = 0;
			nExternalSpecies = 0;
			nReactions = 0;
			sm.clear();
			em.clear();
			ns.clear();
			cv.clear();
			throw ae;
		}
		return OWrite;
	}
	
	/// This method provides the implementation of runMetatool method
	/// Run Metatool with the given SBML string and wait for execution to
	/// complete.
	DataBlockWriter runMetatoolWithSBML(Module from, DataBlockReader reader)
	{
		DataBlockWriter OWrite;
		DataBlockWriter OSbml;
		string sbmlInput;
		string inputFileName;
		
		try
		{
			reader >> sbmlInput;
			CHECK_NOM;
			ServiceMethod loadSBMLMethod = nomService.getMethod("void loadSBML(string)");
			OSbml << sbmlInput;
			loadSBMLMethod.call(OSbml);
			inputFileName = MakeMetaToolFile();
			runMetatool(inputFileName);
			extractInformation();
		}
		catch(SBWApplicationException* ae)
		{
			//Clear the values of variables set in previous runs.
			nInternalSpecies = 0;
			nExternalSpecies = 0;
			nReactions = 0;
			sm.clear();
			em.clear();
			ns.clear();
			cv.clear();
			throw ae;
		}
		return OWrite;
	}

	/// This method provides the implementation of runMetatoolFromNOM method
	/// Run Metatool with the given metatool input file and wait for execution to
	/// complete. If the input file name is empty then the routine uses
	/// the default input metatool file name
	DataBlockWriter runMetatoolFromNOM(Module from, DataBlockReader reader)
	{
		DataBlockWriter OWrite;
		string inputFileName;
		try
		{
			inputFileName = MakeMetaToolFile();
			runMetatool(inputFileName);
			extractInformation();
		}
		catch(SBWApplicationException* ae)
		{
			//Clear the values of variables set in previous runs.
			nInternalSpecies = 0;
			nExternalSpecies = 0;
			nReactions = 0;
			sm.clear();
			em.clear();
			ns.clear();
			cv.clear();
			throw ae;
		}
		return OWrite;
	}

	/// this method provides the implementation of getNumReactions method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getNumReactions(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << nReactions;
	}

	/// this method provides the implementation of getNumInternalSpecies method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getNumInternalSpecies(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << nInternalSpecies;
	}

	/// this method provides the implementation of getNumExternalSpecies method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getNumExternalSpecies(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << nExternalSpecies;
	}

	/// this method provides the implementation of getStoichiometryMatrix method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getStoichiometryMatrix(Module from, DataBlockReader reader)
	{
		double** sm_array;
		DataBlockWriter OWrite;
		int nrows;
		int ncolumns;

		nrows = sm.size();
		//----If the Stoichiometry matrix is empty,
		//    then return empty DataBlockWriter.
		if(nrows == 0)
		{
			unsigned char message[14];
			message[0] = (unsigned char) 5;
			message[1] = (unsigned char) 2;
			message[2] = (unsigned char) 2;
			for(int i=3; i <14; i++)
			{
				message[i] = (unsigned char) 0;
			}
			OWrite.insertData(message, 14);		}
		else
		{
			sm_array = (double** )malloc(nrows * sizeof(double *));
			if(sm_array == NULL)
			{
				SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
					"Exception raised in Module:Metatool, Method:getStoichiometryMatrix");
				throw ae;
			}

			ncolumns = sm[0].size();
			for(int i=0; i < nrows; i++)
			{
				sm_array[i] = (double* ) malloc(ncolumns * sizeof(double));
				if(sm_array[i] == NULL)
				{
					SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
						"Exception raised in Module:Metatool, Method:getStoichiometryMatrix");
					throw ae;

				}
			}

			vector < vector<double> >::iterator sm_rowiterator;
			vector <double>::iterator sm_coliterator;
			int row=0;
			int col;
			for(sm_rowiterator = sm.begin(); sm_rowiterator != sm.end(); sm_rowiterator++)
			{
				col=0;
				for(sm_coliterator = (*sm_rowiterator).begin(); sm_coliterator != (*sm_rowiterator).end(); sm_coliterator++)
				{
					sm_array[row][col] = (*sm_coliterator);
					col++;
				}
				row++;
			}
			OWrite.add(nrows, ncolumns, sm_array);
		}
		return OWrite;
	}

	/// this method provides the implementation of getElementaryModes method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getElementaryModes(Module from, DataBlockReader reader)
	{
		double** em_array;
		DataBlockWriter OWrite;
		int nrows;
		int ncolumns;

		nrows = em.size();
		//----If the ElementaryModes matrix is empty,
		//    then return empty DataBlockWriter.
		if(em.size() == 0)
		{
			unsigned char message[14];
			message[0] = (unsigned char) 5;
			message[1] = (unsigned char) 2;
			message[2] = (unsigned char) 2;
			for(int i=3; i <14; i++)
			{
				message[i] = (unsigned char) 0;
			}
			OWrite.insertData(message, 14);
		}
		else
		{
			em_array = (double** )malloc(nrows * sizeof(double *));
			if(em_array == NULL)
			{
				SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
					"Exception raised in Module:Metatool, Method:getElementaryModes");
				throw ae;
			}

			ncolumns = em[0].size();
			for(int i=0; i < nrows; i++)
			{
				em_array[i] = (double* ) malloc(ncolumns * sizeof(double));
				if(em_array[i] == NULL)
				{
					SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
						"Exception raised in Module:Metatool, Method:getElementaryModes");
					throw ae;

				}
			}

			vector < vector<double> >::iterator em_rowiterator;
			vector <double>::iterator em_coliterator;
			int row=0;
			int col;
			for(em_rowiterator = em.begin(); em_rowiterator != em.end(); em_rowiterator++)
			{
				col=0;
				for(em_coliterator = (*em_rowiterator).begin(); em_coliterator != (*em_rowiterator).end(); em_coliterator++)
				{
					em_array[row][col] = (*em_coliterator);
					col++;
				}
				row++;
			}
			OWrite.add(nrows, ncolumns, em_array);
		}
		return OWrite;
	}

	/// this method provides the implementation of getNullSpaceVectors method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getNullSpaceVectors(Module from, DataBlockReader reader)
	{
		double** ns_array;
		DataBlockWriter OWrite;
		int nrows;
		int ncolumns;

		nrows = ns.size();
		//----If the NullSpace vectors matrix is empty,
		//    then return empty DataBlockWriter.
		if(ns.size() == 0)
		{
			unsigned char message[14];
			message[0] = (unsigned char) 5;
			message[1] = (unsigned char) 2;
			message[2] = (unsigned char) 2;
			for(int i=3; i <14; i++)
			{
				message[i] = (unsigned char) 0;
			}
			OWrite.insertData(message, 14);
		}
		else
		{
			ns_array = (double** )malloc(nrows * sizeof(double *));
			if(ns_array == NULL)
			{
				SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
					"Exception raised in Module:Metatool, Method:getNullSpaceVectors");
				throw ae;
			}

			ncolumns = ns[0].size();
			for(int i=0; i < nrows; i++)
			{
				ns_array[i] = (double* ) malloc(ncolumns * sizeof(double));
				if(ns_array[i] == NULL)
				{
					SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
						"Exception raised in Module:Metatool, Method:getNullSpaceVectors");
					throw ae;
				}
			}

			vector < vector<double> >::iterator ns_rowiterator;
			vector <double>::iterator ns_coliterator;
			int row=0;
			int col;
			for(ns_rowiterator = ns.begin(); ns_rowiterator != ns.end(); ns_rowiterator++)
			{
				col=0;
				for(ns_coliterator = (*ns_rowiterator).begin(); ns_coliterator != (*ns_rowiterator).end(); ns_coliterator++)
				{
					ns_array[row][col] = (*ns_coliterator);
					col++;
				}
				row++;
			}
			OWrite.add(nrows, ncolumns, ns_array);
		}
		return OWrite;
	}

	/// this method provides the implementation of getConservationVectors method
	/// arguments: from - module calling this method
	/// reader - object containing argument data
	/// returns: object containing result data
	DataBlockWriter getConservationVectors(Module from, DataBlockReader reader)
	{
		double** cv_array;
		DataBlockWriter OWrite;
		int nrows;
		int ncolumns;

		nrows = cv.size();
		//----If the NullSpace vectors matrix is empty,
		//    then return empty DataBlockWriter.
		if(nrows == 0)
		{
			unsigned char message[14];
			message[0] = (unsigned char) 5;
			message[1] = (unsigned char) 2;
			message[2] = (unsigned char) 2;
			for(int i=3; i <14; i++)
			{
				message[i] = (unsigned char) 0;
			}
			OWrite.insertData(message, 14);
		}
		else
		{
			cv_array = (double** )malloc(nrows * sizeof(double *));
			if(cv_array == NULL)
			{
				SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
					"Exception raised in Module:Metatool, Method:getConservationVectors");
				throw ae;
			}

			ncolumns = cv[0].size();
			for(int i=0; i < nrows; i++)
			{
				cv_array[i] = (double* ) malloc(ncolumns * sizeof(double));
				if(cv_array[i] == NULL)
				{
					SBWApplicationException* ae = new SBWApplicationException("Error occured when trying to allocate memory",
						"Exception raised in Module:Metatool, Method:getConservationVectors");
					throw ae;
				}
			}

			vector < vector<double> >::iterator cv_rowiterator;
			vector <double>::iterator cv_coliterator;
			int row=0;
			int col;
			for(cv_rowiterator = cv.begin(); cv_rowiterator != cv.end(); cv_rowiterator++)
			{
				col=0;
				for(cv_coliterator = (*cv_rowiterator).begin(); cv_coliterator != (*cv_rowiterator).end(); cv_coliterator++)
				{
					cv_array[row][col] = (*cv_coliterator);
					col++;
				}
				row++;
			}
			OWrite.add(nrows, ncolumns, cv_array);
		}

		return OWrite;
	}

	///
	///This method gets the list of Floating Species
	DataBlockWriter getInternalList(Module from, DataBlockReader reader)
	{
		DataBlockWriter internalList;
		bool found;
		string str;
		string buffer;

		if (metaToolInputFile == "")
		{
#if defined(WIN32)
			str = AppPath + "\\" + INPUT_METATOOL_FILE;
#elif defined(linux)
        	str = AppPath + "/" + INPUT_METATOOL_FILE;
#endif
		}
		else
		{
			str = metaToolInputFile;
		}

		ifstream instream(str.c_str());

		if(!instream.is_open())
		{
	                SBWApplicationException *ae;
			ae = new SBWApplicationException("Unable to read - " + INPUT_METATOOL_FILE,
				" Exception raised in Module:Metatool, Method: getInternalList");
			throw ae;
		}

		found = false;
		do
		{
			getline(instream, str);
			if (str.substr(0, marker13.length()) == marker13)
			{
			        getline(instream, str);
				stringstream ss(str);
				while (ss >> buffer)
					internalList << buffer;
				found = true;
			}
		}while(!found && !instream.eof());

		instream.close();
		return DataBlockWriter() << internalList;
	}

	///
	///This method gets the list of boundary species
	DataBlockWriter getExternalList(Module from, DataBlockReader reader)
	{
		DataBlockWriter externalList;
		bool found;
		string str;
		string buffer;

		if (metaToolInputFile == "")
		{
#if defined(WIN32)
			str = AppPath + "\\" + INPUT_METATOOL_FILE;
#elif defined(linux)
        		str = AppPath + "/" + INPUT_METATOOL_FILE;
#endif
		}
		else
		{
			str = metaToolInputFile;
		}
		ifstream instream(str.c_str());

		if(!instream.is_open())
		{
	                SBWApplicationException *ae;
			ae = new SBWApplicationException("Unable to read - " + str,
				" Exception raised in Module:Metatool, Method: getExternalList");
			throw ae;
		}

		found = false;
		do
		{
			getline(instream, str);
			if (str.substr(0, marker14.length()) == marker14)
			{
			        getline(instream, str);
				stringstream ss(str);
				while (ss >> buffer)
					externalList << buffer;
				found = true;
			}
		}while(!found && !instream.eof());

		instream.close();
		return DataBlockWriter() << externalList;
	}

	///
	///This method gets the list of reaction names
	DataBlockWriter getReactionNameList(Module from, DataBlockReader reader)
	{
		DataBlockWriter reactionNameList;
		bool found;
		string str;
		string buffer;

		if (metaToolInputFile == "")
		{
#if defined(WIN32)
			str = AppPath + "\\" + INPUT_METATOOL_FILE;
#elif defined(linux)
        	str = AppPath + "/" + INPUT_METATOOL_FILE;
#endif
		}
		else
		{
			str = metaToolInputFile;
		}

		ifstream instream(str.c_str());

		if(!instream.is_open())
		{
	        SBWApplicationException *ae;
			ae = new SBWApplicationException("Unable to read - " + INPUT_METATOOL_FILE,
				" Exception raised in Module:Metatool, Method: getReactionNameList");
			throw ae;
		}

		found = false;
		do
		{
			getline(instream, str);
			if (str.substr(0, marker11.length()) == marker11)
			{
			        getline(instream, str);
				stringstream ss(str);
				while (ss >> buffer)
					reactionNameList << buffer;
				found = true;
			}
		}while(!found && !instream.eof());

		found = false;
		do
		{
			getline(instream, str);
			if (str.substr(0, marker12.length()) == marker12)
			{
			        getline(instream, str);
				stringstream ss(str);
				while (ss >> buffer)
					reactionNameList << buffer;
				found = true;
			}
		}while(!found && !instream.eof());
		instream.close();

		return DataBlockWriter() << reactionNameList;
	}

	/**
	* register methods of this class with a module implementation represented by a
	* given MethodTable
	* argument - table object that provides a simple interface to a module implementation
	*/
	static void registerMethods(MethodTable<Metatool> &table)
	{
		table.addMethod(&Metatool::getVersion, "string getVersion()");
		table.addMethod(&Metatool::setMetaToolPath, "int setMetaToolPath(string)");
		table.addMethod(&Metatool::runMetatool, "void runMetatool(string)");
		table.addMethod(&Metatool::runMetatoolWithSBML, "void runMetatoolWithSBML(string)");
		table.addMethod(&Metatool::runMetatoolFromNOM, "void runMetatoolFromNOM()");
		table.addMethod(&Metatool::getNumReactions, "int getNumReactions()");
		table.addMethod(&Metatool::getNumInternalSpecies, "int getNumInternalSpecies()");
		table.addMethod(&Metatool::getNumExternalSpecies, "int getNumExternalSpecies()");
		table.addMethod(&Metatool::getStoichiometryMatrix, "double[][] getStoichiometryMatrix()");
		table.addMethod(&Metatool::getElementaryModes, "double[][] getElementaryModes()");
		table.addMethod(&Metatool::getNullSpaceVectors, "double[][] getNullSpaceVectors()");
		table.addMethod(&Metatool::getConservationVectors, "double[][] getConservationVectors()");
		table.addMethod(&Metatool::getInternalList, "{} getInternalList()");
		table.addMethod(&Metatool::getExternalList, "{} getExternalList()");
		table.addMethod(&Metatool::getReactionNameList, "{} getReactionNameList()");
	}
};

/*
*This class provides an implementation of biospiceAnalyzer service
*/
class BiospiceAnalyze
{
private:
	//*******************************
	//Here starts the public methods:
	//*******************************
public:
	DataBlockWriter mbiospiceInitialise(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << true;
	}

	DataBlockWriter mbiospiceAnalyzeGetName(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << "Data Generator";
	}

	DataBlockWriter mbiospiceAnalyzeGetDisplayName(Module from, DataBlockReader reader)
	{
		return DataBlockWriter() << "Data Generator";
	}

	DataBlockWriter mbiospiceAnalyzeGetInputTypes(Module from, DataBlockReader reader)
	{
		DataBlockWriter inputList;
		inputList << "sbmlL1";
		return DataBlockWriter() << inputList;
	}

	DataBlockWriter mbiospiceAnalyzeGetOutputTypes(Module from, DataBlockReader reader)
	{
		DataBlockWriter outputList;
		outputList << "Matrix2D";
		return DataBlockWriter() << outputList;
	}

	DataBlockWriter mbiospiceAnalyzeGetConfigurationParams(Module from, DataBlockReader reader)
	{
		DataBlockWriter configurationParams;
		return DataBlockWriter() << configurationParams;
	}

	DataBlockWriter mbiospiceAnalyze(Module from, DataBlockReader reader)
	{
		DataBlockWriter OWrite;
		DataBlockWriter resultWrite;
		DataBlockWriter OSbml;
		DataBlockWriter result;
		DataBlockReader sbmlRead;
		string sbmlInput;
		string inputFileName;
		int rows, cols;
		double** read_array;
		try
		{
			reader >> sbmlRead;
			sbmlRead >> sbmlInput;
			
			Module metatoolModule = SBW::getModuleInstance("metatool");
			Service metatoolService = metatoolModule.findServiceByName("mt");
			
			ServiceMethod runMetatoolWithSBMLMethod = metatoolService.getMethod("void runMetatoolWithSBML(string)");
			OSbml << sbmlInput;
			runMetatoolWithSBMLMethod.call(OSbml);
			
			ServiceMethod getElementaryModesMethod = metatoolService.getMethod("double[][] getElementaryModes()");
			DataBlockReader ORead                  = getElementaryModesMethod.call(OWrite);
			ORead.get(rows, cols, read_array);

			ServiceMethod getReactionNameListMethod = metatoolService.getMethod("{} getReactionNameList()");
			DataBlockReader OReadList               = getReactionNameListMethod.call(OWrite);

			DataBlockWriter emColumnList;
			string emColumnName;
			DataBlockReader reactionNameList;
			OReadList.get(reactionNameList);
			DataBlockType OType = reactionNameList.getNextType();

			while(OType != SBWTerminateType)
			{
				reactionNameList.get(emColumnName);
				emColumnList << emColumnName;
				OType = reactionNameList.getNextType();
			}
						
			if (rows!=0)
			{
				result << emColumnList;
				result.add(rows, cols, read_array);
				resultWrite << result;
			}
			else
			{	//Instead of returning an empty two dimensional array
				//let us return a a two dimensional array with single
				//element and column label set to "Empty".
				emColumnName = "Empty";
				emColumnList << emColumnName;
				result << emColumnList;
				read_array = (double** )malloc(1 * sizeof(double *));
				read_array[0] = (double* ) malloc(1 * sizeof(double));
				read_array[0][0] = 0.0;
				result.add(1,1,read_array);
				resultWrite << result;
			}

			ORead.free2DArray(rows, read_array);
		}
		catch(SBWApplicationException* ae)
		{
			throw ae;
		}
		return resultWrite;
	}


	/**
	* register methods of this biospiceAnalyze with a module implementation represented by a
	* given MethodTable
	* argument - table object that provides a simple interface to a module implementation
	*/
	static void registerMethods(MethodTable<BiospiceAnalyze> &table)
	{
		table.addMethod(&BiospiceAnalyze::mbiospiceInitialise, "int initialize()");
		table.addMethod(&BiospiceAnalyze::mbiospiceAnalyzeGetName, "string getName()");
		table.addMethod(&BiospiceAnalyze::mbiospiceAnalyzeGetDisplayName, "string getDisplayName()");
		table.addMethod(&BiospiceAnalyze::mbiospiceAnalyzeGetInputTypes, "{} getInputTypes()");
		table.addMethod(&BiospiceAnalyze::mbiospiceAnalyzeGetOutputTypes, "{} getOutputTypes()");
		table.addMethod(&BiospiceAnalyze::mbiospiceAnalyzeGetConfigurationParams, "{} getConfigurationParams()");
		table.addMethod(&BiospiceAnalyze::mbiospiceAnalyze, "{} analyze({})");
	}
};

/*
* module entry point WinMain/main
* On windows using WinMain as an entry point means that the executable can be an
* invisible process rather than a console application.
* arguments: in windows case all are ignored, see windows documentation
* for details of arguments. In linux case standard main arguments
* result: int : -1 if error occurred 0 otherwise.
*/
#if defined(WIN32)
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
LPSTR lpCmdLine, int nCmdShow)
{
	int argc ;
	char **argv ;
	ModuleImpl::windowsExtractCommandLine(&argc, &argv); /* get windows commandline */
#elif defined(linux)
int main(int argc, char* argv[])
{
#endif
	try
	{
		Metatool::setApplicationPath(argv[0]);

		ModuleImpl modImpl(
		"metatool", // module identification
		"metatool written in C++", // humanly readable name
		UniqueModule); // management scheme

		modImpl.addServiceObject(
		"mt", // service identification
		"mt written in C++", // humanly readable name
		"mt", // category
		new Metatool()); // service implementation

		modImpl.addServiceObject(
		"biospiceAnalyze", // service identification
		"BioSPICE/Analyze", // humanly readable name
		"biospice", // category
		new BiospiceAnalyze()); // service implementation

		// connect to broker providing services
		modImpl.run(argc, argv);
	}
	catch (SBWException *e)
	{
		#if defined(WIN32)
			MessageBox(
				NULL, e->getMessage().c_str(), "Metatool", MB_OK | MB_ICONEXCLAMATION);
		#elif defined(linux)
			fprintf(stderr, "Metatool exception: %s\n", e->getMessage().c_str());
		#endif
			return -1;
	}
	return 0;
}
string Metatool::AppPath;
